# _sqlite.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Object database using Sqlite.

Segmented databases take account of the local density of values per key on
secondary databases.  A primary database, always with a single column integer
primary key, is seen as a sequence of fixed size intervals called segments.
Secondary databases have zero or one keys per segment for each key.  The value
associated with each segment key is a number, a list of numbers, or a bit map,
representing the record numbers in the segment referenced by the key.  The
absence of a segment key means the key has no values in that segment.  Bit maps
are fixed length and record lists are variable length.  The maximum byte size
of a list is less than or equal to the byte size of a bit map.

Idea taken from DPT, an emulation of Model 204 which runs on Microsoft Windows.

The primary database and secondary database terms are from Berkeley DB.

The representation of multiple rows of a table by one row makes it at least
difficult to use SQL statements directly to evaluate queries.  But the reason
for the sqlite3 interface is it's status as the only cross-platform database
engine included in the Python 3 distribution.  Here SQL statements are used to
build emulations of some of the Berkeley DB interface available at Python 2.

"""

import os
import subprocess
from ast import literal_eval
import bisect
import re

import sys
_platform_win32 = sys.platform == 'win32'
_python_version = '.'.join(
    (str(sys.version_info[0]),
     str(sys.version_info[1])))
del sys

from .api.bytebit import Bitarray
from .api.database import (
    DatabaseError,
    Database,
    )
from .api.cursor import (
    Cursor,
    )
from .api.recordset import (
    SegmentBitarray,
    SegmentInt,
    SegmentList,
    RecordsetCursor,
    )
from .api.constants import (
    FLT, SPT,
    FILEDESC, FILEORG, EO,
    PRIMARY_FIELDATTS, SECONDARY_FIELDATTS, SQLITE3_FIELDATTS,
    FIELDS,
    PRIMARY, SECONDARY, INDEXPREFIX, SEGMENTPREFIX,
    SQLITE_VALUE_COLUMN,
    SQLITE_SEGMENT_COLUMN,
    SQLITE_COUNT_COLUMN,
    SQLITE_RECORDS_COLUMN,
    LENGTH_SEGMENT_BITARRAY_REFERENCE,
    LENGTH_SEGMENT_LIST_REFERENCE,
    SUBFILE_DELIMITER,
    )
from .api.segmentsize import SegmentSize


class Sqlite3apiError(DatabaseError):
    pass


class _DatabaseEncoders(object):
    
    """Define default record key encoder and decoder.
    """

    def encode_record_number(self, key):
        """Return repr(key) because this is sqlite3 version of method.

        Typically used to convert primary key to secondary index format,
        using Berkeley DB terminology.
        
        """
        return repr(key)

    def decode_record_number(self, skey):
        """Return literal_eval(skey) because this is sqlite3 version of method.

        Typically used to convert DB primary key held on secondary index,
        using Berkeley DB terminology, to integer.

        """
        return literal_eval(skey)

    def encode_record_selector(self, key):
        """Return key because this is sqlite3 version of method.

        Typically used to convert a secondary index key value to database
        engine format, using Berkeley DB terminology.
        
        """
        return key

    def encode_record_key(self, key):
        """Return key because this is sqlite3 version of method.
        
        Specifically when comparing the key of a record read from database with
        the key used to find it.
        
        """
        return key


class Sqlite3base(Database, _DatabaseEncoders):
    
    """Define a Berkeley DB-like database structure using sqlite3.
    
    Primary databases are created as 'integer primary key'.
    Secondary databases are:
    'index, segment, reference, count, primary key (index, segment)'.

    Primary and secondary terminology comes from Berkeley DB documentation.

    segment moves from the value part of a secondary key:value, (segment,
    reference, count), to the primary key definition in sqlite3 where segment
    follows DPT terminology.  Reference can be a record number relative to
    segment start, a reference to a list of record numbers, or a reference to a
    bitmap representing such record numbers.  Count is the number of records
    referenced by this value.

    Secondary databases are supported by an 'integer primary key' table, for
    lists of record numbers or bitmap representations of record numbers.  The
    reference is the key into the relevant table.
    
    """

    def __init__(
        self,
        primary_class,
        secondary_class,
        database_specification,
        databasefolder,
        *args,
        **kwargs):
        """Define database structure.
        
        database_specification = {
            name:{
                primary:name,
                fields:{
                    name:{property:value, ...},
                    ),
                }, ...
            }
        databasefolder = folder containing database file
        record_class = class implementing access to databases (sqlite3 tables)

        """
        super(Sqlite3base, self).__init__(
            database_specification,
            databasefolder,
            primary_class=primary_class,
            secondary_class=secondary_class,
            **kwargs)

        # sqlite3 database connection object
        self._sqconn = None

        # database_specification processing

        definitions = {k for k in self.get_dbsets()}

        for dbset, definition in self.get_dbsets().items():
            for s in definition.secondary:
                if s in definitions:
                    msg = ' '.join(['Secondary table name', s,
                                    'for', dbset,
                                    'is used in another definition'])
                    raise Sqlite3apiError(msg)
                definitions.add(s)
        self._control = Sqlite3ControlFile()
        for definition in self.get_dbsets().values():
            definition.primary.set_control_database(self._control)

        # The file from databasefolder after validation
        self._home = os.path.join(self._databasefolder,
                                  os.path.split(self._databasefolder)[-1])

    def make_root(self, **kw):
        """Return Sqlite3Definition instance created from **kw arguments.
        """
        return _Sqlite3Definition(**kw)

    @property
    def connection(self):
        """sqlite3.connection object for database."""
        return self._sqconn

    def close_context(self):
        """Close all sqlite3 cursors."""
        if self._sqconn is None:
            return
        for table in self.get_dbsets().values():
            table.primary.close()
            for s in table.secondary.values():
                s.close()

    def close_database(self):
        """Close connection to database."""
        if self._sqconn is None:
            return
        self.close_context()
        self._sqconn.close()
        self._sqconn = None

    def start_transaction(self):
        """Start a transaction."""
        if self._sqconn:
            cursor = self._sqconn.cursor()
            try:
                cursor.execute('begin')
            finally:
                cursor.close()

    def backout(self):
        """Backout tranaction."""
        if self._sqconn:
            cursor = self._sqconn.cursor()
            try:
                cursor.execute('rollback')
            finally:
                cursor.close()
            
    def commit(self):
        """Commit tranaction."""
        if self._sqconn:
            cursor = self._sqconn.cursor()
            try:
                cursor.execute('commit')
            finally:
                cursor.close()

    def db_compatibility_hack(self, record, srkey):
        """Convert to (key, value) format returned by Berkeley DB access.

        sqlite3 is compatible with the conventions for Berkeley DB RECNO
        databases except for a Berkeley DB index where the primary key is not
        held as the value on an index record (maybe the primary key is embedded
        in the secondary key). Here the Berkeley DB index record is (key, None)
        rather than (key, primary key). The correponding sqlite3 structure is
        always (index field value, record number).
        DataClient works to Berkeley DB conventions.
        The user code side of DataClient adopts the appropriate Berkeley DB
        format because it defines the format used. The incompatibility that
        comes from mapping a (key, None) to sqlite3 while using the same user
        code is dealt with in this method.

        """
        key, value = record
        if value is None:
            return (key, self.decode_record_number(srkey))
        else:
            return record

    def files_exist(self):
        """Return True if all defined files exist in self._home folder."""
        return os.path.exists(self._home)

    def get_associated_indicies(self, dbset):
        """Return frozenset of all table indicies for dbset."""
        s = self.get_dbsets()[dbset]
        return frozenset(set((s._dbset,)).union(s.secondary))

    def get_database_filename(self):
        """Return database file name."""
        return self._home

    def get_database_folder(self):
        """return database folder name."""
        return os.path.dirname(self._home)
    
    def get_database(self, dbset, dbname):
        """Return DB for dbname in dbset."""
        return self.get_dbsets()[dbset].primary.sq_conn

    def is_recno(self, dbset, dbname):
        """Return True if dbname is primary table in dbset."""
        return self.get_dbsets()[dbset].associate(dbname).is_primary()

    def open_context(self):
        """Open all tables on database."""
        f, b = os.path.split(self._home)
        if not os.path.exists(f):
            os.makedirs(f, mode=0o700)
        self.make_connection()
        for table in self.get_dbsets().values():
            table.primary.open_root(self.connection)
            for s in table.secondary.values():
                s.open_root(self.connection)
        return True

    def allocate_and_open_contexts(self, closed_contexts):
        """Open closed_contexts which had been closed.

        This method is intended for use only when re-opening a table after
        closing it temporarily so another thread can make and commit changes
        to a table.

        The method name comes from the dptbase module where it describes
        extactly what is done.  For sqlite3api it is a callback hook where the
        name is already chosen.

        """
        self.open_context()

    def get_packed_key(self, dbset, instance):
        """Convert instance.key for use as database value.

        For sqlite3 just return instance.key.pack().
        dbname is relevant to Berkeley DB and retained for compatibility.

        """
        return instance.key.pack()

    def decode_as_primary_key(self, dbset, pkey):
        """Convert pkey for use as database key.

        For sqlite3 just return integer form of pkey.

        """
        #KEYCHANGE
        # Avoid isinstance test?
        if isinstance(pkey, int):
            return pkey
        else:
            return self.decode_record_number(pkey)

    def encode_primary_key(self, dbset, instance):
        """Convert instance.key for use as database value.

        For sqlite3 just return self.get_packed_key() converted to string.

        """
        # Should this be like Berkeley DB version of method?
        return self.encode_record_number(self.get_packed_key(dbset, instance))

    def do_database_task(
        self,
        taskmethod,
        logwidget=None,
        taskmethodargs={},
        use_specification_items=None,
        ):
        """Open new connection to database, run method, then close database.

        This method is intended for use in a separate thread from the one
        dealing with the user interface.  If the normal user interface thread
        also uses a separate thread for it's normal, quick, database actions
        there is probably no need to use this method at all.

        """
        db = self.__class__(
            self.get_database_folder(),
            use_specification_items=use_specification_items)
        db.open_context()
        try:
            taskmethod(db, logwidget, **taskmethodargs)
        finally:
            db.close_database()

    def do_deferred_updates(self, pyscript, filepath):
        """Invoke a deferred update process and wait for it to finish.

        pyscript is the script to do the deferred update.
        filepath is a file or a sequence of files containing updates.

        """
        if _platform_win32:
            args = ['pythonw']
        else:
            args = [''.join(('python', _python_version))]
        
        if not os.path.isfile(pyscript):
            msg = ' '.join([repr(pyscript),
                            'is not an existing file'])
            raise Sqlite3apiError(msg)

        args.append(pyscript)
        
        if isinstance(filepath, str):
            filepath = (filepath,)
        for fp in filepath:
            if not os.path.isfile(fp):
                msg = ' '.join([repr(fp),
                                'is not an existing file'])
                raise Sqlite3apiError(msg)

        args.append(os.path.abspath(os.path.dirname(self._home)))
        args.extend(filepath)

        return subprocess.Popen(args)

    def cede_contexts_to_process(self, close_contexts):
        """Close all contexts so another process, or thread, can commit.

        close_contexts is ignored by this module's version of the method.

        """
        # Closing just the tables in close_contexts seems to be insufficient
        # and the complete shutdown in close_database() seems unnecessary.
        self.close_context()

    def create_recordset_cursor(self, recordset):
        """Create and return a cursor for this recordset."""
        return RecordsetCursorSqlite3(recordset)


class Sqlite3api(object):
    
    """Define a Berkeley DB-like database structure using Sqlite."""

    def close_context(self):
        """Close main and deferred update databases and environment."""
        self._control.close()
        super().close_context()

    def open_context(self):
        """Open all DBs."""
        super().open_context()
        self._control.open_root(self.connection)
        return True

    def do_deferred_updates(self, pyscript, filepath):
        """Invoke a deferred update process and wait for it to finish.

        pyscript is the script to do the deferred update.
        filepath is a file or a sequence of files containing updates.

        """
        if _platform_win32:
            args = ['pythonw']
        else:
            args = [''.join(('python', _python_version))]
        
        if not os.path.isfile(pyscript):
            msg = ' '.join([repr(pyscript),
                            'is not an existing file'])
            raise Sqlite3apiError(msg)

        args.append(pyscript)
        
        if isinstance(filepath, str):
            filepath = (filepath,)
        for fp in filepath:
            if not os.path.isfile(fp):
                msg = ' '.join([repr(fp),
                                'is not an existing file'])
                raise Sqlite3apiError(msg)

        args.append(os.path.abspath(os.path.dirname(self._home)))
        args.extend(filepath)

        return subprocess.Popen(args)

            
class Sqlite3File(object):
    
    """Define a sqlite3 table with open_root and close methods.
    """

    def __init__(self, name, sqlite3desc, primaryname):
        """Define a sqlite3 table.
        
        name = table description name
        sqlite3desc = description of related tables
        primaryname = primary table description name

        Interpret primary in the Berleley DB sense of primary and secondary
        databases for the relationship between primaryname and name.
        
        """
        super(Sqlite3File, self).__init__()

        self._fd_name = name
        self._primaryname = primaryname
        self._primary = name == primaryname
        self._fieldatts = dict()
        self._sq_conn = None
        self._object = None # a sqlite3 cursor on a table
        
        if self._primary:
            fieldatts = PRIMARY_FIELDATTS
        else:
            fieldatts = SECONDARY_FIELDATTS
        for attr in SQLITE3_FIELDATTS:
            if attr in fieldatts:
                self._fieldatts[attr] = fieldatts[attr]
        description = sqlite3desc[FIELDS][name]
        if description == None:
            description = dict()
        if not isinstance(description, dict):
            msg = ' '.join(['Attributes for column', repr(name),
                            'in table', repr(self.dataname),
                            "must be a 'dict' or 'None'"])
            raise Sqlite3apiError(msg)
        
        for attr in description:
            if attr not in fieldatts:
                msg = ' '.join(['Attribute', repr(attr),
                                'for column', repr(name),
                                'in table', self.dataname,
                                'is not allowed'])
                raise Sqlite3apiError(msg)
            
            if not isinstance(description[attr], type(fieldatts[attr])):
                msg = ' '.join([attr, 'for column', repr(name),
                                'in table', self.dataname, 'is wrong type'])
                raise Sqlite3apiError(msg)
            
            if attr == SPT:
                if (description[attr] < 0 or
                    description[attr] > 100):
                    msg = ' '.join(['Split percentage for field',
                                    repr(name), 'in file', self.dataname,
                                    'is invalid'])
                    raise Sqlite3apiError(msg)

            if attr in SQLITE3_FIELDATTS:
                self._fieldatts[attr] = description[attr]

        # When using either pickle.dumps() and pickle.loads(). or repr() and
        # ast.literal_eval(), the stored bytestring is returned for consistency
        # with the other database engines.
        #sqlite3.register_converter('TEXT', lambda bs: bs)

    def close(self):
        """Close DB and cursor."""
        # The _connection fills the role of DB
        self._sq_conn = None
        self._object = None

    def open_root(self, sqconn):
        """Open sqlite3 database cursor and create table unless it exists."""
        self._sq_conn = sqconn
        self._object = [self.dataname]

    @property
    def sq_object(self):
        """Return database table name."""
        return self._object[0]

    @property
    def sq_object_list(self):
        """Return list of database table names."""
        return self._object

    @property
    def dataname(self):
        """Return database file name."""
        return self._fd_name

    def is_primary(self):
        """Return True."""
        return self._primary

    @property
    def primaryname(self):
        """Return primary column name.

        Usually the name of the primary table, used in secondary tables as
        column name holding rowid reference to primary table row.

        """
        return self._primaryname

    @property
    def sq_conn(self):
        """Return sqlite3 database connection."""
        return self._sq_conn

            
class Sqlite3PrimaryFile(Sqlite3File):
    
    """Define a sqlite3 table with open_root and close methods.
    """

    def __init__(self, name, sqlite3desc, primaryname):
        """Bitmapped primary database file for name in description."""
        super().__init__(name, sqlite3desc, primaryname)
        self._segmentname = ''.join((SEGMENTPREFIX, self.dataname))

        # Interpret EO (from DPT) as 'integer primary key autoincrement'
        if sqlite3desc[FILEDESC][FILEORG] == EO:
            self._autoincrementprimary = True
        else:
            self._autoincrementprimary = False

        # Description to be provided
        self._control_database = None

        # Existence bit map control structure (reuse record numbers)
        self._control_primary = FileControlPrimary(self)

        # Record number existence bit map for this primary database
        self._existence_bits = Sqlite3ExistenceBitMap(self.dataname)

    def open_root(self, *args):
        """Open sqlite3 database cursor and create table unless it exists."""
        super(Sqlite3PrimaryFile, self).open_root(*args)
        cursor = self.sq_conn.cursor()
        try:
            if self._autoincrementprimary:
                statement = ' '.join((
                    'create table if not exists', self.sq_object,
                    '(',
                    self.dataname, 'integer primary key autoincrement ,',
                    SQLITE_VALUE_COLUMN,
                    ')',
                    ))
                cursor.execute(statement)
            else:
                statement = ' '.join((
                    'create table if not exists', self.sq_object,
                    '(',
                    self.dataname, 'integer primary key ,',
                    SQLITE_VALUE_COLUMN,
                    ')',
                    ))
                cursor.execute(statement)
            statement = ' '.join((
                'create table if not exists', self._segmentname,
                '(',
                SQLITE_RECORDS_COLUMN,
                ')',
                ))
            cursor.execute(statement)
        finally:
            cursor.close()
        self._existence_bits.open_root(self.sq_conn)
        # Commit must be at higher level: in open_context() at least.
        #sqconn.commit()

    def close(self):
        """Close inverted index databases then primary database."""
        super(Sqlite3PrimaryFile, self).close()

    def get_control_database(self):
        """Return the database containing segment control data."""
        return self._control_database.get_control_database()

    def get_existence_bits(self):
        """Return the existence bit map control data."""
        return self._existence_bits

    def get_existence_bits_database(self):
        """Return the database containing existence bit map."""
        # Use instance.get_existence_bits().put(...) and so forth in sqlite3
        return self._existence_bits._seg_object

    def set_control_database(self, database):
        """Set reference to segment control databases."""
        self._control_database = database

    def get_control_primary(self):
        """Return the re-use record number control data."""
        return self._control_primary

    def get_segment_records(self, rownumber):
        """Return the record list or bitmap in self._segmentname rownumber."""
        statement = ' '.join((
            'select',
            SQLITE_RECORDS_COLUMN,
            'from',
            self._segmentname,
            'where rowid == ?',
            ))
        values = (rownumber,)
        cursor = self.sq_conn.cursor()
        try:
            return cursor.execute(statement, values).fetchone()[0]
        finally:
            cursor.close()

    def set_segment_records(self, values):
        """Update self._segmentname row using values."""
        statement = ' '.join((
            'update',
            self._segmentname,
            'set',
            SQLITE_RECORDS_COLUMN, '= ?',
            'where rowid == ?',
            ))
        cursor = self.sq_conn.cursor()
        try:
            cursor.execute(statement, values)
        finally:
            cursor.close()

    def delete_segment_records(self, values):
        """Delete self._segmentname row using values."""
        statement = ' '.join((
            'delete from',
            self._segmentname,
            'where rowid == ?',
            ))
        cursor = self.sq_conn.cursor()
        try:
            cursor.execute(statement, values)
        finally:
            cursor.close()

    def insert_segment_records(self, values):
        """Insert self._segmentname row using values."""
        statement = ' '.join((
            'insert into',
            self._segmentname,
            '(',
            SQLITE_RECORDS_COLUMN,
            ')',
            'values ( ? )',
            ))
        #self.sq_conn.execute(statement, values)
        #return self.sq_conn.lastrowid
        #return self.sq_conn.execute(statement, values).lastrowid
        cursor = self.sq_conn.cursor()
        try:
            cursor.execute(statement, values)
            return cursor.execute(
                ' '.join((
                    'select last_insert_rowid() from',
                    self._segmentname))).fetchone()[0]
        finally:
            cursor.close()

            
class Sqlite3SecondaryFile(Sqlite3File):
    
    """Define a sqlite3 table with open_root and close methods.
    """

    def __init__(self, *args):
        """Bitmapped secondary database.  See superclass for *args."""
        super(Sqlite3SecondaryFile, self).__init__(*args)
        self._primary_database = None
        self._indexname = [''.join((INDEXPREFIX, self.dataname))]

    @property
    def indexname(self):
        """Return index name."""
        return self._indexname[0]

    @property
    def indexname_list(self):
        """Return index name list."""
        return self._indexname

    def open_root(self, *args):
        """Open sqlite3 database cursor and create table unless it exists."""
        super(Sqlite3SecondaryFile, self).open_root(*args)
        cursor = self.sq_conn.cursor()
        try:
            statement = ' '.join((
                'create table if not exists', self.sq_object,
                '(',
                self.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self.primaryname,
                ')',
                ))
            cursor.execute(statement)
            statement = ' '.join((
                'create unique index if not exists', self.indexname,
                'on', self.sq_object,
                '(',
                self.dataname, ',',
                SQLITE_SEGMENT_COLUMN,
                ')',
                ))
            cursor.execute(statement)
            # Commit must be at higher level: in open_context() at least.
            #self.sq_conn.commit()
        finally:
            cursor.close()

    def set_primary_database(self, database):
        """Set reference to primary database to access segment databases."""
        self._primary_database = database

    def get_primary_database(self):
        """Set reference to primary database to access segment databases."""
        return self._primary_database


class Sqlite3Primary(Sqlite3PrimaryFile):
    
    """Define a sqlite3 table with record access and deferred update methods.
    """

    def __init__(self, name, sqlite3desc, primaryname):
        """Define a sqlite3 table.

        See superclass for argument descriptions.
        
        """
        super(Sqlite3Primary, self).__init__(
            name, sqlite3desc, primaryname)
        self._clientcursors = dict()
        self._recordsets = dict()
    
    def close(self):
        """Close DB and any cursors or recordsets."""
        for c in list(self._clientcursors.keys()):
            c.close()
        self._clientcursors.clear()
        for rs in list(self._recordsets.keys()):
            rs.close()
        self._recordsets.clear()
        super(Sqlite3Primary, self).close()

    def delete(self, key, value):
        """Delete (key, value) from database.

        Actually update to null because the rowid must be preserved to fit
        the segment structure when record numbers are re-used.

        """
        cursor = self.sq_conn.cursor()
        try:
            statement = ' '.join((
                'update',
                self.sq_object,
                'set',
                SQLITE_VALUE_COLUMN, '= null',
                'where',
                self.primaryname, '== ?',
                ))
            values = (key,)
            cursor.execute(statement, values)

            # See comment in Sqlite3api.delete_instance()
            #self.get_control_primary().note_freed_record_number(key)
            
        except:
            pass
        finally:
            cursor.close()

    def get_first_primary_key_for_index_key(self, key):
        """Return the record number on primary table given key on index.

        This method should be used only on indexed columns whose keys each
        reference a single row. The intended use is where a key for a
        column in table has been derived from a row in some other table.

        """
        raise Sqlite3apiError((
            'get_first_primary_key_for_index_key for primary table'))
    
    def get_primary_record(self, key):
        """Return the instance given the record number in key."""
        if key is None:
            return None
        statement = ' '.join((
            'select * from',
            self.sq_object,
            'where',
            self.dataname, '== ?',
            'order by',
            self.dataname,
            'limit 1',
            ))
        values = (key,)
        cursor = self.sq_conn.cursor()
        try:
            return cursor.execute(statement, values).fetchone()
        finally:
            cursor.close()

    def make_cursor(self, dbobject, keyrange):
        """Create a cursor on the dbobject positiioned at start of keyrange."""
        c = CursorSqlite3Primary(dbobject, keyrange)
        if c:
            self._clientcursors[c] = True
        return c

    def put(self, key, value):
        """Put (key, value) on database and return key for new rows."""
        cursor = self.sq_conn.cursor()
        try:
            if not key: #key == 0:  # Change test to "key is None" when sure
                statement = ' '.join((
                    'insert into',
                    self.sq_object,
                    '(', SQLITE_VALUE_COLUMN, ')',
                    'values ( ? )',
                    ))
                values = (value,)
                #return self.sq_conn.execute(statement, values).lastrowid
                cursor.execute(statement, values)
                return cursor.execute(
                    ' '.join((
                        'select last_insert_rowid() from',
                        self.dataname))).fetchone()[0]
            else:
                statement = ' '.join((
                    'update',
                    self.sq_object,
                    'set',
                    SQLITE_VALUE_COLUMN, '= ?',
                    'where',
                    self.primaryname, '== ?',
                    ))
                values = (value, key)
                cursor.execute(statement, values)
                return None
        finally:
            cursor.close()

    def replace(self, key, oldvalue, newvalue):
        """Replace (key, oldvalue) with (key, newvalue) on table.
        
        (key, newvalue) is put on table only if (key, oldvalue) is on table.
        
        """
        cursor = self.sq_conn.cursor()
        try:
            statement = ' '.join((
                'update',
                self.sq_object,
                'set',
                SQLITE_VALUE_COLUMN, '= ?',
                'where',
                self.primaryname, '== ?',
                ))
            values = (newvalue, key)
            cursor.execute(statement, values)
        except:
            pass
        finally:
            cursor.close()

    def populate_recordset_key_like(self, recordset, key):
        """Raise exception - populate_recordset_key_like primary db."""
        raise Sqlite3apiError(
            ''.join(
                ('populate_recordset_key_like not available ',
                 'on primary database')))

    def populate_recordset_key(self, recordset, key=None):
        """Return recordset on database containing records for key."""
        statement = ' '.join((
            'select',
            self.dataname,
            'from',
            self.sq_object,
            'where',
            self.dataname, '== ?',
            'order by',
            self.dataname,
            'limit 1',
            ))
        values = (key,)
        cursor = self.sq_conn.cursor()
        try:
            if len(cursor.execute(statement, values).fetchone()):
                s, rn = divmod(key, SegmentSize.db_segment_size)
                recordset[s] = SegmentList(
                    s, None, records=rn.to_bytes(2, byteorder='big'))
        finally:
            cursor.close()

    def populate_recordset_key_startswith(self, recordset, key):
        """Raise exception - populate_recordset_key_startswith primary db."""
        raise Sqlite3apiError(
            ''.join(
                ('populate_recordset_key_startswith not available ',
                 'on primary database')))

    def populate_recordset_key_range(
        self, recordset, keystart=None, keyend=None):
        """Return recordset on database containing records for key range."""
        if keystart is None and keyend is None:
            self.populate_recordset_all(recordset)
            return
        
        if keystart is None:
            segment_start, recnum_start = 0, 1
        else:
            segment_start, recnum_start = divmod(keystart,
                                                 SegmentSize.db_segment_size)
        if keyend is not None:
            segment_end, record_number_end = divmod(keyend,
                                                    SegmentSize.db_segment_size)

        if keyend is None:
            statement = ' '.join((
                'select',
                SQLITE_VALUE_COLUMN,
                'from',
                self.get_existence_bits()._seg_dbfile,
                'where',
                self.get_existence_bits()._seg_dbfile, '>= ?',
                ))
            values = (segment_start,)
        elif keystart is None:
            statement = ' '.join((
                'select',
                SQLITE_VALUE_COLUMN,
                'from',
                self.get_existence_bits()._seg_dbfile,
                'where',
                self.get_existence_bits()._seg_dbfile, '<= ?',
                ))
            values = (segment_end,)
        else:
            statement = ' '.join((
                'select',
                SQLITE_VALUE_COLUMN,
                'from',
                self.get_existence_bits()._seg_dbfile,
                'where',
                self.get_existence_bits()._seg_dbfile, '>= ? and',
                self.get_existence_bits()._seg_dbfile, '<= ?',
                ))
            values = (segment_start, segment_end)
        cursor = self.get_existence_bits_database().cursor()
        try:
            for r in cursor.execute(statement, values):
                recordset[r[0] - 1] = SegmentBitarray(
                    r[0] - 1, None, records=r[1])
        finally:
            cursor.close()
        try:
            recordset[segment_start][:recnum_start] = False
        except KeyError:
            pass
        try:
            recordset[segment_end][recnum_end + 1:] = False
        except KeyError:
            pass
    
    def populate_recordset_all(self, recordset):
        """Return recordset containing all referenced records."""
        statement = ' '.join((
            'select',
            self.get_existence_bits()._seg_dbfile, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self.get_existence_bits()._seg_dbfile,
            ))
        values = ()
        cursor = self.get_existence_bits_database().cursor()
        try:
            for r in cursor.execute(statement, values):
                recordset[r[0] - 1] = SegmentBitarray(
                    r[0] - 1, None, records=r[1])
        finally:
            cursor.close()
    
    def file_records_under(self, recordset, key):
        """Raise exception for Sqlite3Primary.file_records_under()."""
        raise DatabaseError(
            'file_records_under not implemented for Sqlite3Primary')
    
    def unfile_records_under(self, key):
        """Raise exception for Sqlite3Primary.unfile_records_under()."""
        raise DatabaseError(
            'unfile_records_under not implemented for Sqlite3Primary')

    def segment_delete(self, segment, record_number):
        """Remove record_number from existence bit map for segment."""
        # See dbduapi.py DBduPrimary.defer_put for model.  Main difference
        # is the write back to database is done immediately (and delete!!).
        # Get the segment existence bit map from database
        ebmb = self.get_existence_bits().get(segment + 1)
        if ebmb is None:
            # It does not exist so raise an exception
            raise Sqlite3apiError(
                'Existence bit map for segment does not exist')
        else:
            # It does exist so convert database representation to bitarray
            ebm = Bitarray()
            ebm.frombytes(ebmb)
            # Set bit for record number and write segment back to database
            ebm[record_number] = False
            self.get_existence_bits().put(segment + 1, ebm.tobytes())

    def segment_put(self, segment, record_number):
        """Add record_number to existence bit map for segment."""
        # See dbduapi.py DBduPrimary.defer_put for model.  Main difference
        # is the write back to database is done immediately.
        # Get the segment existence bit map from database
        ebmb = self.get_existence_bits().get(segment + 1)
        if ebmb is None:
            # It does not exist so create a new empty one
            ebm = SegmentSize.empty_bitarray.copy()
            # Set bit for record number and write segment to database
            ebm[record_number] = True
            self.get_existence_bits().append(ebm.tobytes())
        else:
            # It does exist so convert database representation to bitarray
            ebm = Bitarray()
            ebm.frombytes(ebmb)
            # Set bit for record number and write segment back to database
            ebm[record_number] = True
            self.get_existence_bits().put(segment + 1, ebm.tobytes())

    def find_values(self, valuespec):
        """Yield values meeting valuespec specification."""
        raise DatabaseError(
            'Sqlite3Primary.find_values not implemented')

    def get_high_record(self):
        """Return record with highest record number."""
        statement = ' '.join((
            'select',
            self.dataname, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self.sq_object,
            'order by',
            self.dataname, 'desc',
            'limit 1',
            ))
        values = ()
        cursor = self.sq_conn.cursor()
        try:
            return cursor.execute(statement, values).fetchone()
        finally:
            cursor.close()


class Sqlite3Secondary(Sqlite3SecondaryFile):
    
    """Define a sqlite3 table with record access and deferred update methods.
    """

    def __init__(self, name, sqlite3desc, primaryname):
        """Define a sqlite3 table.

        See superclass for argument descriptions.
        
        """
        super(Sqlite3Secondary, self).__init__(
            name, sqlite3desc, primaryname)
        self._clientcursors = dict()
        self._recordsets = dict()
    
    def close(self):
        """Close DB and any cursors or recordsets."""
        for c in list(self._clientcursors.keys()):
            c.close()
        self._clientcursors.clear()
        for rs in list(self._recordsets.keys()):
            rs.close()
        self._recordsets.clear()
        super(Sqlite3Secondary, self).close()

    def make_cursor(self, dbobject, keyrange):
        """Create a cursor on the dbobject positiioned at start of keyrange."""
        c = CursorSqlite3Secondary(dbobject, keyrange)
        if c:
            self._clientcursors[c] = True
        return c

    def populate_recordset_key_like(self, recordset, key=None):
        """Return recordset containing database records with keys like key."""
        statement = ' '.join((
            'select',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            ))
        db_segment_size_bytes = SegmentSize.db_segment_size_bytes
        matcher = re.compile('.*?' + key, flags=re.IGNORECASE|re.DOTALL)
        get_segment_records = self.get_primary_database().get_segment_records
        cursor = self.sq_conn.cursor()
        try:
            for record in cursor.execute(statement):
                if not matcher.match(record[0]):
                    continue
                column = record[1]
                if record[2] == 1:
                    if column in recordset:
                        recordset[column] |= SegmentInt(
                            column,
                            None,
                            records=record[3].to_bytes(2, byteorder='big'))
                    else:
                        recordset[column] = SegmentInt(
                            column,
                            None,
                            records=record[3].to_bytes(2, byteorder='big'))
                else:
                    bs = get_segment_records(record[3])
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if len(bs) == db_segment_size_bytes:
                        if column in recordset:
                            recordset[column] |= SegmentBitarray(
                            column, None, records=bs)
                        else:
                            recordset[column] = SegmentBitarray(
                            column, None, records=bs)
                    else:
                        if column in recordset:
                            recordset[column] |= SegmentList(
                                column, None, records=bs)
                        else:
                            recordset[column] = SegmentList(
                                column, None, records=bs)
        finally:
            cursor.close()

    def populate_recordset_key(self, recordset, key=None):
        """Return recordset on database containing records for key."""
        statement = ' '.join((
            'select',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            'where',
            self.dataname, '== ?',
            ))
        db_segment_size_bytes = SegmentSize.db_segment_size_bytes
        values = (key,)
        get_segment_records = self.get_primary_database().get_segment_records
        cursor = self.sq_conn.cursor()
        try:
            for record in cursor.execute(statement, values):
                if record[2] == 1:
                    recordset[record[1]] = SegmentInt(
                        record[1],
                        None,
                        records=record[3].to_bytes(2, byteorder='big'))
                else:
                    bs = get_segment_records(record[3])
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if len(bs) == db_segment_size_bytes:
                        recordset[record[1]] = SegmentBitarray(
                            record[1], None, records=bs)
                    else:
                        recordset[record[1]] = SegmentList(
                            record[1], None, records=bs)
        finally:
            cursor.close()

    def populate_recordset_key_startswith(self, recordset, key):
        """Return recordset on database containing records for keys starting."""
        statement = ' '.join((
            'select',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            'where',
            self.dataname, 'glob ?',
            ))
        values = (
            b''.join(
                (key.encode() if isinstance(key, str) else key,
                 b'*',
                 )),)
        db_segment_size_bytes = SegmentSize.db_segment_size_bytes
        get_segment_records = self.get_primary_database().get_segment_records
        cursor = self.sq_conn.cursor()
        try:
            for record in cursor.execute(statement, values):
                if record[2] == 1:
                    segment = SegmentInt(
                        record[1],
                        None,
                        records=record[3].to_bytes(2, byteorder='big'))
                else:
                    bs = get_segment_records(record[3])
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if len(bs) == db_segment_size_bytes:
                        segment = SegmentBitarray(record[1], None, records=bs)
                    else:
                        segment = SegmentList(record[1], None, records=bs)
                if record[1] not in recordset:
                    recordset[record[1]] = segment.promote()
                else:
                    recordset[record[1]] |= segment
        finally:
            cursor.close()

    def populate_recordset_key_range(
        self, recordset, keystart=None, keyend=None):
        """Return recordset on database containing records for key range."""
        if keystart is None and keyend is None:
            self.populate_recordset_all(recordset)
            return

        if keyend is None:
            statement = ' '.join((
                'select',
                self.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self.primaryname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '>= ?',
                ))
            values = (keystart,)
        elif keystart is None:
            statement = ' '.join((
                'select',
                self.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self.primaryname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '<= ?',
                ))
            values = (keyend,)
        else:
            statement = ' '.join((
                'select',
                self.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self.primaryname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '>= ? and',
                self.dataname, '<= ?',
                ))
            values = (keystart, keyend)
        db_segment_size_bytes = SegmentSize.db_segment_size_bytes
        get_segment_records = self.get_primary_database().get_segment_records
        cursor = self.sq_conn.cursor()
        try:
            for record in cursor.execute(statement, values):
                if record[2] == 1:
                    segment = SegmentInt(
                        record[1],
                        None,
                        records=record[3].to_bytes(2, byteorder='big'))
                else:
                    bs = get_segment_records(record[3])
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if len(bs) == db_segment_size_bytes:
                        segment = SegmentBitarray(record[1], None, records=bs)
                    else:
                        segment = SegmentList(record[1], None, records=bs)
                if record[1] not in recordset:
                    recordset[record[1]] = segment.promote()
                else:
                    recordset[record[1]] |= segment
        finally:
            cursor.close()
    
    def populate_recordset_all(self, recordset):
        """Return recordset containing all referenced records."""
        statement = ' '.join((
            'select',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            ))
        values = ()
        db_segment_size_bytes = SegmentSize.db_segment_size_bytes
        get_segment_records = self.get_primary_database().get_segment_records
        cursor = self.sq_conn.cursor()
        try:
            for record in cursor.execute(statement, values):
                if record[2] == 1:
                    segment = SegmentInt(
                        record[1],
                        None,
                        records=record[3].to_bytes(2, byteorder='big'))
                else:
                    bs = get_segment_records(record[3])
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if len(bs) == db_segment_size_bytes:
                        segment = SegmentBitarray(record[1], None, records=bs)
                    else:
                        segment = SegmentList(record[1], None, records=bs)
                if record[1] not in recordset:
                    recordset[record[1]] = segment.promote()
                else:
                    recordset[record[1]] |= segment
        finally:
            cursor.close()

    def populate_segment(self, segment):
        """Helper for segment_delete and segment_put methods."""
        if segment[2] == 1:
            return SegmentInt(
                segment[1],
                None,
                records=segment[3].to_bytes(2, byteorder='big'))
        else:
            bs = self.get_primary_database().get_segment_records(segment[3])
            if bs is None:
                raise DatabaseError('Segment record missing')
            if len(bs) == SegmentSize.db_segment_size_bytes:
                return SegmentBitarray(segment[1], None, records=bs)
            else:
                return SegmentList(segment[1], None, records=bs)

    def make_segment(self, key, segment_number, record_count, records):
        """Return a Segment subclass instance created from arguments."""
        if record_count == 1:
            return SegmentInt(
                segment_number,
                None,
                records=records.to_bytes(2, byteorder='big'))
        else:
            if len(records) == SegmentSize.db_segment_size_bytes:
                return SegmentBitarray(segment_number, None, records=records)
            else:
                return SegmentList(segment_number, None, records=records)
    
    def file_records_under(self, recordset, key):
        """File records in recordset under key.

        The existing references are replaced, not added to as in segment_put
        nor removed from as in segment_delete.

        """
        gpd = self.get_primary_database()

        # select (segment number, record count, key reference) statement for
        # (index value, key) which finds existing segments for index value.
        select_existing_segments = ' '.join((
            'select',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            'indexed by',
            self.indexname,
            'where',
            self.dataname, '== ?',
            ))

        # Update (record count, key reference) statement
        # for (index value, segment number) used when exactly one record is
        # replaced by more than one record or more than one record is replaced
        # by exactly one record.  The former is associated with creation of a
        # segment reference record and the latter with a deletion.
        update_count_and_reference = ' '.join((
            'update',
            self.sq_object,
            'set',
            SQLITE_COUNT_COLUMN, '= ? ,',
            self.primaryname, '= ?',
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # insert or replace (index value, segment number, record count,
        # key reference) statement used when a new segment is created or an
        # existing one replaced.  Key reference is the segment reference if
        # more than one record is in the segment for the index value, or the
        # record key when one record is referenced.
        insert_new_segment = ' '.join((
            'insert or replace into',
            self.sq_object,
            '(',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            ')',
            'values ( ? , ? , ? , ? )',
            ))

        # delete (index value, segment number) statement used when a segment
        # which references one or more records is replaced by an empty segment.
        delete_existing_segment = ' '.join((
            'delete from',
            self.sq_object,
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # Find the segments and sort by record counts.
        # The segment references are reused, not necessarily for the same
        # segment, or deleted.
        # Originally DB_CONVERSION_LIMIT marked the boundary, in record count,
        # between lists and bitarrays and was used to stay close to the bsddb
        # version.  But sorting by record count is just as good here.
        cursor = self.sq_conn.cursor()
        try:
            rows = cursor.execute(select_existing_segments, (key,)).fetchall()
        finally:
            cursor.close()
        old_rows = [s[1] for s in sorted((r[1], r) for r in rows if r[1] > 1)]
        rows = {r[0] for r in rows}
        recordset.normalize()

        # Process the segments in segment number order.
        cursor = self.sq_conn.cursor()
        try:
            for sn in recordset.sorted_segnums:
                rs_segment = recordset.rs_segments[sn]

                if isinstance(rs_segment, SegmentBitarray):

                    # Use an existing bitarray in preference to a list.
                    if len(old_rows):
                        sk = old_rows.pop()[2]
                        gpd.set_segment_records(
                            (rs_segment._bitarray.tobytes(), sk))
                        rows.remove(sn)
                        cursor.execute(
                            update_count_and_reference,
                            (rs_segment.count_records(),
                             sk,
                             key,
                             sn))
                    else:
                        sk = gpd.insert_segment_records(
                            (rs_segment._bitarray.tobytes(),))
                        cursor.execute(
                            insert_new_segment,
                            (key,
                             sn,
                             rs_segment.count_records(),
                             sk))
                        
                elif isinstance(rs_segment, SegmentList):
                    rnlist = b''.join(
                        [rn.to_bytes(2, byteorder='big')
                         for rn in rs_segment._list])

                    # Use an existing list in preference to a bitarray.
                    if len(old_rows):
                        sk = old_rows.pop(0)[2]
                        gpd.set_segment_records((rnlist, sk))
                        rows.remove(sn)
                        cursor.execute(
                            update_count_and_reference,
                            (rs_segment.count_records(),
                             sk,
                             key,
                             sn))
                    else:
                        sk = gpd.insert_segment_records((rnlist,))
                        cursor.execute(
                            insert_new_segment,
                            (key,
                             sn,
                             rs_segment.count_records(),
                             sk))
                        
                elif isinstance(rs_segment, SegmentInt):

                    # Use direct reference to record number.
                    if sn in rows:
                        rows.remove(sn)
                        cursor.execute(
                            update_count_and_reference,
                            (rs_segment.count_records(),
                             int.from_bytes(
                                 rs_segment.tobytes(), 'big'),
                             key,
                             sn))
                    else:
                        cursor.execute(
                            insert_new_segment,
                            (key,
                             sn,
                             rs_segment.count_records(),
                             int.from_bytes(
                                 rs_segment.tobytes(), 'big')
                             ))
        finally:
            cursor.close()
                    
        # Delete any references not reused by file_records_under.
        cursor = self.sq_conn.cursor()
        try:
            for reuse in rows:
                cursor.execute(delete_existing_segment, (key, reuse))
        finally:
            cursor.close()
        for sk in old_rows:
            gpd.delete_segment_records((sk[2],))
    
    def unfile_records_under(self, key):
        """Delete the reference to records in file under key.

        The existing reference by key, usually created by file_records_under,
        is deleted.

        """
        gpd = self.get_primary_database()

        # select (segment number, record count, key reference) statement for
        # (index value, key) which finds existing segments for index value.
        select_existing_segments = ' '.join((
            'select',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            'indexed by',
            self.indexname,
            'where',
            self.dataname, '== ?',
            ))

        # delete (index value, segment number) statement used when a segment
        # which references one or more records is replaced by an empty segment.
        delete_existing_segment = ' '.join((
            'delete from',
            self.sq_object,
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # Find the segments.
        # The segment references are deleted.
        # Originally DB_CONVERSION_LIMIT marked the boundary, in record count,
        # between lists and bitarrays and was used to stay close to the bsddb
        # version.
        cursor = self.sq_conn.cursor()
        try:
            rows = cursor.execute(select_existing_segments, (key,)).fetchall()
        finally:
            cursor.close()
        old_rows = [s[1] for s in sorted((r[1], r) for r in rows if r[1] > 1)]
        rows = {r[0] for r in rows}
                    
        # Delete all references.
        cursor = self.sq_conn.cursor()
        try:
            for reuse in rows:
                cursor.execute(delete_existing_segment, (key, reuse))
        finally:
            cursor.close()
        for sk in old_rows:
            gpd.delete_segment_records((sk[2],))

    def get_first_primary_key_for_index_key(self, key):
        """Return the record number on primary table given key on index.

        This method should be used only on indexed columns whose keys each
        reference a single row. The intended use is where a key for a
        column in table has been derived from a row in some other table.

        """
        statement = ' '.join((
            'select',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            'where',
            self.dataname, '== ?',
            'order by',
            self.primaryname,
            'limit 1',
            ))
        values = (key,)
        cursor = self.sq_conn.cursor()
        try:
            s, c, n = cursor.execute(statement, values).fetchone()
        except TypeError:
            return None
        finally:
            cursor.close()
        if c != 1:
            raise Sqlite3apiError('Index must refer to unique record')
        return s * SegmentSize.db_segment_size + n
    
    def segment_delete(self, key, segment, record_number):
        """Remove record_number from segment for key and write to database.

        """
        gpd = self.get_primary_database()

        # select (index value, segment number, record count, key reference)
        # statement for (index value, segment number).  Execution returns None
        # if no record to delete.
        select_existing_segment = ' '.join((
            'select',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # Update (record count) statement for (index value, segment number) used
        # when at least one record remains after deletion.
        update_record_count = ' '.join((
            'update',
            self.sq_object,
            'set',
            SQLITE_COUNT_COLUMN, '= ?',
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # Update (record count, key reference) statement
        # for (index value, segment number) used when exactly one record
        # remains after deletion.
        update_count_and_reference = ' '.join((
            'update',
            self.sq_object,
            'set',
            SQLITE_COUNT_COLUMN, '= ? ,',
            self.primaryname, '= ?',
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # delete (index value, segment number) statement used when only record
        # in segment is deleted.
        delete_existing_segment = ' '.join((
            'delete from',
            self.sq_object,
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # Get the existing segment if there is one.
        cursor = self.sq_conn.cursor()
        try:
            s = cursor.execute(select_existing_segment, (key, segment)
                               ).fetchone()
        finally:
            cursor.close()

        if s is None:

            # Assume record already deleted.
            return
        
        # Remove record_number from existing segment.
        seg = SegmentInt(
            segment,
            None,
            records=record_number.to_bytes(2, byteorder='big')
            )
        existing_segment = self.populate_segment(s)
        seg = (seg & existing_segment) ^ existing_segment

        count = seg.count_records()
        if count == existing_segment.count_records():

            # Assume record already deleted.
            return

        if not isinstance(existing_segment, SegmentBitarray):
            seg = seg.normalize()
        else:
            seg = seg.normalize(use_upper_limit=False)
        if count > 1:

            # Update segment and record count.
            gpd.set_segment_records((seg.tobytes(), s[3]))
            cursor = self.sq_conn.cursor()
            try:
                cursor.execute(update_record_count, (count, key, segment))
            finally:
                cursor.close()
            return

        if count == 1:

            # Assume count was >1 before delete and delete segment leaving a
            # direct reference to the remaining record number.
            gpd.delete_segment_records((s[3],))
            rn = seg.get_record_number_at_position(0)
            cursor = self.sq_conn.cursor()
            try:
                cursor.execute(
                    update_count_and_reference,
                    (count,
                     rn % (segment * SegmentSize.db_segment_size
                           ) if segment else rn,
                     key,
                     segment))
            finally:
                cursor.close()
            return

        # Assume count was 1 before delete and delete segment.
        cursor = self.sq_conn.cursor()
        try:
            cursor.execute(delete_existing_segment, (key, segment))
        finally:
            cursor.close()
        return
    
    def segment_put(self, key, segment, record_number):
        """Add record_number to segment for key and write to database.

        """
        gpd = self.get_primary_database()

        # All puts of new records except the first in a segment are splicing
        # into an existing segment, so follow the example of sort_and_write()
        # in sqlite3duapi.py.

        # select (index value, segment number, record count, key reference)
        # statement for (index value, segment number).  Execution returns None
        # if no splicing needed.
        select_existing_segment = ' '.join((
            'select',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            'from',
            self.sq_object,
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # Update (record count) statement for (index value, segment number) used
        # when splicing needed.
        update_record_count = ' '.join((
            'update',
            self.sq_object,
            'set',
            SQLITE_COUNT_COLUMN, '= ?',
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # Update (record count, key reference) statement
        # for (index value, segment number) used when record count increased
        # from 1.
        update_count_and_reference = ' '.join((
            'update',
            self.sq_object,
            'set',
            SQLITE_COUNT_COLUMN, '= ? ,',
            self.primaryname, '= ?',
            'where',
            self.dataname, '== ? and',
            SQLITE_SEGMENT_COLUMN, '== ?',
            ))

        # insert (index value, segment number, record count, key reference)
        # statement.
        insert_new_segment = ' '.join((
            'insert into',
            self.sq_object,
            '(',
            self.dataname, ',',
            SQLITE_SEGMENT_COLUMN, ',',
            SQLITE_COUNT_COLUMN, ',',
            self.primaryname,
            ')',
            'values ( ? , ? , ? , ? )',
            ))

        # Get the existing segment if there is one.
        cursor = self.sq_conn.cursor()
        try:
            s = cursor.execute(select_existing_segment, (key, segment)
                               ).fetchone()
        finally:
            cursor.close()

        if s is None:

            # Create a new segment referring to one record.
            cursor = self.sq_conn.cursor()
            try:
                cursor.execute(insert_new_segment,
                               (key, segment, 1, record_number))
            finally:
                cursor.close()
            return

        # Splice record_number into existing segment.
        # If the existing segment record for a segment had a record count > 1
        # before being updated, a subsidiary table record already exists.
        # Otherwise it must be created.
        # Key reference is a record number if the record count is 1.
        existing_segment = self.populate_segment(s)
        seg = SegmentInt(
            segment,
            None,
            records=record_number.to_bytes(2, byteorder='big')
            ) | existing_segment
        if not isinstance(existing_segment, SegmentBitarray):
            seg = seg.normalize()
        if s[2] > 1:
            gpd.set_segment_records((seg.tobytes(), s[3]))
            cursor = self.sq_conn.cursor()
            try:
                cursor.execute(update_record_count, (s[2] + 1, key, segment))
            finally:
                cursor.close()
        else:
            nv = gpd.insert_segment_records((seg.tobytes(),))
            cursor = self.sq_conn.cursor()
            try:
                cursor.execute(update_count_and_reference,
                               (s[2] + 1, nv, key, s[1]))
            finally:
                cursor.close()

    def find_values(self, valuespec):
        """Yield values meeting valuespec specification."""
        if valuespec.above_value and valuespec.below_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '> ? and',
                self.dataname, '< ?',
                ))
            values = valuespec.above_value, valuespec.below_value
        elif valuespec.above_value and valuespec.to_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '> ? and',
                self.dataname, '<= ?',
                ))
            values = valuespec.above_value, valuespec.to_value
        elif valuespec.from_value and valuespec.to_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '>= ? and',
                self.dataname, '<= ?',
                ))
            values = valuespec.from_value, valuespec.to_value
        elif valuespec.from_value and valuespec.below_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '>= ? and',
                self.dataname, '< ?',
                ))
            values = valuespec.from_value, valuespec.below_value
        elif valuespec.above_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '> ?',
                ))
            values = valuespec.above_value,
        elif valuespec.from_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '>= ?',
                ))
            values = valuespec.from_value,
        elif valuespec.to_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '<= ?',
                ))
            values = valuespec.to_value,
        elif valuespec.below_value:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                'where',
                self.dataname, '< ?',
                ))
            values = valuespec.below_value,
        else:
            statement = ' '.join((
                'select distinct',
                self.dataname,
                'from',
                self.sq_object,
                ))
            values = ()
        apply_to_value = valuespec.apply_pattern_and_set_filters_to_value
        cursor = self.sq_conn.cursor()
        try:
            for r in cursor.execute(statement, values):
                if apply_to_value(r[0]):
                    yield r[0]
        finally:
            cursor.close()


class _Sqlite3Definition(object):
    """A definition derived from a FileSpec entry fitted to sqlite3."""

    def __init__(self,
                 dbset=None,
                 specification=None,
                 primary_class=None,
                 secondary_class=None,
                 field_name_converter=None,
                 **kw):
        """Create primary and secondary instances from specification for dbset.

        This class should be instantiated via the make_root() call which is
        called after much of the specification validation, which cannot be
        done internally, has been done.

        """
        super().__init__()
        for a in (dbset,
                  specification,
                  primary_class,
                  secondary_class,
                  field_name_converter):
            if a is None:
                msg = ' '.join(['Cannot construct definition from',
                                'specification',
                                'because a mandatory argument is missing'])
                raise Sqlite3apiError(msg)
        self._dbset = dbset
        primary = specification[PRIMARY]
        self.primary = primary_class(
            primary,
            specification,
            primary)
        self.secondary = {}
        self.dbname_to_secondary_key = {}
        if SECONDARY in specification:
            for sname, secondary in specification[SECONDARY].items():
                if not isinstance(sname, str):
                    msg = ' '.join(['Secondary table name',
                                    'for', dbset,
                                    'must be a string'])
                    raise Sqlite3apiError(msg)

                if secondary is None:
                    secondary = field_name_converter(sname)
                
                if secondary == primary:
                    msg = ' '.join(['Secondary table name', secondary,
                                    'for', dbset,
                                    'cannot be same as primary'])
                    raise Sqlite3apiError(msg)

                if secondary not in specification[FIELDS]:
                    msg = ' '.join(['Secondary table name',
                                    secondary,
                                    'for', dbset, 'does not have',
                                    'a column description'])
                    raise Sqlite3apiError(msg)

                self.secondary[secondary] = secondary_class(
                    secondary,
                    specification,
                    primary)
                self.dbname_to_secondary_key[sname] = secondary
        for v in self.secondary.values():
            v.set_primary_database(self.primary)

    def associate(self, dbname):
        """Return primary definition for dbname."""
        if dbname == self._dbset:
            return self.primary
        return self.secondary[self.dbname_to_secondary_key[dbname]]


class CursorSqlite3(Cursor):
    
    """Define bsddb3 style cursor methods on a sqlite table.

    Primary and secondary database, and others, should be read as the Berkeley
    DB usage.  This class emulates interaction with a Berkeley DB database via
    the Python bsddb3 module.

    Segmented should be read as the DPT database engine usage.

    The value part of (key, value) on primary or secondary databases is either:

        primary key (segment and record number)
        reference to a list of primary keys for a segment
        reference to a bit map of primary keys for a segment

    References are to records on RECNO databases, one each for lists and bit
    maps, containing the primary keys.

    Each rowid is mapped to a bit in the bitmap associated with a segment.

    """

    def __init__(self, dbset, keyrange=None):
        """Define a cursor using the sqlite3 engine."""
        super(CursorSqlite3, self).__init__(dbset)
        self._most_recent_row_read = False
        if dbset.sq_object_list is not None:
            self._cursor = dbset.sq_conn.cursor()
        self._current_segment = None
        self._current_segment_number = None
        self._current_record_number_in_segment = None

    def close(self):
        """Delete database cursor."""
        try:
            del self._dbset._clientcursors[self]
        except:
            pass
        if self._cursor:
            self._cursor.close()
        self._cursor = None
        self._dbset = None
        self.set_partial_key(None)
        self._most_recent_row_read = False

    def database_cursor_exists(self):
        """Return True if database cursor exists and False otherwise."""
        return bool(self._cursor)

    def get_partial(self):
        """return self._partial"""
        return self._partial

    def get_converted_partial(self):
        """return self._partial as it would be held on database."""
        # See comment at get_converted_partial_with_wildcard().
        return self._partial#.encode()

    def get_partial_with_wildcard(self):
        """return self._partial with wildcard suffix appended."""
        raise DatabaseError('get_partial_with_wildcard not implemented')

    def get_converted_partial_with_wildcard(self):
        """return converted self._partial with wildcard suffix appended."""
        # Code replaced:
        #return self._dbset.get_converter(USE_BYTES)(
        #    ''.join((self._partial, '*')))
        # implies self._partial is str always: so calling the converter, which
        # tests if argument is str. is pointless.
        # Sqlite will encode str internally so the encode() is commented.
        return ''.join((self._partial, '*'))#.encode()

    def refresh_recordset(self, instance=None):
        """Refresh records for datagrid access after database update.

        Do nothing in sqlite3.  The cursor (for the datagrid) accesses
        database directly.  There are no intervening data structures which
        could be inconsistent.

        """
        pass


class CursorSqlite3Primary(CursorSqlite3):
    
    """Define bsddb3 cursor methods for primary database on a segmented table.
    """

    def count_records(self):
        """return record count or None if cursor is not usable."""

        # Quicker than executing 'select count ( * ) ...' for many records.
        statement = ' '.join((
            'select',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.get_existence_bits()._seg_dbfile,
            ))
        return sum(SegmentBitarray(0, None, records=r[0]).count_records()
                   for r in self._cursor.execute(statement))

    def first(self):
        """Return first record taking partial key into account."""
        statement = ' '.join((
            'select',
            self._dbset.dataname, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.dataname,
            'order by',
            self._dbset.dataname,
            'limit 1',
            ))
        values = ()
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        return self._most_recent_row_read

    def get_position_of_record(self, record=None):
        """return position of record in file or 0 (zero)."""
        if record is None:
            return 0

        # Quicker than executing 'select count ( * ) ...' for many records.
        statement = ' '.join((
            'select',
            'rowid', ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.get_existence_bits()._seg_dbfile,
            'order by rowid',
            ))
        dss = SegmentSize.db_segment_size
        position = 0
        rowid = record[0]
        for r in self._cursor.execute(statement):
            segment = SegmentBitarray(r[0]-1, None, records=r[1])
            if r[0] * dss <= rowid:
                position += segment.count_records()
                continue
            position += segment.get_position_of_record_number(
                rowid % dss)
            break
        return position

    def get_record_at_position(self, position=None):
        """return record for positionth record in file or None."""
        if position is None:
            return None
        if position < 0:
            statement = ' '.join((
                'select * from',
                self._dbset.dataname,
                'order by',
                self._dbset.dataname, 'desc',
                'limit 1',
                'offset ?',
                ))
            values = (str(-1 - position),)
        else:
            statement = ' '.join((
                'select * from',
                self._dbset.dataname,
                'order by',
                self._dbset.dataname,
                'limit 1',
                'offset ?',
                ))
            values = (str(position - 1),)
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        return self._most_recent_row_read

    def last(self):
        """Return last record taking partial key into account."""
        statement = ' '.join((
            'select',
            self._dbset.dataname, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.dataname,
            'order by',
            self._dbset.dataname, 'desc',
            'limit 1',
            ))
        values = ()
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        return self._most_recent_row_read

    def nearest(self, key):
        """Return nearest record to key taking partial key into account."""
        statement = ' '.join((
            'select',
            self._dbset.dataname, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.dataname,
            'where',
            self._dbset.dataname, '>= ?',
            'order by',
            self._dbset.dataname,
            'limit 1',
            ))
        values = (key,)
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        return self._most_recent_row_read

    def next(self):
        """Return next record taking partial key into account."""
        if self._most_recent_row_read is False:
            return self.first()
        elif self._most_recent_row_read is None:
            return None
        statement = ' '.join((
            'select',
            self._dbset.dataname, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.dataname,
            'where',
            self._dbset.dataname, '> ?',
            'order by',
            self._dbset.dataname,
            'limit 1',
            ))
        values = (self._most_recent_row_read[0],)
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        return self._most_recent_row_read

    def prev(self):
        """Return previous record taking partial key into account."""
        if self._most_recent_row_read is False:
            return self.last()
        elif self._most_recent_row_read is None:
            return None
        statement = ' '.join((
            'select',
            self._dbset.dataname, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.dataname,
            'where',
            self._dbset.dataname, '< ?',
            'order by',
            self._dbset.dataname, 'desc',
            'limit 1',
            ))
        values = (self._most_recent_row_read[0],)
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        return self._most_recent_row_read

    def setat(self, record):
        """Return current record after positioning cursor at record.

        Take partial key into account.
        
        Words used in bsddb3 (Python) to describe set and set_both say
        (key, value) is returned while Berkeley DB description seems to
        say that value is returned by the corresponding C functions.
        Do not know if there is a difference to go with the words but
        bsddb3 works as specified.

        """
        if self.get_partial() is False:
            return None
        if self.get_partial() is not None:
            if not record[0].startswith(self.get_partial()):
                return None
        statement = ' '.join((
            'select',
            self._dbset.dataname, ',',
            SQLITE_VALUE_COLUMN,
            'from',
            self._dbset.dataname,
            'where',
            self._dbset.dataname, '== ?',
            'order by',
            self._dbset.dataname,
            'limit 1',
            ))
        values = (record[0],)
        row = self._cursor.execute(statement, values).fetchone()
        if row:
            self._most_recent_row_read = row
        return row

    def set_partial_key(self, partial):
        """Set partial key to None for primary cursor."""
        #self._partial = partial
        #self._most_recent_row_read = False
        self._partial = None

    def refresh_recordset(self, instance=None):
        """Refresh records for datagrid access after database update.

        The bitmap for the record set may not match the existence bitmap.

        """
        #raise Sqlite3apiError('refresh_recordset not implemented')


class CursorSqlite3Secondary(CursorSqlite3):
    
    """Define bsddb3 cursor methods for secondary database on a segmented table.
    """

    def count_records(self):
        """Return record count."""
        #if self.get_partial() is None:
        if self.get_partial() in (None, False):
            statement = ' '.join((
                'select',
                SQLITE_COUNT_COLUMN, # to avoid sum() overflow
                'from',
                self._dbset.dataname,
                ))
            values = ()
        else:
            statement = ' '.join((
                'select',
                SQLITE_COUNT_COLUMN, # to avoid sum() overflow
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ?',
                ))
            values = (self.get_converted_partial_with_wildcard(),)
        count = 0
        for r in self._cursor.execute(statement, values):
            count += r[0]
        return count

    def first(self):
        """Return first record taking partial key into account."""
        if self.get_partial() is None:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = ()
        elif self.get_partial() is False:
            return None
        else:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = (self.get_converted_partial_with_wildcard(),)
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        if self._most_recent_row_read is None:
            return None
        return self.set_current_segment(self._most_recent_row_read).first()

    def get_position_of_record(self, record=None):
        """Return position of record in file or 0 (zero)."""
        if record is None:
            return 0
        key, value = record
        segment_number, record_number = divmod(value,
                                               SegmentSize.db_segment_size)
        # Get position of record relative to start point
        if not self.get_partial():
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                ))
            values = ()
        else:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                ))
            values = (self.get_converted_partial_with_wildcard(),)
        db_segment_size_bytes = SegmentSize.db_segment_size_bytes
        gpd = self._dbset.get_primary_database()
        position = 0
        for r in self._cursor.execute(statement, values):
            if r[0] < key:
                position += r[2]
            elif r[0] > key:
                break
            elif r[1] < segment_number:
                position += r[2]
            elif r[1] > segment_number:
                break
            else:
                if r[2] == 1:
                    segment = SegmentInt(
                        segment_number,
                        None,
                        records=r[3].to_bytes(2, byteorder='big'))
                else:
                    bs = gpd.get_segment_records(r[3])
                    if len(bs) == db_segment_size_bytes:
                        segment = SegmentBitarray(
                            segment_number, None, records=bs)
                    else:
                        segment = SegmentList(segment_number, None, records=bs)
                position += segment.get_position_of_record_number(record_number)
        return position

    def get_record_at_position(self, position=None):
        """Return record for positionth record in file or None."""
        if position is None:
            return None
        # Start at first or last record whichever is likely closer to position
        if position < 0:
            is_step_forward = False
            position = -1 - position
            if not self.get_partial():
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'order by',
                    self._dbset.dataname, 'desc', ',',
                    SQLITE_SEGMENT_COLUMN, 'desc',
                    ))
                values = ()
            else:
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'where',
                    self._dbset.dataname, 'glob ?',
                    'order by',
                    self._dbset.dataname, 'desc', ',',
                    SQLITE_SEGMENT_COLUMN, 'desc',
                    ))
                values = (self.get_converted_partial_with_wildcard(),)
        else:
            is_step_forward = True
            if not self.get_partial():
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'order by',
                    self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                    ))
                values = ()
            else:
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'where',
                    self._dbset.dataname, 'glob ?',
                    'order by',
                    self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                    ))
                values = (self.get_converted_partial_with_wildcard(),)
        # Get record at position relative to start point
        db_segment_size_bytes = SegmentSize.db_segment_size_bytes
        gpd = self._dbset.get_primary_database()
        count = 0
        for r in self._cursor.execute(statement, values):
            count += r[2]
            if count < position:
                continue
            count -= position
            if r[2] == 1:
                segment = SegmentInt(
                    r[1],
                    None,
                    records=r[3].to_bytes(2, byteorder='big'))
            else:
                bs = gpd.get_segment_records(r[3])
                if len(bs) == db_segment_size_bytes:
                    segment = SegmentBitarray(r[1], None, records=bs)
                else:
                    segment = SegmentList(r[1], None, records=bs)
            record_number = segment.get_record_number_at_position(
                count, is_step_forward)
            if record_number is not None:
                return r[0], record_number
            break
        return None

    def last(self):
        """Return last record taking partial key into account."""
        if self.get_partial() is None:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'order by',
                self._dbset.dataname, 'desc', ',',
                SQLITE_SEGMENT_COLUMN, 'desc',
                'limit 1',
                ))
            values = ()
        elif self.get_partial() is False:
            return None
        else:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ?',
                'order by',
                self._dbset.dataname, 'desc', ',',
                SQLITE_SEGMENT_COLUMN, 'desc',
                'limit 1',
                ))
            values = (self.get_converted_partial_with_wildcard(),)
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        if self._most_recent_row_read is None:
            return None
        return self.set_current_segment(self._most_recent_row_read).last()

    def nearest(self, key):
        """Return nearest record to key taking partial key into account."""
        if self.get_partial() is None:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, '>= ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = (key,)
        elif self.get_partial() is False:
            return None
        else:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ? and',
                self._dbset.dataname, '>= ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = (self.get_converted_partial_with_wildcard(), key)
        self._most_recent_row_read = self._cursor.execute(
            statement, values).fetchone()
        if self._most_recent_row_read is None:
            return None
        return self.set_current_segment(self._most_recent_row_read).first()

    def next(self):
        """Return next record taking partial key into account."""
        if self._most_recent_row_read is False:
            return self.first()
        elif self._most_recent_row_read is None:
            return None
        record = self._current_segment.next()
        if record is not None:
            return record
        if self.get_partial() is None:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, '== ? and',
                SQLITE_SEGMENT_COLUMN, '> ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = (self._current_segment._key, self._current_segment_number)
            self._most_recent_row_read = self._cursor.execute(
                statement, values).fetchone()
            if self._most_recent_row_read is None:
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'where',
                    self._dbset.dataname, '> ?',
                    'order by',
                    self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                    'limit 1',
                    ))
                values = (self._current_segment._key,)
                self._most_recent_row_read = self._cursor.execute(
                    statement, values).fetchone()
        elif self.get_partial() is False:
            return None
        else:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ? and',
                self._dbset.dataname, '== ? and',
                SQLITE_SEGMENT_COLUMN, '> ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = (
                self.get_converted_partial_with_wildcard(),
                self._current_segment._key,
                self._current_segment_number,
                )
            self._most_recent_row_read = self._cursor.execute(
                statement, values).fetchone()
            if self._most_recent_row_read is None:
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'where',
                    self._dbset.dataname, 'glob ? and',
                    self._dbset.dataname, '> ?',
                    'order by',
                    self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                    'limit 1',
                    ))
                values = (
                    self.get_converted_partial_with_wildcard(),
                    self._current_segment._key,
                    )
                self._most_recent_row_read = self._cursor.execute(
                    statement, values).fetchone()
        if self._most_recent_row_read is None:
            return None
        return self.set_current_segment(self._most_recent_row_read).first()

    def prev(self):
        """Return previous record taking partial key into account."""
        if self._most_recent_row_read is False:
            return self.first()
        elif self._most_recent_row_read is None:
            return None
        record = self._current_segment.prev()
        if record is not None:
            return record
        if self.get_partial() is None:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, '== ? and',
                SQLITE_SEGMENT_COLUMN, '< ?',
                'order by',
                self._dbset.dataname, 'desc', ',',
                SQLITE_SEGMENT_COLUMN, 'desc',
                'limit 1',
                ))
            values = (self._current_segment._key, self._current_segment_number)
            self._most_recent_row_read = self._cursor.execute(
                statement, values).fetchone()
            if self._most_recent_row_read is None:
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'where',
                    self._dbset.dataname, '< ?',
                    'order by',
                    self._dbset.dataname, 'desc', ',',
                    SQLITE_SEGMENT_COLUMN, 'desc',
                    'limit 1',
                    ))
                values = (self._current_segment._key,)
                self._most_recent_row_read = self._cursor.execute(
                    statement, values).fetchone()
        elif self.get_partial() is False:
            return None
        else:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ? and',
                self._dbset.dataname, '== ? and',
                SQLITE_SEGMENT_COLUMN, '< ?',
                'order by',
                self._dbset.dataname, 'desc', ',',
                SQLITE_SEGMENT_COLUMN, 'desc',
                'limit 1',
                ))
            values = (
                self.get_converted_partial_with_wildcard(),
                self._current_segment._key,
                self._current_segment_number,
                )
            self._most_recent_row_read = self._cursor.execute(
                statement, values).fetchone()
            if self._most_recent_row_read is None:
                statement = ' '.join((
                    'select',
                    self._dbset.dataname, ',',
                    SQLITE_SEGMENT_COLUMN, ',',
                    SQLITE_COUNT_COLUMN, ',',
                    self._dbset.primaryname,
                    'from',
                    self._dbset.dataname,
                    'where',
                    self._dbset.dataname, 'glob ? and',
                    self._dbset.dataname, '< ?',
                    'order by',
                    self._dbset.dataname, 'desc', ',',
                    SQLITE_SEGMENT_COLUMN, 'desc',
                    'limit 1',
                    ))
                values = (
                    self.get_converted_partial_with_wildcard(),
                    self._current_segment._key,
                    )
                self._most_recent_row_read = self._cursor.execute(
                    statement, values).fetchone()
        if self._most_recent_row_read is None:
            return None
        return self.set_current_segment(self._most_recent_row_read).last()

    def setat(self, record):
        """Return current record after positioning cursor at record.

        Take partial key into account.
        
        Words used in bsddb3 (Python) to describe set and set_both say
        (key, value) is returned while Berkeley DB description seems to
        say that value is returned by the corresponding C functions.
        Do not know if there is a difference to go with the words but
        bsddb3 works as specified.

        """
        if self.get_partial() is False:
            return None
        if self.get_partial() is not None:
            if not record[0].startswith(self.get_partial()):
                return None
        segment_number, record_number = divmod(record[1],
                                               SegmentSize.db_segment_size)
        if self.get_partial() is not None:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, 'glob ? and',
                self._dbset.dataname, '== ? and',
                SQLITE_SEGMENT_COLUMN, '== ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = (
                self.get_converted_partial_with_wildcard(),
                record[0],
                segment_number,
                )
        else:
            statement = ' '.join((
                'select',
                self._dbset.dataname, ',',
                SQLITE_SEGMENT_COLUMN, ',',
                SQLITE_COUNT_COLUMN, ',',
                self._dbset.primaryname,
                'from',
                self._dbset.dataname,
                'where',
                self._dbset.dataname, '== ? and',
                SQLITE_SEGMENT_COLUMN, '== ?',
                'order by',
                self._dbset.dataname, ',', SQLITE_SEGMENT_COLUMN,
                'limit 1',
                ))
            values = (record[0], segment_number)
        row = self._cursor.execute(statement, values).fetchone()
        if row is None:
            return None
        segment = self.get_segment(*row)
        if record_number not in segment:
            return None
        self._current_segment = segment
        self._current_segment_number = row[1]
        self._most_recent_row_read = row
        return segment.setat(record[1])

    def set_partial_key(self, partial):
        """Set partial key."""
        self._partial = partial

    def get_segment(self, key, segment_number, count, record_number):
        """Return a SegmentBitarray, SegmentInt, or SegmentList instance.

        Arguments are the 4-tuple segment reference returned by fetchone().

        """
        if count == 1:
            return SegmentInt(
                segment_number,
                key,
                records=record_number.to_bytes(2, byteorder='big'))
        if self._current_segment_number == segment_number:
            if key == self._current_segment._key:
                return self._current_segment
        records=self._dbset.get_primary_database().get_segment_records(
            record_number)
        if len(records) == SegmentSize.db_segment_size_bytes:
            return SegmentBitarray(segment_number, key, records=records)
        else:
            return SegmentList(segment_number, key, records=records)

    def set_current_segment(self, segment_reference):
        """Return a SegmentBitarray, SegmentInt, or SegmentList instance.

        Argument is the 4-tuple segment reference returned by fetchone().

        """
        self._current_segment = self.get_segment(*segment_reference)
        self._current_segment_number = segment_reference[1]
        return self._current_segment

    def refresh_recordset(self, instance=None):
        """Refresh records for datagrid access after database update.

        The bitmap for the record set may not match the existence bitmap.

        """
        # See set_selection() hack in chesstab subclasses of DataGrid.
        
        #raise Sqlite3apiError('refresh_recordset not implemented')


class RecordsetCursorSqlite3(RecordsetCursor):
    
    """Provide a bsddb3 style cursor for a recordset of arbitrary records.

    The cursor does not support partial keys because the records in the
    recordset do not have an implied order (apart from the accidential order
    of existence on the database).

    """

    # Hack to get round self._dbset._database being a Sqlite.Cursor which means
    # the Recordset.get_record method does not work here because it does:
    # record = self._database.get(record_number)
    # All self._dbset.get_record(..) calls replaced by self._get_record(..) in
    # this module (hope no external use for now).
    # Maybe Recordset should not have a get_record method.
    def _get_record(self, record_number, use_cache=False):
        """Return record_number from database using cache if requested."""
        dbset = self._dbset
        primary = dbset.dbhome.get_database_instance(dbset.dbset, dbset.dbset
                                                     ).dataname
        if use_cache:
            record = dbset.record_cache.get(record_number)
            if record is not None:
                return record # maybe (record_number, record)
        segment, recnum = divmod(record_number, SegmentSize.db_segment_size)
        if segment not in dbset.rs_segments:
            return # maybe raise
        if recnum not in dbset.rs_segments[segment]:
            return # maybe raise
        statement = ' '.join((
            'select',
            SQLITE_VALUE_COLUMN,
            'from',
            primary,
            'where',
            primary, '== ?',
            'limit 1',
            ))
        values = (record_number,)
        database_cursor = dbset._database.cursor()
        try:
            record = database_cursor.execute(statement, values).fetchone()[0]
        finally:
            database_cursor.close()
        # maybe raise if record is None (if not, None should go on cache)
        if use_cache:
            dbset.record_cache[record_number] = record
            dbset.record.deque.append(record_number)
        return (record_number, record)

            
class Sqlite3Segment(object):
    
    """Define a sqlite3 table to store inverted record number lists or bitmaps.
    """

    def __init__(self, dbfile, segment_type):
        """Define segment file for segment type of name in file description.
        
        dbfile=file name relative to environment home directory
        segment_type=representation style for inverted list of record numbers

        For large numbers of records fixed sized bit maps representing record
        numbers relative to the base for the segment.  Otherwise a list of
        record numbers relative to the base.
        
        """
        super(Sqlite3Segment, self).__init__()
        self._seg_dbfile = (SUBFILE_DELIMITER * 2).join((dbfile, segment_type))
        self._seg_object = None

    def close(self):
        """Close inverted index DB."""
        self._seg_object = None

    def open_root(self, sqconn):
        """Create inverted index DB in dbenv."""
        try:
            self._seg_object = sqconn
        except:
            raise
        cursor = self._seg_object.cursor()
        try:
            statement = ' '.join((
                'create table if not exists', self._seg_dbfile,
                '(',
                self._seg_dbfile,
                'integer primary key', ',',
                SQLITE_VALUE_COLUMN,
                ')',
                ))
            cursor.execute(statement)
        except:
            self._seg_object = None
            raise
        finally:
            cursor.close()

    def get(self, key):
        """Get a segment record from the database."""
        statement = ' '.join((
            'select',
            SQLITE_VALUE_COLUMN,
            'from',
            self._seg_dbfile,
            'where',
            self._seg_dbfile, '== ?',
            'limit 1',
            ))
        values = (key,)
        cursor = self._seg_object.cursor()
        try:
            return cursor.execute(statement, values).fetchone()[0]
        except TypeError:
            return None
        finally:
            cursor.close()

    def delete(self, key):
        """Delete a segment record from the database."""
        statement = ' '.join((
            'delete from',
            self._seg_dbfile,
            'where',
            self._seg_dbfile, '== ?',
            ))
        values = (key,)
        cursor = self._seg_object.cursor()
        try:
            cursor.execute(statement, values)
        finally:
            cursor.close()

    def put(self, key, value):
        """Put a segment record on the database using key."""
        statement = ' '.join((
            'update',
            self._seg_dbfile,
            'set',
            SQLITE_VALUE_COLUMN, '= ?',
            'where',
            self._seg_dbfile, '== ?',
            ))
        values = (value, key)
        cursor = self._seg_object.cursor()
        try:
            cursor.execute(statement, values)
        finally:
            cursor.close()

    def append(self, value):
        """Append a segment record on the database using a new key."""
        statement = ' '.join((
            'insert into',
            self._seg_dbfile,
            '(',
            SQLITE_VALUE_COLUMN,
            ')',
            'values ( ? )',
            ))
        values = (value,)
        #return self._seg_object.execute(statement, values).lastrowid
        cursor = self._seg_object.cursor()
        try:
            return cursor.execute(statement, values).execute(
                    ' '.join((
                        'select last_insert_rowid() from',
                        self._seg_dbfile))).fetchone()[0]
        finally:
            cursor.close()

            
class Sqlite3ExistenceBitMap(Sqlite3Segment):
    
    """Sqlite3 table to store record existence bit map keyed by segment number.
    """

    def __init__(self, dbfile):
        """Define dbfile in environment for segment record bit maps.
        
        dbfile=file name relative to environment home directory
        
        """
        super(Sqlite3ExistenceBitMap, self).__init__(dbfile, 'exist')
        self._segment_count = None

    @property
    def segment_count(self):
        """Return number of records in segment."""
        return self._segment_count

    @segment_count.setter
    def segment_count(self, segment_number):
        """Set segment count from 0-based segment_number if greater."""
        if segment_number > self._segment_count:
            self._segment_count = segment_number + 1
    
    def open_root(self, sqconn):
        """Create inverted index DB in dbenv."""
        super(Sqlite3ExistenceBitMap, self).open_root(sqconn)
        statement = ' '.join(('select count(*) from', self._seg_dbfile))
        cursor = self._seg_object.cursor()
        try:
            self._segment_count = cursor.execute(statement).fetchone()[0]
        finally:
            cursor.close()

            
class Sqlite3ControlFile(object):
    
    """Define sqlite3 table for control information about the sqlite3 tables.

    The method names used in Sqlite3apiPrimaryFile and superclasses are used
    where possible.  But this file is primary DB_BTREE NODUP.  It is used by
    all FileControl instances.
    
    """

    def __init__(self, control_file='control'):
        """File control database for all DB files."""
        super(Sqlite3ControlFile, self).__init__()
        self._control_file = ''.join((SUBFILE_DELIMITER * 3, control_file))
        self._control_object = None

    def open_root(self, sqconn):
        """Create file control database in environment."""
        self._control_object = sqconn
        cursor = self._control_object.cursor()
        try:
            statement = ' '.join((
                'create table if not exists', self._control_file,
                '(',
                self._control_file, ',',
                SQLITE_VALUE_COLUMN, ',',
                'primary key',
                '(',
                self._control_file, ',',
                SQLITE_VALUE_COLUMN,
                ') )',
                ))
            cursor.execute(statement)
        except:
            self._control_object = None
            raise
        finally:
            cursor.close()

    def close(self):
        """Close file control database."""
        self._control_object = None

    def get_control_database(self):
        """Return the database containing file control records."""
        return self._control_object

    @property
    def control_file(self):
        """Return the name which is both primary column and table name."""
        return self._control_file


class FileControl(object):
    
    """Freed resource data for a segmented file.
    """

    def __init__(self, dbfile):
        """Define the file control data."""
        super(FileControl, self).__init__()

        # Primary or Secondary file instance whose segment reuse is handled
        self._dbfile = dbfile


class FileControlPrimary(FileControl):
    
    """Freed resource data for a segmented file.

    Introduced to keep track of pages on the existence bit map that contain
    freed record numbers that can be reused.

    The list of existence bit map page numbers containing freed record numbers
    is cached.

    """

    def __init__(self, *args):
        """Define the file control data for primary files."""
        super(FileControlPrimary, self).__init__(*args)
        self._freed_record_number_pages = None

    @property
    def freed_record_number_pages(self):
        """Return existence bit map record numbers available for re-use."""
        if self._freed_record_number_pages is None:
            return None
        return bool(self._freed_record_number_pages)

    def note_freed_record_number(self, record_number):
        """Adjust segment of high and low freed record numbers."""
        self.note_freed_record_number_segment(
            *divmod(record_number, SegmentSize.db_segment_size))

    def note_freed_record_number_segment(
        self, segment, record_number_in_segment):
        """Adjust segment of high and low freed record numbers."""
        if self._freed_record_number_pages is None:
            self._freed_record_number_pages = []
            statement = ' '.join((
                'select',
                SQLITE_VALUE_COLUMN,
                'from',
                self._dbfile._control_database.control_file,
                'where',
                self._dbfile._control_database.control_file, '== ?',
                'order by',
                SQLITE_VALUE_COLUMN,
                ))
            values = (b'E',)
            cursor = self._dbfile.get_control_database().cursor()
            try:
                for record in cursor.execute(statement, values):
                    self._freed_record_number_pages.append(
                        int.from_bytes(record[0], byteorder='big'))
            finally:
                cursor.close()
        insert = bisect.bisect_left(self._freed_record_number_pages, segment)
        if self._freed_record_number_pages:
            if insert < len(self._freed_record_number_pages):
                if self._freed_record_number_pages[insert] == segment:
                    return
        self._freed_record_number_pages.insert(insert, segment)
        statement = ' '.join((
            'insert into',
            self._dbfile._control_database.control_file,
            '(',
            self._dbfile._control_database.control_file,
            ',',
            SQLITE_VALUE_COLUMN,
            ')',
            'values ( ? , ? )',
            ))
        values = (b'E', segment)
        cursor = self._dbfile.get_control_database().cursor()
        try:
            cursor.execute(statement, values)
        finally:
            cursor.close()

    def get_lowest_freed_record_number(self):
        """Return low record number in segments with freed record numbers."""
        if self._freed_record_number_pages is None:
            self._freed_record_number_pages = []
            statement = ' '.join((
                'select',
                SQLITE_VALUE_COLUMN,
                'from',
                self._dbfile._control_database.control_file,
                'where',
                self._dbfile._control_database.control_file, '== ?',
                'order by',
                SQLITE_VALUE_COLUMN,
                ))
            values = (b'E',)
            cursor = self._dbfile.get_control_database().cursor()
            try:
                for record in cursor.execute(statement, values):
                    self._freed_record_number_pages.append(record[0])
            finally:
                cursor.close()
        db_segment_size = SegmentSize.db_segment_size
        while len(self._freed_record_number_pages):
            s = self._freed_record_number_pages[0]
            lfrns = self._read_exists_segment(s)
            if lfrns is None:
                # Do not reuse record number on segment of high record number
                return 0
            try:
                first_zero_bit = lfrns.index(False, 0 if s else 1)
            except ValueError:
                statement = ' '.join((
                    'delete from',
                    self._dbfile._control_database.control_file,
                    'where',
                    self._dbfile._control_database.control_file, '== ? and',
                    SQLITE_VALUE_COLUMN, '== ?',
                    ))
                values = (b'E', s)
                cursor = self._dbfile.get_control_database().cursor()
                try:
                    cursor.execute(statement, values)
                finally:
                    cursor.close()
                del self._freed_record_number_pages[0]
                continue
            return s * db_segment_size + first_zero_bit
        else:
            return 0 # record number when inserting into RECNO database

    def _read_exists_segment(self, segment_number):
        """Return existence bit map for segment_number if not high segment."""
        # record keys are 1-based but segment_numbers are 0-based
        if segment_number < self._dbfile.get_existence_bits().segment_count - 1:
            ebm = Bitarray()
            ebm.frombytes(
                self._dbfile.get_existence_bits().get(segment_number + 1))
            return ebm
        return None

