# dbapi.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Object database using bsddb3, a Python wrapper for Berkeley DB.  A
repr(object) is held on the database, recoverable by the ast.literal_eval()
method.  Indicies whose values are derived from the object are supported.

Segmented databases take account of the local density of values per key on
secondary databases.  A primary database, always RECNO, is seen as a sequence
of fixed size intervals called segments.  Secondary databases have zero or one
keys per segment for each key.  The value associated with each segment key is a
number, a list of numbers, or a bit map, representing the record numbers in the
segment referenced by the key.  The absence of a segment key means the key has
no values in that segment.  Bit maps are fixed length and record lists are
variable length.  The maximum byte size of a list is less than or equal to the
byte size of a bit map.

Idea taken from DPT, an emulation of Model 204 which runs on Microsoft Windows.

"""

import os
import subprocess
from ast import literal_eval
import bisect
import re

import sys
_platform_win32 = sys.platform == 'win32'
_python_version = '.'.join(
    (str(sys.version_info[0]), 
     str(sys.version_info[1])))
del sys

# bsddb removed from Python 3.n
try:
    from bsddb3.db import (
        DB_KEYLAST, DB_CURRENT, DB_DUP, DB_DUPSORT, DB_NODUPDATA,
        DB_BTREE, DB_HASH, DB_RECNO, DB_UNKNOWN,
        DBEnv, DB, DB_CREATE, DB_FAST_STAT,
        DBKeyExistError, DBNotFoundError,
        )
except ImportError:
    from bsddb.db import (
        DB_KEYLAST, DB_CURRENT, DB_DUP, DB_DUPSORT, DB_NODUPDATA,
        DB_BTREE, DB_HASH, DB_RECNO, DB_UNKNOWN,
        DBEnv, DB, DB_CREATE, DB_FAST_STAT,
        DBKeyExistError, DBNotFoundError,
        )

from .api.bytebit import Bitarray, SINGLEBIT
from .api.database import (
    DatabaseError,
    Database,
    )
from .api.cursor import (
    Cursor,
    )
from .api.recordset import (
    SegmentBitarray,
    SegmentInt,
    SegmentList,
    RecordsetCursor,
    )
from .api.constants import (
    DB_DEFER_FOLDER, SECONDARY_FOLDER,
    PRIMARY, SECONDARY, FILE,
    PRIMARY_FIELDATTS, SECONDARY_FIELDATTS, DB_FIELDATTS,
    ACCESS_METHOD, HASH,
    FIELDS, SPT,
    LENGTH_SEGMENT_BITARRAY_REFERENCE,
    LENGTH_SEGMENT_LIST_REFERENCE,
    LENGTH_SEGMENT_RECORD_REFERENCE,
    SUBFILE_DELIMITER,
    )
from .api.segmentsize import SegmentSize


class DBapiError(DatabaseError):
    pass

        
class _DatabaseEncoders(object):
    
    """Define default record key encoder and decoder.
    """

    def encode_record_number(self, key):
        """Return repr(key).encode() because this is bsddb3 version of method.

        Typically used to convert primary key to secondary index format,
        using Berkeley DB terminology.
        
        """
        return repr(key).encode()

    def decode_record_number(self, skey):
        """Return literal_eval(skey.decode()) because this is bsddb3 version.

        Typically used to convert DB primary key held on secondary index,
        using Berkeley DB terminology, to integer.

        """
        return literal_eval(skey.decode())

    def encode_record_selector(self, key):
        """Return key.encode() because this is bsddb3 version of method.

        Typically used to convert a secondary index key value to database
        engine format, using Berkeley DB terminology.
        
        """
        return key.encode()

    def encode_record_key(self, key):
        """Return key.encode() because this is bsddb3 version of method.
        
        Specifically when comparing the key of a record read from database with
        the key used to find it.
        
        """
        return key.encode()


class DBbase(Database, _DatabaseEncoders):
    
    """Define a Berkeley DB database structure.
    
    Primary databases are created as DB_RECNO.
    Secondary databases are DB_BTREE with DB_DUPSORT set.

    Primary and secondary terminology comes from Berkeley DB documentation but
    the association technique is not used.

    The value part of a secondary key:value is a (segment, reference, count)
    tuple where segment follows DPT terminology.  Reference can be a record
    number relative to segment start, a reference to a list of record numbers,
    or a reference to a bitmap representing such record numbers.  Count is the
    number of records referenced by this value.

    Secondary databases are supported by two DB_RECNO databases, one for lists
    of record numbers and one for bitmap representations of record numbers. The
    reference is the key into the relevant DB_RECNO database.
    
    """

    def __init__(
        self,
        primary_class,
        secondary_class,
        database_specification,
        databasefolder,
        DBenvironment,
        *args,
        **kwargs):
        """Define database structure.
        
        databasefolder = full path for database directory
        database_specification = {name:{primary:name,
                         secondary:{name:name...},
                         }, ...}
        DBenvironment = {<DB property>:<value>, ...}
        primary_class = class implementing access to primary databases
        secondary_class = class implementing access to secondary databases

        """
        super(DBbase, self).__init__(
            database_specification,
            databasefolder,
            primary_class=primary_class,
            secondary_class=secondary_class,
            **kwargs)
        
        # The DBenv object
        self._dbenv = None
        
        # The active transaction object
        self._dbtxn = None
        
        # Parameters for setting up the DBenv object
        self._DBenvironment = DBenvironment

        # database_specification processing

        definitions = {k for k in self.get_dbsets()}

        for dbset, definition in self.get_dbsets().items():
            for s in definition.secondary:
                if s in definitions:
                    msg = ' '.join(['Secondary name', s,
                                    'for', dbset,
                                    'is used in another definition'])
                    raise DBapiError(msg)
                definitions.add(s)
        
        # Home directory for the DBenv
        self._home = self._databasefolder
        
        # The segment control object
        self._control = None

    def make_root(self, **kw):
        """Return Sqlite3Definition instance created from **kw arguments.
        """
        return _DBDefinition(database_instance=self, **kw)

    def backout(self):
        """Abort the active transaction and remove binding to txn object"""
        if self._dbtxn is not None:
            self._dbtxn.abort()
            self._dbtxn = None
            self._dbenv.txn_checkpoint(5)

    def close_context(self):
        """Close main and deferred update databases and environment."""
        if self._control:
            self._control.close()
        for table in self.get_dbsets().values():
            table.primary.close()
            for s in table.secondary.values():
                s.close()
        if self._dbenv is not None:
            self._dbenv.txn_checkpoint()
            self._dbenv.close()
            self._dbenv = None
        m = self.get_dbsets()
        database_specification = self._dbspec
        for n in database_specification:
            m[n].primary.set_control_database(None)
            for k, v in database_specification[n][SECONDARY].items():
                m[n].associate(k).set_primary_database(None)

    def close_database(self):
        """Close main and deferred update databases and environment.

        Introduced for compatibility with DPT.  There is a case for closing
        self._dbenv in this method rather than doing it all in close_context.
        
        """
        self.close_context()
            
    def commit(self):
        """Commit the active transaction and remove binding to txn object"""
        if self._dbtxn is not None:
            self._dbtxn.commit()
            self._dbtxn = None
            self._dbenv.txn_checkpoint(5)

    def db_compatibility_hack(self, record, srkey):
        """Convert record and return in (key, value) format.
        
        Do nothing as record is in (key, value) format on Berkeley DB.
        Added for compatibility with DPT.
        
        """
        return record

    def do_deferred_updates(self, pyscript, filepath):
        """Invoke a deferred update process and wait for it to finish.

        pyscript is the script to do the deferred update.
        filepath is a file or a sequence of files containing updates.

        """
        if _platform_win32:
            args = ['pythonw']
        else:
            args = [''.join(('python', _python_version))]
        
        if not os.path.isfile(pyscript):
            msg = ' '.join([repr(pyscript),
                            'is not an existing file'])
            raise DBapiError(msg)

        args.append(pyscript)
        
        try:
            if os.path.exists(filepath):
                paths = (filepath,)
            else:
                msg = ' '.join([repr(filepath),
                                'is not an existing file'])
                raise DBapiError(msg)
        except:
            paths = tuple(filepath)
            for fp in paths:
                if not os.path.isfile(fp):
                    msg = ' '.join([repr(fp),
                                    'is not an existing file'])
                    raise DBapiError(msg)

        args.append(os.path.abspath(self._home))
        args.extend(paths)

        return subprocess.Popen(args)

    def files_exist(self):
        """Return True if all defined files exist in self._home folder."""
        fileset = set(self.get_dbsets())
        filecount = len(fileset)
        for f in os.listdir(self._home):
            if f in fileset:
                fileset.remove(f)
        if len(fileset) == filecount:
            return None
        return len(fileset) == 0

    def get_associated_indicies(self, dbset):
        """Return frozenset of all table indicies for dbset."""
        s = self.get_dbsets()[dbset]
        return frozenset(set((s._dbset,)).union(s.secondary))

    def get_database_folder(self):
        """Return database folder name"""
        return self._home

    def get_database_filenames(self):
        """Return list of database file names"""
        names = ['___control']
        for f in os.listdir(self._home):
            b, e = os.path.splitext(f)
            if b == 'log' and e[1:].isdigit():
                names.append(f)
        for mv in self.get_dbsets().values():
            names.extend(mv.primary.get_filenames())
            for s in mv.secondary.values():
                names.extend(s.get_filenames())
        return names
    
    def get_database(self, dbset, dbname):
        """Return DB for dbname in dbset."""
        return self.get_dbsets()[dbset].primary.db_object
    
    def get_transaction(self):
        """Return object created by an earlier DBEnv.txn_begin() call or None
        if a DBEnv.abort() or DBEnv.commit() call is more recent."""
        return self._dbtxn

    def is_recno(self, dbset, dbname):
        """Return True if DB dbname in dbset is RECNO."""
        return self.get_dbsets()[dbset].associate(dbname).is_recno()

    def open_context(self):
        """Open all DBs."""
        try:
            os.mkdir(self._home)
        except FileExistsError:
            if not os.path.isdir(self._home):
                raise

        # Set up control database for segments with unused record numbers.
        self._control = DBControlFile(self)
        
        gbytes = self._DBenvironment.get('gbytes', 0)
        bytes_ = self._DBenvironment.get('bytes', 0)
        flags = self._DBenvironment.get('flags', 0)
        self._dbenv = DBEnv()
        if gbytes or bytes_:
            self._dbenv.set_cachesize(gbytes, bytes_)
        self._dbenv.open(self._home, flags)
        self.start_transaction()
        for table in self.get_dbsets().values():
            table.primary.open_root(self._dbenv)
            for s in table.secondary.values():
                s.open_root(self._dbenv)
        self._control.open_root(self._dbenv)
        self.commit()

        # Refer to primary from secondary for access to segment databases
        # Link each primary to control file for segment management
        m = self.get_dbsets()
        database_specification = self._dbspec
        for n in database_specification:
            m[n].primary.set_control_database(self._control)
            for k, v in database_specification[n][SECONDARY].items():
                m[n].associate(k).set_primary_database(m[n].primary)
        return True

    def allocate_and_open_contexts(self, closed_contexts):
        """Do nothing, present for DPT compatibility."""
        pass

    def get_packed_key(self, dbset, instance):
        """Return instance.key converted to string for dbset.

        encode_record_number provides this for RECNO databases.
        packed_key method of instance does conversion otherwise.

        """
        return self.encode_record_number(instance.key.pack())

    def decode_as_primary_key(self, dbset, pkey):
        """Return primary key after converting from secondary database format.

        No conversion is required if the primary DB is not RECNO.
        
        """
        return self.decode_record_number(pkey)

    def encode_primary_key(self, dbset, instance):
        """Convert instance.key for use as database value.
        
        For Berkeley DB just return self.get_packed_key().
        
        """
        return self.get_packed_key(dbset, instance)

    def do_database_task(
        self,
        taskmethod,
        logwidget=None,
        taskmethodargs={},
        use_specification_items=None,
        ):
        """Run taskmethod to perform database task.

        This method is structured to be compatible with the requirements of
        the sqlite3 version which is intended for use in a separate thread and
        must open a separate connection to the database.  Such action seems to
        be unnecessary in Berkeley DB so far.

        """
        # See sqlite3api.py for code which justifies existence of this method.
        taskmethod(self, logwidget, **taskmethodargs)

    def start_transaction(self):
        """Start transaction if none and bind txn object to self._dbtxn."""
        if self._dbtxn is None:
            self._dbtxn = self._dbenv.txn_begin()

    def cede_contexts_to_process(self, close_contexts):
        """Do nothing. Added for compatibility with apsw Sqlite3 interface."""
        pass

    def create_recordset_cursor(self, recordset):
        """Create and return a cursor for this recordset."""
        return RecordsetCursorDB(recordset, self)


class DBapi(DBbase):
    
    """Define a Berkeley DB database structure.
    
    Primary databases are created as DB_RECNO.
    Secondary databases are DB_BTREE with DB_DUPSORT set.

    Primary and secondary terminology comes from Berkeley DB documentation but
    the association technique is not used.

    The value part of a secondary key:value is a (segment, reference, count)
    tuple where segment follows DPT terminology.  Reference can be a record
    number relative to segment start, a reference to a list of record numbers,
    or a reference to a bitmap representing such record numbers.  Count is the
    number of records referenced by this value.

    Secondary databases are supported by two DB_RECNO databases, one for lists
    of record numbers and one for bitmap representations of record numbers. The
    reference is the key into the relevant DB_RECNO database.
    
    """

    def __init__(self, *args, **kargs):
        """Define database structure.  See superclass for *args and **kargs."""
        super(DBapi, self).__init__(
            DBPrimary, DBSecondary, *args, **kargs)

            
class DBFile(object):
    
    """Define a DB file with a cursor and open_root and close methods.
    """

    def __init__(
        self,
        dbfile,
        dbdesc,
        fieldatts,
        description,
        dbname,
        database_instance,
        ):
        """Define a DB file.
        
        dbfile=file name relative to environment home directory
        dbname=db name within file
        primary=True|False. DB is primary or secondary
        
        """
        super(DBFile, self).__init__()

        self._object = None
        self._dbfile = dbfile
        self._dbname = dbname
        self._fieldatts = dict()
        
        for attr in DB_FIELDATTS:
            self._fieldatts[attr] = fieldatts[attr]
        if description == None:
            description = dict()
        if not isinstance(description, dict):
            msg = ' '.join(['Attributes for index', dbname,
                            'in file', repr(dbdesc[PRIMARY]),
                            'must be a dictionary or "None"'])
            raise DBapiError(msg)
        
        for attr in description:
            if attr not in fieldatts:
                msg = ' '.join(['Attribute', repr(attr),
                                'for index', dbname,
                                'in file', repr(dbdesc[PRIMARY]),
                                'is not allowed'])
                raise DBapiError(msg)
            
            if not isinstance(description[attr], type(fieldatts[attr])):
                msg = ' '.join([attr, 'for field', dbname,
                                'in file', repr(dbdesc[PRIMARY]),
                                'is wrong type'])
                raise DBapiError(msg)
            
            if attr == SPT:
                if (description[attr] < 0 or
                    description[attr] > 100):
                    msg = ' '.join(['Split percentage for field',
                                    dbname, 'in file', repr(dbdesc[PRIMARY]),
                                    'is invalid'])
                    raise DBapiError(msg)

            if attr in DB_FIELDATTS:
                self._fieldatts[attr] = description[attr]

        self._transaction = database_instance.get_transaction

    @property
    def transaction(self):
        """Return the active transaction or None."""
        return self._transaction()

    def close(self):
        """Close DB and cursor."""
        if self._object is not None:
            self.db_object.close()
            self._object = None

    def open_root(self, dbenv):
        """Create DB in dbenv."""
        try:

            # Use a list for compatibility with dbduapi.py deferred updates.
            self._object = [DB(dbenv)]
            
        except:
            raise

    @property
    def db_object(self):
        """Return database file"""
        try:
            return self._object[0]
        except TypeError:
            if self._object is not None:
                raise
            return None

    @property
    def db_object_list(self):
        """Return list of database files"""
        return self._object

    def get_database_file(self):
        """Return database file name"""
        return self._dbfile

    def get_filenames(self):
        """Return tuple of database file names"""
        return self._dbfile,

            
class DBPrimaryFile(DBFile):
    
    """Define a DB file with a cursor and open_root and close methods.
    """

    def __init__(self, dbfile, dbdesc, *args):
        """Primary database file for name in description.
        
        dbfile=file name relative to environment home directory
        dbdesc=file description containing secondary database dbname
        
        """
        super(DBPrimaryFile, self).__init__(
            dbfile,
            dbdesc,
            PRIMARY_FIELDATTS,
            dbdesc[FIELDS][dbdesc[PRIMARY]],
            *args)

        # Description to be provided
        self._control_database = None

        # Existence bit map control structure (reuse record numbers)
        self._control_primary = FileControlPrimary(self)

        # Freed record list and bit map segment control structure
        self._control_secondary = FileControlSecondary(self)

        # Record number existence bit map for this primary database
        self._existence_bits = DBExistenceBitMap(self)

        # Inverted index record number lists for this primary database
        self._segment_list = DBSegmentList(self)

        # Inverted index record number bit maps for this primary database
        self._segment_bits = DBSegmentBitMap(self)

    def is_primary(self):
        """Return True."""
        # Maybe ask self._object
        return True

    def is_primary_recno(self):
        """Return True."""
        # Maybe ask self._object
        return True

    def is_recno(self):
        """Return True."""
        # Maybe ask self._object
        return True

    def is_value_recno(self):
        """Return False."""
        # Maybe ask self._object
        return False

    def open_root(self, *args):
        """Open primary database.  See superclass for *args"""
        super(DBPrimaryFile, self).open_root(*args)
        try:
            self.db_object.open(
                self._dbfile,
                self._dbname,
                DB_RECNO,
                DB_CREATE,
                txn=self.transaction)
        except:
            self._object = None
            raise
        self._segment_list.open_root(*args)
        self._segment_bits.open_root(*args)
        self._existence_bits.open_root(*args)

    def close(self):
        """Close inverted index databases then primary database."""
        self._segment_list.close()
        self._segment_bits.close()
        super(DBPrimaryFile, self).close()

    def get_control_database(self):
        """Return the database containing segment control data."""
        return self._control_database.get_control_database()

    # Added for sqlite3 compatibility.
    # Berkeley DB uses get_segment_list_database.
    def get_segment_list(self):
        """Return the segment list control data."""
        return self._segment_list

    def get_segment_list_database(self):
        """Return the database containing segment record number lists."""
        # Maybe use instance.get_segment_list().get_seg_object() instead of
        # instance.get_segment_list_database()
        return self._segment_list._seg_object

    # Added for sqlite3 compatibility.
    # Berkeley DB uses get_segment_bits_database.
    def get_segment_bits(self):
        """Return the segment bits control data."""
        return self._segment_bits

    def get_segment_bits_database(self):
        """Return the database containing segment record number bit maps."""
        # Maybe use instance.get_segment_bits().get_seg_object() instead of
        # instance.get_segment_bits_database()
        return self._segment_bits._seg_object

    def get_existence_bits(self):
        """Return the existence bit map control data."""
        return self._existence_bits

    def get_existence_bits_database(self):
        """Return the database containing existence bit map."""
        # Maybe use instance.get_existence_bits().get_seg_object() instead of
        # instance.get_existence_bits_database()
        return self._existence_bits._seg_object

    def get_control_primary(self):
        """Return the re-use record number control data."""
        return self._control_primary

    def get_control_secondary(self):
        """Return the segment control data."""
        return self._control_secondary

    def get_filenames(self):
        """Return tuple of database file names."""
        return (self.get_database_file(),
                self._existence_bits._seg_dbfile,
                self._segment_list._seg_dbfile,
                self._segment_bits._seg_dbfile)

    def set_control_database(self, database):
        """Set reference to segment control databases."""
        self._control_database = database

            
class DBSecondaryFile(DBFile):
    
    """Define a DB file with a cursor and open_root and close methods.
    """

    def __init__(self, dbfile, dbdesc, dbname, *args):
        """Secondary database file for name in description.
        
        dbfile=file name relative to environment home directory
        dbdesc=file description containing secondary database dbname
        dbname=entry in file description for this file
        
        """
        super(DBSecondaryFile, self).__init__(
            dbfile,
            dbdesc,
            SECONDARY_FIELDATTS,
            dbdesc[FIELDS][dbname],
            dbname,
            *args)
        self._primary_database = None

    def is_primary(self):
        """Return False."""
        # Maybe ask self._object
        return False

    def is_primary_recno(self):
        """Return True."""
        # Maybe ask self._object
        return True

    def is_recno(self):
        """Return False."""
        # Maybe ask self._object
        return False

    def is_value_recno(self):
        """Return True."""
        # Maybe ask self._object
        return True

    def open_root(self, *args):
        """Open secondary database.  See superclass for *args"""
        super(DBSecondaryFile, self).open_root(*args)
        try:
            self.db_object.set_flags(DB_DUPSORT)
            self.db_object.open(
                self._dbfile,
                self._dbname,
                DB_HASH if self._fieldatts[ACCESS_METHOD] == HASH else DB_BTREE,
                DB_CREATE,
                txn=self.transaction)
        except:
            self._object = None
            raise

    def get_primary_database(self):
        """Set reference to primary database to access segment databases."""
        return self._primary_database

    def get_primary_segment_bits(self):
        """Return the segment bitmap database of primary database."""
        return self._primary_database.get_segment_bits_database()

    def get_primary_segment_list(self):
        """Return the segment list database of primary database."""
        return self._primary_database.get_segment_list_database()

    def set_primary_database(self, database):
        """Set reference to primary database to access segment databases."""
        self._primary_database = database


class DBPrimary(DBPrimaryFile, _DatabaseEncoders):
    
    """Define a DB file with record access and deferred update methods.
    """

    def __init__(self, *args):
        """Primary database.  See superclass for *args."""
        super(DBPrimary, self).__init__(*args)
        self._clientcursors = dict()
        self._recordsets = dict()

    def make_cursor(self, dbobject, keyrange):
        """Create a cursor on the dbobject positiioned at start of keyrange."""
        c = CursorDBPrimary(dbobject,
                            keyrange=keyrange,
                            transaction=self.transaction)
        if c:
            self._clientcursors[c] = True
        return c
    
    def close(self):
        """Close DB and any cursors or recordsets."""
        for c in list(self._clientcursors.keys()):
            c.close()
        self._clientcursors.clear()
        for rs in list(self._recordsets.keys()):
            rs.close()
        self._recordsets.clear()
        super(DBPrimary, self).close()

    def delete(self, key, value):
        """Delete (key, value) from database."""
        # Transferred encode() from delete_instance, edit_instance, and
        # put_instance, calls (the only ones) so those methods are identical
        # to sqlite3 versions: well would have if value were mentioned.
        try:
            # Primary assumed to be not DUPSORT nor DUP
            cursor = self.db_object.cursor(txn=self.transaction)
            try:
                if cursor.set(key):
                    cursor.delete()

                    # See comment in DBapi.delete_instance()
                    #self.get_control_primary().note_freed_record_number(key)

            finally:
                cursor.close()
        except:
            pass

    def put(self, key, value):
        """Put (key, value) on database and return key for new RECNO records.

        The DB put method, or append for new RECNO records,is
        used for primary DBs with associated secondary DBs. The
        cursor put method is used otherwise.
        
        """
        # Transferred encode() from delete_instance, edit_instance, and
        # put_instance, calls (the only ones) so those methods are identical
        # to sqlite3 versions.
        # Primary assumed to be not DUPSORT nor DUP
        if not key: #key == 0:  # Change test to "key is None" when sure
            return self.db_object.append(value.encode(), txn=self.transaction)
        else:
            self.db_object.put(key, value.encode(), txn=self.transaction)
            return None

    def replace(self, key, oldvalue, newvalue):
        """Replace (key, oldvalue) with (key, newvalue) on DB.
        
        (key, newvalue) is put on DB only if (key, oldvalue) is on DB.
        
        """
        # Transferred encode() from delete_instance, edit_instance, and
        # put_instance, calls (the only ones) so those methods are identical
        # to sqlite3 versions.
        try:
            # Primary assumed to be not DUPSORT nor DUP
            cursor = self.db_object.cursor(txn=self.transaction)
            try:
                if cursor.set(key):
                    cursor.put(key, newvalue.encode(), DB_CURRENT)
            finally:
                cursor.close()
        except:
            pass

    def populate_recordset_key_like(self, recordset, key):
        """Raise exception - populate_recordset_key_like primary db."""
        raise DBapiError(
            ''.join(
                ('populate_recordset_key_like not available ',
                 'on primary database')))

    # Same as non bitmapped DBPrimary for one key.
    def populate_recordset_key(self, recordset, key=None):
        """Return recordset on database containing records for key."""
        r = self.db_object.get(key, txn=self.transaction)
        if r:
            s, rn = divmod(key, SegmentSize.db_segment_size)
            recordset[s] = SegmentList(
                s, None, records=rn.to_bytes(2, byteorder='big'))

    def populate_recordset_key_startswith(self, key, cache_size=1):
        """Raise DBapiError - populate_recordset_key_startswith primary db."""
        raise DBapiError(
            ''.join(
                ('populate_recordset_key_startswith not available ',
                 'on primary database')))

    def populate_recordset_key_range(
        self, recordset, keystart=None, keyend=None):
        """Return recordset on database containing records for key range."""
        if keystart is None:
            segment_start, recnum_start = 0, 1
        else:
            segment_start, recnum_start = divmod(keystart,
                                                 SegmentSize.db_segment_size)
        if keyend is not None:
            segment_end, record_number_end = divmod(keyend,
                                                    SegmentSize.db_segment_size)
        c = self.get_existence_bits_database().cursor(txn=self.transaction)
        try:
            r = c.set(segment_start + 1)
            while r:
                if keyend is not None:
                    if r[0] - 1 > segment_end:
                        break
                recordset[r[0] - 1] = SegmentBitarray(r[0] - 1,
                                                      None,
                                                      records=r[1])
                r = c.next()
        finally:
            c.close()
        try:
            recordset[segment_start][:recnum_start] = False
        except KeyError:
            pass
        try:
            recordset[segment_end][recnum_end + 1:] = False
        except KeyError:
            pass
    
    def populate_recordset_all(self, recordset):
        """Return recordset containing all referenced records."""
        c = self.get_existence_bits_database().cursor(txn=self.transaction)
        try:
            r = c.first()
            while r:
                recordset[r[0] - 1] = SegmentBitarray(r[0] - 1,
                                                      None,
                                                      records=r[1])
                r = c.next()
        finally:
            c.close()
    
    def file_records_under(self, recordset, key):
        """Raise exception for DBPrimary.file_records_under()."""
        raise DatabaseError(
            'file_records_under not implemented for DBPrimary')
    
    def unfile_records_under(self, key):
        """Raise exception for DBPrimary.unfile_records_under()."""
        raise DatabaseError(
            'unfile_records_under not implemented for DBPrimary')

    def get_first_primary_key_for_index_key(self, key):
        """Return first primary key for secondary key in dbname for dbname.

        Consider restricting use of this method to secondary DBs whose keys
        each have a unique value.
        
        """
        raise DBapiError((
            'get_first_primary_key_for_index_key for primary index'))

    def get_primary_record(self, key):
        """Return primary record (key, value) given primary key on dbset."""
        if key is None:
            return None
        c = self.db_object.cursor(txn=self.transaction)
        try:
            record = c.set(key)
        finally:
            c.close()
        try:
            return key, record[1].decode()
        except:
            if record is None:
                return record
            raise

    def find_values(self, valuespec):
        """Yield values meeting valuespec specification."""
        raise DatabaseError(
            'DBPrimary.find_values not implemented')

    def segment_delete(self, segment, record_number):
        """Remove record_number from existence bit map for segment."""
        # See dbduapi.py DBduPrimary.defer_put for model.  Main difference
        # is the write back to database is done immediately (and delete!!).
        # Get the segment existence bit map from database
        ebmb = self.get_existence_bits_database().get(segment + 1,
                                                      txn=self.transaction)
        if ebmb is None:
            # It does not exist so raise an exception
            raise DBapiError('Existence bit map for segment does not exist')
        else:
            # It does exist so convert database representation to bitarray
            ebm = Bitarray()
            ebm.frombytes(ebmb)
            # Set bit for record number and write segment back to database
            ebm[record_number] = False
            self.get_existence_bits_database().put(segment + 1,
                                                   ebm.tobytes(),
                                                   txn=self.transaction)

    def segment_put(self, segment, record_number):
        """Add record_number to existence bit map for segment."""
        # See dbduapi.py DBduPrimary.defer_put for model.  Main difference
        # is the write back to database is done immediately.
        # Get the segment existence bit map from database
        ebmb = self.get_existence_bits_database().get(segment + 1,
                                                      txn=self.transaction)
        if ebmb is None:
            # It does not exist so create a new empty one
            ebm = SegmentSize.empty_bitarray.copy()
        else:
            # It does exist so convert database representation to bitarray
            ebm = Bitarray()
            ebm.frombytes(ebmb)
        # Set bit for record number and write segment back to database
        ebm[record_number] = True
        self.get_existence_bits_database().put(segment + 1,
                                               ebm.tobytes(),
                                               txn=self.transaction)

    def get_high_record(self):
        """Return record with highest record number."""
        c = self.db_object.cursor(txn=self.transaction)
        try:
            return c.last()
        finally:
            c.close()


class DBSecondary(DBSecondaryFile, _DatabaseEncoders):
    
    """Define a DB file with record access and deferred update methods.
    """

    def __init__(self, *args):
        """Secondary database.  See superclass for *args."""
        super(DBSecondary, self).__init__(*args)
        self._clientcursors = dict()

    def make_cursor(self, dbobject, keyrange):
        """Create a cursor on the dbobject positiioned at start of keyrange."""
        c = CursorDBSecondary(dbobject,
                              keyrange=keyrange,
                              transaction=self.transaction)
        if c:
            self._clientcursors[c] = True
        return c
    
    def close(self):
        """Close DB and any cursors."""
        for c in list(self._clientcursors.keys()):
            c.close()
        self._clientcursors.clear()
        super(DBSecondary, self).close()
    
    def populate_recordset_key_like(self, recordset, key):
        """Return recordset containing database records with keys like key."""
        pattern = b'.*?' + key
        cursor = self.make_cursor(self, key)
        try:
            c = cursor._cursor
            r = c.first()
            while r:
                k, v = r
                if re.match(pattern, k, flags=re.IGNORECASE|re.DOTALL):
                    s = int.from_bytes(v[:4], byteorder='big')
                    if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                        srn = int.from_bytes(v[6:], byteorder='big')
                        bs = self.get_primary_segment_list().get(srn)
                        if bs is None:
                            raise DatabaseError('Segment record missing')
                        if s in recordset:
                            recordset[s] |= SegmentList(s, None, records=bs)
                        else:
                            recordset[s] = SegmentList(s, None, records=bs)
                    elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                        srn = int.from_bytes(v[7:], byteorder='big')
                        bs = self.get_primary_segment_bits().get(srn)
                        if bs is None:
                            raise DatabaseError('Segment record missing')
                        if s in recordset:
                            recordset[s] |= SegmentBitarray(s, None, records=bs)
                        else:
                            recordset[s] = SegmentBitarray(s, None, records=bs)
                    else:
                        if s in recordset:
                            recordset[s] |= SegmentList(s, None, records=v[4:])
                        else:
                            recordset[s] = SegmentList(s, None, records=v[4:])
                r = c.next()
        finally:
            cursor.close()
    
    def populate_recordset_key(self, recordset, key):
        """Return recordset of segments containing records for key."""
        cursor = self.make_cursor(self, key)
        try:
            c = cursor._cursor
            r = c.set_range(key)
            while r:
                k, v = r
                if k != key:
                    break
                s = int.from_bytes(v[:4], byteorder='big')
                if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn = int.from_bytes(v[6:], byteorder='big')
                    bs = self.get_primary_segment_list().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    recordset[s] = SegmentList(s, None, records=bs)
                elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn = int.from_bytes(v[7:], byteorder='big')
                    bs = self.get_primary_segment_bits().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    recordset[s] = SegmentBitarray(s, None, records=bs)
                else:
                    recordset[s] = SegmentList(s, None, records=v[4:])
                r = c.next()
        finally:
            cursor.close()

    def populate_recordset_key_startswith(self, recordset, key):
        """Return recordset on database containing records for keys starting."""
        cursor = self.make_cursor(self, key)
        try:
            c = cursor._cursor
            r = c.set_range(key)
            while r:
                if not r[0].startswith(key):
                    break
                v = r[1]
                s = int.from_bytes(v[:4], byteorder='big')
                if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn = int.from_bytes(v[6:], byteorder='big')
                    bs = self.get_primary_segment_list().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    sba = recordset[s]._bitarray# needs tidying
                    for i in range(0, len(bs), 2):
                        sba[int.from_bytes(bs[i:i+2], byteorder='big')] = True
                elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn = int.from_bytes(v[7:], byteorder='big')
                    bs = self.get_primary_segment_bits().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    sba = SegmentBitarray(s, None, records=bs)
                    recordset[s] |= sba
                else:
                    rn = int.from_bytes(v[4:], byteorder='big')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    recordset[s]._bitarray[rn] = True# needs tidying
                r = c.next()
        finally:
            cursor.close()

    def populate_recordset_key_range(
        self, recordset, keystart=None, keyend=None):
        """Return recordset on database containing records for key range."""
        cursor = self.make_cursor(self, keystart)
        try:
            c = cursor._cursor
            if keystart is None:
                r = c.first()
            else:
                r = c.set_range(keystart)
            while r:
                if keyend is not None:
                    if r[0] > keyend:
                        break
                v = r[1]
                s = int.from_bytes(v[:4], byteorder='big')
                if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn = int.from_bytes(v[6:], byteorder='big')
                    bs = self.get_primary_segment_list().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    sba = recordset[s]._bitarray# needs tidying
                    for i in range(0, len(bs), 2):
                        sba[int.from_bytes(bs[i:i+2], byteorder='big')] = True
                elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn = int.from_bytes(v[7:], byteorder='big')
                    bs = self.get_primary_segment_bits().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    sba = SegmentBitarray(s, None, records=bs)
                    recordset[s] |= sba
                else:
                    rn = int.from_bytes(v[4:], byteorder='big')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    recordset[s]._bitarray[rn] = True# needs tidying
                r = c.next()
        finally:
            cursor.close()
    
    def populate_recordset_all(self, recordset):
        """Return recordset containing all referenced records."""
        cursor = self.make_cursor(self, None)
        try:
            c = cursor._cursor
            r = c.first()
            while r:
                v = r[1]
                s = int.from_bytes(v[:4], byteorder='big')
                if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn = int.from_bytes(v[6:], byteorder='big')
                    bs = self.get_primary_segment_list().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    sba = recordset[s]._bitarray# needs tidying
                    for i in range(0, len(bs), 2):
                        sba[int.from_bytes(bs[i:i+2], byteorder='big')] = True
                elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn = int.from_bytes(v[7:], byteorder='big')
                    bs = self.get_primary_segment_bits().get(srn)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    sba = SegmentBitarray(s, None, records=bs)
                    recordset[s] |= sba
                else:
                    rn = int.from_bytes(v[4:], byteorder='big')
                    if s not in recordset:
                        recordset[s] = SegmentBitarray(s, None)
                    recordset[s]._bitarray[rn] = True# needs tidying
                r = c.next()
        finally:
            cursor.close()

    def populate_segment(self, segment):
        """Helper for segment_delete and segment_put methods."""
        k, v = segment
        s = int.from_bytes(k, byteorder='big')
        if len(v) + len(k) == LENGTH_SEGMENT_LIST_REFERENCE:
            srn = int.from_bytes(v[2:], byteorder='big')
            bs = self.get_primary_segment_list().get(srn)
            if bs is None:
                raise DatabaseError('Segment record missing')
            return SegmentList(s, None, records=bs)
        elif len(v) + len(k) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
            srn = int.from_bytes(v[3:], byteorder='big')
            bs = self.get_primary_segment_bits().get(srn)
            if bs is None:
                raise DatabaseError('Segment record missing')
            return SegmentBitarray(s, None, records=bs)
        else:
            return SegmentList(s, None, records=v)

    def make_segment(self, key, segment_number, record_count, records):
        """Return a Segment subclass instance created from arguments."""
        if record_count == 1:
            return SegmentInt(
                segment_number,
                None,
                records=records.to_bytes(2, byteorder='big'))
        else:
            if len(records) == SegmentSize.db_segment_size_bytes:
                return SegmentBitarray(segment_number, None, records=records)
            else:
                return SegmentList(segment_number, None, records=records)
    
    def file_records_under(self, recordset, key):
        """Replace records for index dbname[key] with recordset records."""
        # Delete existing segments for key
        self.unfile_records_under(key)
        # Put new segments for key
        cursor = self.db_object.cursor(txn=self.transaction)
        try:
            recordset.normalize()
            for sn in recordset.sorted_segnums:
                if isinstance(recordset.rs_segments[sn], SegmentBitarray):
                    count = recordset.rs_segments[sn].count_records()
                    # stub call to get srn_bits from reuse stack
                    srn_bits = self.get_primary_database(
                        ).get_control_secondary().get_freed_bits_page()
                    if srn_bits == 0:
                        srn_bits = self.get_primary_segment_bits(
                            ).append(recordset.rs_segments[sn].tobytes(),
                                     txn=self.transaction)
                    else:
                        self.get_primary_segment_bits(
                            ).put(srn_bits,
                                  recordset.rs_segments[sn].tobytes(),
                                  txn=self.transaction)
                    cursor.put(
                        key,
                        b''.join(
                            (sn.to_bytes(4, byteorder='big'),
                             count.to_bytes(3, byteorder='big'),
                             srn_bits.to_bytes(4, byteorder='big'))),
                        DB_KEYLAST)
                elif isinstance(recordset.rs_segments[sn], SegmentList):
                    count = recordset.rs_segments[sn].count_records()
                    # stub call to get srn_list from reuse stack
                    srn_list = self.get_primary_database(
                        ).get_control_secondary().get_freed_list_page()
                    if srn_list == 0:
                        srn_list = self.get_primary_segment_list().append(
                            recordset.rs_segments[sn].tobytes(),
                            txn=self.transaction)
                    else:
                        self.get_primary_segment_list().put(
                            srn_list,
                            recordset.rs_segments[sn].tobytes(),
                            txn=self.transaction)
                    cursor.put(
                        key,
                        b''.join(
                            (sn.to_bytes(4, byteorder='big'),
                             count.to_bytes(2, byteorder='big'),
                             srn_list.to_bytes(4, byteorder='big'))),
                        DB_KEYLAST)
                elif isinstance(recordset.rs_segments[sn], SegmentInt):
                    cursor.put(
                        key,
                        b''.join(
                            (sn.to_bytes(4, byteorder='big'),
                             recordset.rs_segments[sn].tobytes())),
                        DB_KEYLAST)
        finally:
            cursor.close()
    
    def unfile_records_under(self, key):
        """Delete the reference to records in file under key.

        The existing reference by key, usually created by file_records_under,
        is deleted.

        """
        # Delete existing segments for key
        cursor = self.db_object.cursor(txn=self.transaction)
        try:
            r = cursor.set_range(key)
            while r:
                k, v = r
                if k != key:
                    break
                sr = int.from_bytes(v[:4], byteorder='big')
                if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn_list = int.from_bytes(v[6:], byteorder='big')
                    # stub call to put srn_list on reuse stack
                    self.get_primary_database().get_control_secondary(
                        ).note_freed_list_page(srn_list)
                    # ok if reuse bitmap but not if reuse stack
                    self.get_primary_segment_list(
                        ).delete(srn_list, txn=self.transaction)
                    #cursor.delete()
                elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn_bits = int.from_bytes(v[7:], byteorder='big')
                    # stub call to put srn_bits on reuse stack
                    self.get_primary_database().get_control_secondary(
                        ).note_freed_bits_page(srn_bits)
                    # ok if reuse bitmap but not if reuse stack
                    self.get_primary_segment_bits(
                        ).delete(srn_bits, txn=self.transaction)
                    #cursor.delete()
                #elif record_number == int.from_bytes(v[4:], byteorder='big'):
                    #cursor.delete()
                r = cursor.next()
            try:
                self.db_object.delete(key, txn=self.transaction)
            except DBNotFoundError:
                pass
        finally:
            cursor.close()

    def get_first_primary_key_for_index_key(self, key):
        """Return first primary key for secondary key in dbname for dbname.

        This method should be ued only on secondary DBs whose keys each have a
        unique value.
        
        """
        if isinstance(key, str):
            key = key.encode()
        try:
            c = self.db_object.cursor(txn=self.transaction)
            try:
                v = c.set(key)[1]
            finally:
                c.close()
        except:
            if not isinstance(key, bytes):
                raise
            return None
        if len(v) != LENGTH_SEGMENT_RECORD_REFERENCE:
            raise DBapiError('Index must refer to unique record')
        return (
            int.from_bytes(v[:4],
                           byteorder='big') * SegmentSize.db_segment_size +
            int.from_bytes(v[4:], byteorder='big'))

    def find_values(self, valuespec):
        """Yield values meeting valuespec specification."""
        cursor = self.db_object.cursor(txn=self.transaction)
        try:
            if valuespec.above_value and valuespec.below_value:
                r = cursor.set_range(valuespec.above_value.encode())
                if r:
                    if r[0] == valuespec.above_value.encode():
                        r = cursor.next_nodup()
                while r:
                    k = r[0].decode()
                    if k >= valuespec.below_value:
                        break
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            elif valuespec.above_value and valuespec.to_value:
                r = cursor.set_range(valuespec.above_value.encode())
                if r:
                    if r[0] == valuespec.above_value.encode():
                        r = cursor.next_nodup()
                while r:
                    k = r[0].decode()
                    if k > valuespec.to_value:
                        break
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            elif valuespec.from_value and valuespec.to_value:
                r = cursor.set_range(valuespec.from_value.encode())
                while r:
                    k = r[0].decode()
                    if k > valuespec.to_value:
                        break
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            elif valuespec.from_value and valuespec.below_value:
                r = cursor.set_range(valuespec.from_value.encode())
                while r:
                    k = r[0].decode()
                    if k >= valuespec.below_value:
                        break
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            elif valuespec.above_value:
                r = cursor.set_range(valuespec.above_value.encode())
                if r:
                    if r[0] == valuespec.above_value.encode():
                        r = cursor.next_nodup()
                while r:
                    k = r[0].decode()
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            elif valuespec.from_value:
                r = cursor.set_range(valuespec.from_value.encode())
                while r:
                    k = r[0].decode()
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            elif valuespec.to_value:
                r = cursor.first()
                while r:
                    k = r[0].decode()
                    if k > valuespec.to_value:
                        break
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            elif valuespec.below_value:
                r = cursor.first()
                while r:
                    k = r[0].decode()
                    if k >= valuespec.below_value:
                        break
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
            else:
                r = cursor.first()
                while r:
                    k = r[0].decode()
                    if valuespec.apply_pattern_and_set_filters_to_value(k):
                        yield k
                    r = cursor.next_nodup()
        finally:
            cursor.close()
    
    def segment_delete(self, key, segment, record_number):
        """Remove record_number from segment for key and write to database."""
        # Transferred from delete_instance, edit_instance, and put_instance,
        # calls (the only ones) so those methods are identical to sqlite3
        # versions.
        key = key.encode()
        # See DBSecondary.segment_put (in this class definition) for model.
        cursor = self.db_object.cursor(txn=self.transaction)
        try:
            r = cursor.set_range(key)
            while r:
                k, v = r
                if k != key:
                    # Assume that multiple requests to delete an index value
                    # have been made for a record.  The segment_put method uses
                    # sets to avoid adding multiple entries.  Consider using
                    # set rather than list in the pack method of the subclass
                    # of Value if this will happen a lot.
                    return
                sr = int.from_bytes(v[:4], byteorder='big')
                if sr < segment:
                    r = cursor.next_dup()
                    continue
                elif sr > segment:
                    return
                if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn_list = int.from_bytes(v[6:], byteorder='big')
                    bs = self.get_primary_segment_list(
                        ).get(srn_list, txn=self.transaction)
                    recnums = {int.from_bytes(bs[i:i+2], byteorder='big')
                               for i in range(0, len(bs), 2)}
                    # ignore possibility record_number already absent
                    recnums.discard(record_number)
                    count = len(recnums)
                    if count < 2:
                        for rn in recnums:
                            ref = b''.join(
                                (segment.to_bytes(4, byteorder='big'),
                                 rn.to_bytes(2, byteorder='big')))
                        # stub call to put srn_list on reuse stack
                        self.get_primary_database().get_control_secondary(
                            ).note_freed_list_page(srn_list)
                        # ok if reuse bitmap but not if reuse stack
                        self.get_primary_segment_list(
                            ).delete(srn_list, txn=self.transaction)
                        cursor.delete()
                        if count:
                            cursor.put(k, ref, DB_KEYLAST)
                    else:
                        seg = b''.join(tuple(
                            rn.to_bytes(length=2, byteorder='big')
                            for rn in sorted(recnums)))
                        self.get_primary_segment_list(
                            ).put(srn_list, seg, txn=self.transaction)
                        cursor.delete()
                        cursor.put(
                            k,
                            b''.join(
                                (v[:4],
                                 count.to_bytes(2, byteorder='big'),
                                 v[6:])),
                            DB_KEYLAST)
                elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn_bits = int.from_bytes(v[7:], byteorder='big')
                    bs = self.get_primary_segment_bits(
                        ).get(srn_bits, txn=self.transaction)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    recnums = Bitarray()
                    recnums.frombytes(bs)
                    # ignore possibility record_number already absent
                    recnums[record_number] = False
                    count = recnums.count()
                    if count > SegmentSize.db_lower_conversion_limit:
                        self.get_primary_segment_bits().put(
                            srn_bits, recnums.tobytes(), txn=self.transaction)
                        cursor.delete()
                        cursor.put(
                            k,
                            b''.join(
                                (v[:4],
                                 recnums.count().to_bytes(3, byteorder='big'),
                                 v[7:])),
                            DB_KEYLAST)
                    else:
                        recnums = {rn for rn in recnums.search(SINGLEBIT)}
                        # stub call to get srn_list from reuse stack
                        srn_list = self.get_primary_database(
                            ).get_control_secondary().get_freed_list_page()
                        if srn_list == 0:
                            srn_list = self.get_primary_segment_list().append(
                                b''.join(
                                    [rn.to_bytes(2, byteorder='big')
                                     for rn in sorted(recnums)]),
                                txn=self.transaction)
                        else:
                            self.get_primary_segment_list().put(
                                srn_list,
                                b''.join(
                                    [rn.to_bytes(2, byteorder='big')
                                     for rn in sorted(recnums)]),
                                txn=self.transaction)
                        cursor.delete()
                        cursor.put(
                            k,
                            b''.join(
                                (v[:4],
                                 len(recnums).to_bytes(2, byteorder='big'),
                                 srn_list.to_bytes(4, byteorder='big'))),
                            DB_KEYLAST)
                        # stub call to put srn_bits on reuse stack
                        self.get_primary_database().get_control_secondary(
                            ).note_freed_bits_page(srn_bits)
                        # ok if reuse bitmap but not if reuse stack
                        self.get_primary_segment_bits(
                            ).delete(srn_bits, txn=self.transaction)
                elif record_number == int.from_bytes(v[4:], byteorder='big'):
                    cursor.delete()
                return
        finally:
            cursor.close()
    
    def segment_put(self, key, segment, record_number):
        """Add record_number to segment for key and write to database."""
        # Transferred from delete_instance, edit_instance, and put_instance,
        # calls (the only ones) so those methods are identical to sqlite3
        # versions.
        key = key.encode()
        # See dbduapi.py DBduSecondary.defer_put for model.
        # The dance to find the segment record is a reason to convert these
        # secondary databases from DUP to NODUP.  Also a NODUP database allows
        # implementation equivalent to DPT 'FOR EACH VALUE' directly and easy
        # counting of values for manipulation of scrollbar sliders.
        # Assumption is that new records usually go in last segment for value.
        cursor = self.db_object.cursor(txn=self.transaction)
        try:
            r = cursor.set_range(key)
            while r:
                k, v = r
                if k != key:
                    # No index entry for key yet
                    cursor.put(
                        key,
                        b''.join(
                            (segment.to_bytes(4, byteorder='big'),
                             record_number.to_bytes(2, byteorder='big'))),
                        DB_KEYLAST)
                    return
                sr = int.from_bytes(v[:4], byteorder='big')
                if sr < segment:
                    r = cursor.next_dup()
                    continue
                elif sr > segment:
                    cursor.put(
                        k,
                        b''.join(
                            (segment.to_bytes(4, byteorder='big'),
                             record_number.to_bytes(2, byteorder='big'))),
                        DB_KEYLAST)
                    return
                if len(v) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn_list = int.from_bytes(v[6:], byteorder='big')
                    bs = self.get_primary_segment_list(
                        ).get(srn_list, txn=self.transaction)
                    recnums = {int.from_bytes(bs[i:i+2], byteorder='big')
                               for i in range(0, len(bs), 2)}
                    # ignore possibility record_number already present
                    recnums.add(record_number)
                    count = len(recnums)
                    if count > SegmentSize.db_upper_conversion_limit:
                        seg = SegmentSize.empty_bitarray.copy()
                        for rn in recnums:
                            seg[rn] = True
                        # stub call to put srn_list on reuse stack
                        self.get_primary_database().get_control_secondary(
                            ).note_freed_list_page(srn_list)
                        # ok if reuse bitmap but not if reuse stack
                        self.get_primary_segment_list(
                            ).delete(srn_list, txn=self.transaction)
                        # stub call to get srn_bits from reuse stack
                        srn_bits = self.get_primary_database(
                            ).get_control_secondary().get_freed_bits_page()
                        if srn_bits == 0:
                            srn_bits = self.get_primary_segment_bits(
                                ).append(seg.tobytes(), txn=self.transaction)
                        else:
                            self.get_primary_segment_bits(
                                ).put(srn_bits,
                                      seg.tobytes(),
                                      txn=self.transaction)
                        cursor.delete()
                        cursor.put(
                            k,
                            b''.join(
                                (v[:4],
                                 count.to_bytes(3, byteorder='big'),
                                 srn_bits.to_bytes(4, byteorder='big'))),
                            DB_KEYLAST)
                    else:
                        seg = b''.join(tuple(
                            rn.to_bytes(length=2, byteorder='big')
                            for rn in sorted(recnums)))
                        self.get_primary_segment_list(
                            ).put(srn_list, seg, txn=self.transaction)
                        cursor.delete()
                        cursor.put(
                            k,
                            b''.join(
                                (v[:4],
                                 count.to_bytes(2, byteorder='big'),
                                 v[6:])),
                            DB_KEYLAST)
                elif len(v) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn = int.from_bytes(v[7:], byteorder='big')
                    bs = self.get_primary_segment_bits(
                        ).get(srn, txn=self.transaction)
                    if bs is None:
                        raise DatabaseError('Segment record missing')
                    recnums = Bitarray()
                    recnums.frombytes(bs)
                    recnums[record_number] = True
                    self.get_primary_segment_bits(
                        ).put(srn, recnums.tobytes(), txn=self.transaction)
                    cursor.delete()
                    cursor.put(
                        k,
                        b''.join(
                            (v[:4],
                             recnums.count().to_bytes(3, byteorder='big'),
                             v[7:])),
                        DB_KEYLAST)
                else:
                    rn = int.from_bytes(v[4:], byteorder='big')
                    if rn > record_number:
                        # stub call to get srn_list from reuse stack
                        srn_list = self.get_primary_database(
                            ).get_control_secondary().get_freed_list_page()
                        if srn_list == 0:
                            srn_list = self.get_primary_segment_list().append(
                                b''.join(
                                    (record_number.to_bytes(
                                        length=2, byteorder='big'),
                                     rn.to_bytes(length=2, byteorder='big'))),
                                txn=self.transaction)
                        else:
                            self.get_primary_segment_list().put(
                                srn_list,
                                b''.join(
                                    (record_number.to_bytes(
                                        length=2, byteorder='big'),
                                     rn.to_bytes(length=2, byteorder='big'))),
                                txn=self.transaction)
                        cursor.delete()
                        cursor.put(
                            k,
                            b''.join(
                                (v[:4],
                                 b'\x00\x02',
                                 srn_list.to_bytes(4, byteorder='big'))),
                            DB_KEYLAST)
                    elif rn < record_number:
                        # stub call to get srn_list from reuse stack
                        srn_list = self.get_primary_database(
                            ).get_control_secondary().get_freed_list_page()
                        if srn_list == 0:
                            srn_list = self.get_primary_segment_list().append(
                                b''.join(
                                    (rn.to_bytes(length=2, byteorder='big'),
                                     record_number.to_bytes(
                                        length=2, byteorder='big'))),
                                txn=self.transaction)
                        else:
                            self.get_primary_segment_list().put(
                                srn_list,
                                b''.join(
                                    (rn.to_bytes(length=2, byteorder='big'),
                                     record_number.to_bytes(
                                        length=2, byteorder='big'))),
                                txn=self.transaction)
                        cursor.delete()
                        cursor.put(
                            k,
                            b''.join(
                                (v[:4],
                                 b'\x00\x02',
                                 srn_list.to_bytes(4, byteorder='big'))),
                            DB_KEYLAST)
                return
            else:
                # No index entry for key yet because database empty
                cursor.put(
                    key,
                    b''.join(
                        (segment.to_bytes(4, byteorder='big'),
                         record_number.to_bytes(2, byteorder='big'))),
                    DB_KEYLAST)
        finally:
            cursor.close()


class _DBDefinition(object):
    """A definition derived from a FileSpec entry fitted to Berkeley DB."""

    def __init__(self,
                 dbset=None,
                 specification=None,
                 primary_class=None,
                 secondary_class=None,
                 field_name_converter=None,
                 database_instance=None,
                 **kw):
        """Create primary and secondary instances from specification for dbset.

        This class should be instantiated via the make_root() call which is
        called after much of the specification validation, which cannot be
        done internally, has been done.

        """
        super().__init__()
        for a in (dbset,
                  specification,
                  primary_class,
                  secondary_class,
                  field_name_converter):
            if a is None:
                msg = ' '.join(['Cannot construct definition from',
                                'specification'
                                'because a mandatory argument is missing'])
                raise DBapiError(msg)
        self._dbset = dbset
        primary = specification[PRIMARY]
        self.primary = primary_class(
            primary,
            specification,
            primary,
            database_instance)
        self.secondary = {}
        self.dbname_to_secondary_key = {}
        if SECONDARY in specification:
            for sname, secondary in specification[SECONDARY].items():
                if not isinstance(sname, str):
                    msg = ' '.join(['Secondary name',
                                    'for', dbset,
                                    'must be a string'])
                    raise DBapiError(msg)

                if secondary is None:
                    secondary = field_name_converter(sname)
                
                if secondary == primary:
                    msg = ' '.join(['Secondary name', secondary,
                                    'for', dbset,
                                    'cannot be same as primary'])
                    raise DBapiError(msg)

                if secondary not in specification[FIELDS]:
                    msg = ' '.join(['Secondary name',
                                    secondary,
                                    'for', dbset, 'does not have',
                                    'a description'])
                    raise DBapiError(msg)

                self.secondary[secondary] = secondary_class(
                    SUBFILE_DELIMITER.join((dbset, secondary)),
                    specification,
                    secondary,
                    database_instance)
                self.dbname_to_secondary_key[sname] = secondary
        for v in self.secondary.values():
            v.set_primary_database(self.primary)

    def associate(self, dbname):
        """Return primary definition for dbname."""
        if dbname == self._dbset:
            return self.primary
        return self.secondary[self.dbname_to_secondary_key[dbname]]


class CursorDB(Cursor):
    
    """Define bsddb3 style cursor methods on a Berkeley DB database.

    Segmented should be read as the DPT database engine usage.

    The value part of (key, value) on secondary databases is either:

        primary key (segment and record number)
        reference to a list of primary keys for a segment
        reference to a bit map of primary keys for a segment

    References are to records on RECNO databases, one each for lists and bit
    maps, containing the primary keys.
    
    Mostly thin wrappers around the identically named bsddb3 cursor methods.

    Keys and values are converted to integer and str from Berkeley DB internal
    representation.

    Sibling classes in other ~api modules provide this interface for other
    database engines.

    The Berkeley DB Cursor put methods are not wrapped.

    The intended use is, using put_record as the example:

        cursor = <Database subclass instance>.database_cursor(...)
        record = cursor.<get method>(...)
        <Record subclass instance>.value.<attributes> = <Function(record)>
        <Record subclass instance>.put_record(...)

    where put_record is boilerplate, creating cursors if needed, and all the
    work to make the correct thing happen is in <Function>.
    
    """

    def __init__(self, dbset, keyrange=None, transaction=None):
        """Define a cursor using the Berkeley DB engine."""
        super(CursorDB, self).__init__(dbset)
        self._transaction = transaction
        if dbset.db_object_list is not None:
            self._cursor = dbset.db_object.cursor(txn=transaction)
        self._current_segment = None
        self._current_segment_number = None
        self._current_record_number_in_segment = None

    def close(self):
        """Delete database cursor."""
        try:
            del self._dbset._clientcursors[self]
        except:
            pass
        try:
            self._cursor.close()
        except:
            pass
        self._cursor = None
        self._dbset = None
        self.set_partial_key(None)

    def database_cursor_exists(self):
        """Return True if database cursor exists and False otherwise."""
        return bool(self._cursor)

    def refresh_recordset(self, instance=None):
        """Refresh records for datagrid access after database update.

        Do nothing in Berkeley DB.  The cursor (for the datagrid) accesses
        database directly.  There are no intervening data structures which
        could be inconsistent.

        """
        pass

    def get_partial(self):
        """Return self._partial."""
        return self._partial

    def get_converted_partial(self):
        """Return self._partial as it would be held on database."""
        return self._partial.encode()

    def get_partial_with_wildcard(self):
        """Return self._partial with wildcard suffix appended."""
        raise DatabaseError('get_partial_with_wildcard not implemented')

    def get_converted_partial_with_wildcard(self):
        """Return converted self._partial with wildcard suffix appended."""
        # Berkeley DB uses a 'startswith(...)' technique
        return self._partial.encode()


class CursorDBPrimary(CursorDB):
    
    """Define bsddb3 cursor methods for primary segmented Berkeley DB database.

    The database must be a RECNO database.
    
    """

    def count_records(self):
        """Return record count."""
        return self._dbset.db_object.stat(flags=DB_FAST_STAT,
                                          txn=self._transaction)['ndata']

    def first(self):
        """Return first record taking partial key into account."""
        return self._decode_record(self._cursor.first())

    def get_position_of_record(self, record=None):
        """Return position of record in file or 0 (zero)."""
        ebd = self._dbset.get_existence_bits_database()
        try:
            segment, record_number = divmod(record[0],
                                            SegmentSize.db_segment_size)
            position = 0
            for i in range(segment):
                sebm = Bitarray()
                sebm.frombytes(ebd.get(i + 1, txn=self._transaction))
                position += sebm.count()
            sebm = Bitarray()
            sebm.frombytes(ebd.get(segment + 1, txn=self._transaction))
            try:
                position += sebm.search(SINGLEBIT).index(record_number)
            except ValueError:
                position += bisect.bisect_left(
                    record_number, sebm.search(SINGLEBIT))
            return position
        except:
            if record is None:
                return 0
            raise

    def get_record_at_position(self, position=None):
        """Return record for positionth record in file or None."""
        if position is None:
            return None
        ebd = self._dbset.get_existence_bits_database()
        count = 0
        abspos = abs(position)
        ebdc = ebd.cursor(txn=self._transaction)
        try:
            if position < 0:
                r = ebdc.last()
                while r:
                    sebm = Bitarray()
                    sebm.frombytes(r[1])
                    sc = sebm.count()
                    if count + sc < abspos:
                        count += sc
                        r = ebdc.prev()
                        continue
                    recno = sebm.search(SINGLEBIT)[position + count] + (
                        (r[0] - 1) * SegmentSize.db_segment_size)
                    ebdc.close()
                    return self._decode_record(self._cursor.set(recno))
            else:
                r = ebdc.first()
                while r:
                    sebm = Bitarray()
                    sebm.frombytes(r[1])
                    sc = sebm.count()
                    if count + sc < abspos:
                        count += sc
                        r = ebdc.next()
                        continue
                    recno = sebm.search(SINGLEBIT)[position - count] + (
                        (r[0] - 1) * SegmentSize.db_segment_size)
                    ebdc.close()
                    return self._decode_record(self._cursor.set(recno))
        finally:
            ebdc.close()
        return None

    def last(self):
        """Return last record taking partial key into account."""
        return self._decode_record(self._cursor.last())

    def nearest(self, key):
        """Return nearest record to key taking partial key into account."""
        return self._decode_record(self._cursor.set_range(key))

    def next(self):
        """Return next record taking partial key into account."""
        return self._decode_record(self._cursor.next())

    def prev(self):
        """Return previous record taking partial key into account."""
        return self._decode_record(self._cursor.prev())

    def setat(self, record):
        """Return current record after positioning cursor at record.

        Take partial key into account.
        
        Words used in bsddb3 (Python) to describe set and set_both say
        (key, value) is returned while Berkeley DB description seems to
        say that value is returned by the corresponding C functions.
        Do not know if there is a difference to go with the words but
        bsddb3 works as specified.

        """
        return self._decode_record(self._cursor.set(record[0]))

    def set_partial_key(self, partial):
        """Set partial key to None for primary cursor."""
        self._partial = None

    def _decode_record(self, record):
        """Return decoded (key, value) of record."""
        try:
            k, v = record
            return k, v.decode()
        except:
            if record is None:
                return record
            raise

    def _get_record(self, record):
        """Return record matching key or partial key or None if no match."""
        raise DBapiError('_get_record not implemented')

    def refresh_recordset(self, instance=None):
        """Refresh records for datagrid access after database update.

        The bitmap for the record set may not match the existence bitmap.

        """
        #raise DBapiError('refresh_recordset not implemented')


class CursorDBSecondary(CursorDB):
    
    """Define bsddb3 cursor methods on secondary segmented Berkeley DB database.

    The database must be a BTREE database.
    
    """

    def __init__(self, *a, **k):#dbset, keyrange=None):
        """Define a cursor using the Berkeley DB engine."""
        super(CursorDBSecondary, self).__init__(*a, **k)#dbset, keyrange=keyrange)
        self._segment_bits = self._dbset.get_primary_segment_bits()
        self._segment_list = self._dbset.get_primary_segment_list()

    def count_records(self):
        """Return record count."""
        #if self.get_partial() is None:
        if self.get_partial() in (None, False):
            count = 0
            r = self._cursor.first()
            while r:
                if len(r[1]) == LENGTH_SEGMENT_LIST_REFERENCE:
                    count += int.from_bytes(r[1][4:6], byteorder='big')
                elif len(r[1]) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    count += int.from_bytes(r[1][4:7], byteorder='big')
                else:
                    count += 1
                r = self._cursor.next()
            return count
        else:
            count = 0
            r = self._cursor.set_range(
                self.get_converted_partial_with_wildcard())
            while r:
                if not r[0].startswith(self.get_converted_partial()):
                    break
                if len(r[1]) == LENGTH_SEGMENT_LIST_REFERENCE:
                    count += int.from_bytes(r[1][4:6], byteorder='big')
                elif len(r[1]) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    count += int.from_bytes(r[1][4:7], byteorder='big')
                else:
                    count += 1
                r = self._cursor.next()
            return count

    def first(self):
        """Return first record taking partial key into account."""
        if self.get_partial() is None:
            try:
                k, v = self._first()
            except TypeError:
                return None
            return k.decode(), v
        elif self.get_partial() is False:
            return None
        else:
            return self.nearest(self.get_converted_partial())

    def get_position_of_record(self, record=None):
        """Return position of record in file or 0 (zero)."""
        if record is None:
            return 0
        key, value = record
        segment_number, record_number = divmod(value,
                                               SegmentSize.db_segment_size)
        # Define lambdas to handle presence or absence of partial key
        low = lambda rk, recordkey: rk < recordkey
        if not self.get_partial():
            high = lambda rk, recordkey: rk > recordkey
        else:
            high = lambda rk, partial: not rk.startswith(partial)
        # Get position of record relative to start point
        position = 0
        if not self.get_partial():
            r = self._cursor.first()
        else:
            r = self._cursor.set_range(
                self.get_converted_partial_with_wildcard())
        while r:
            if low(r[0].decode(), key):
                if len(r[1]) == LENGTH_SEGMENT_LIST_REFERENCE:
                    position += int.from_bytes(r[1][4:6], byteorder='big')
                elif len(r[1]) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    position += int.from_bytes(r[1][4:7], byteorder='big')
                else:
                    position += 1
            elif high(r[0].decode(), key):
                break
            else:
                sr = int.from_bytes(r[1][:4], byteorder='big')
                if sr < segment_number:
                    if len(r[1]) == LENGTH_SEGMENT_LIST_REFERENCE:
                        position += int.from_bytes(
                            r[1][4:6], byteorder='big')
                    elif len(r[1]) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                        position += int.from_bytes(
                            r[1][4:7], byteorder='big')
                    else:
                        position += 1
                elif sr > segment_number:
                    break
                else:
                    if len(r[1]) == LENGTH_SEGMENT_LIST_REFERENCE:
                        srn = int.from_bytes(r[1][6:], byteorder='big')
                        segment = SegmentList(
                            segment_number,
                            None,
                            records=self._segment_list.get(
                                srn, txn=self._transaction))
                    elif len(r[1]) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                        srn = int.from_bytes(r[1][7:], byteorder='big')
                        segment = SegmentBitarray(
                            segment_number,
                            None,
                            records=self._segment_bits.get(
                                srn, txn=self._transaction))
                    else:
                        segment = SegmentInt(
                            segment_number,
                            None,
                            records=r[1][4:])
                    position += segment.get_position_of_record_number(
                        record_number)
                    break
            r = self._cursor.next()
        return position

    def get_record_at_position(self, position=None):
        """Return record for positionth record in file or None."""
        if position is None:
            return None
        # Start at first or last record whichever is likely closer to position
        # and define lambdas to handle presence or absence of partial key.
        if not self.get_partial():
            get_partial = self.get_partial
        else:
            get_partial = self.get_converted_partial
        if position < 0:
            is_step_forward = False
            step = self._cursor.prev
            position = -1 - position
            if not self.get_partial():
                start = lambda partial: self._cursor.last()
            else:
                start = lambda partial: self._last_partial(partial)
        else:
            is_step_forward = True
            step = self._cursor.next
            if not self.get_partial():
                start = lambda partial: self._cursor.first()
            else:
                start = lambda partial: self._first_partial(partial)
        # Get record at position relative to start point
        count = 0
        r = start(get_partial())
        while r:
            if len(r[1]) == LENGTH_SEGMENT_LIST_REFERENCE:
                sc = int.from_bytes(r[1][4:6], byteorder='big')
            elif len(r[1]) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                sc = int.from_bytes(r[1][4:7], byteorder='big')
            else:
                sc = 1
            count += sc
            if count < position:
                r = step()
            else:
                count -= position
                if len(r[1]) == LENGTH_SEGMENT_LIST_REFERENCE:
                    srn = int.from_bytes(r[1][6:], byteorder='big')
                    segment = SegmentList(
                        int.from_bytes(r[1][:4], byteorder='big'),
                        None,
                        records=self._segment_list.get(srn,
                                                       txn=self._transaction))
                elif len(r[1]) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
                    srn = int.from_bytes(r[1][7:], byteorder='big')
                    segment = SegmentBitarray(
                        int.from_bytes(r[1][:4], byteorder='big'),
                        None,
                        records=self._segment_bits.get(srn,
                                                       txn=self._transaction))
                else:
                    segment = SegmentInt(
                        int.from_bytes(r[1][:4], byteorder='big'),
                        None,
                        records=r[1][4:])
                record_number = segment.get_record_number_at_position(
                    count, is_step_forward)
                if record_number is not None:
                    return r[0].decode(), record_number
                break
        return None

    def last(self):
        """Return last record taking partial key into account."""
        if self.get_partial() is None:
            try:
                k, v = self._last()
            except TypeError:
                return None
            return k.decode(), v
        elif self.get_partial() is False:
            return None
        else:
            c = list(self.get_partial())
            while True:
                try:
                    c[-1] = chr(ord(c[-1]) + 1)
                except ValueError:
                    c.pop()
                    if not len(c):
                        try:
                            k, v = self._cursor.last()
                        except TypeError:
                            return None
                        return k.decode(), v
                    continue
                self._set_range(''.join(c).encode())
                return self.prev()

    def nearest(self, key):
        """Return nearest record to key taking partial key into account."""
        try:
            k, v = self._set_range(key)
        except TypeError:
            return None
        return k.decode(), v

    def next(self):
        """Return next record taking partial key into account."""
        try:
            k, v = self._next()
        except TypeError:
            return None
        return k.decode(), v

    def prev(self):
        """Return previous record taking partial key into account."""
        try:
            k, v = self._prev()
        except TypeError:
            return None
        return k.decode(), v

    def setat(self, record):
        """Return current record after positioning cursor at record.

        Take partial key into account.
        
        Words used in bsddb3 (Python) to describe set and set_both say
        (key, value) is returned while Berkeley DB description seems to
        say that value is returned by the corresponding C functions.
        Do not know if there is a difference to go with the words but
        bsddb3 works as specified.

        """
        if self.get_partial() is False:
            return None
        key, value = record
        if self.get_partial() is not None:
            if not key.startswith(self.get_partial()):
                return None
        try:
            k, v = self._set_both(key.encode(), value)
        except TypeError:
            return None
        return k.decode(), v

    def set_partial_key(self, partial):
        """Set partial key."""
        self._partial = partial

    def _get_record(self, record):
        """Return record matching key or partial key or None if no match."""
        raise DBapiError('_get_record not implemented')

    def set_current_segment(self, key, reference):
        """Return a SegmentBitarray, SegmentInt, or SegmentList instance."""
        segment_number = int.from_bytes(reference[:4], byteorder='big')
        if len(reference) == LENGTH_SEGMENT_LIST_REFERENCE:
            if self._current_segment_number == segment_number:
                #return self._current_segment
                if key == self._current_segment._key:
                    return self._current_segment
            record_number = int.from_bytes(reference[6:], byteorder='big')
            segment = SegmentList(
                segment_number,
                key,
                records=self._segment_list.get(record_number,
                                               txn=self._transaction))
        elif len(reference) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
            if self._current_segment_number == segment_number:
                #return self._current_segment
                if key == self._current_segment._key:
                    return self._current_segment
            record_number = int.from_bytes(reference[7:], byteorder='big')
            segment = SegmentBitarray(
                segment_number,
                key,
                records=self._segment_bits.get(record_number,
                                               txn=self._transaction))
        else:
            segment = SegmentInt(segment_number, key, records=reference[4:])
        self._current_segment = segment
        self._current_segment_number = segment_number
        return segment

    def _first(self):
        """Return first record taking partial key into account."""
        record = self._cursor.first()
        if record is None:
            return None
        return self.set_current_segment(*record).first()

    def _last(self):
        """Return last record taking partial key into account."""
        record = self._cursor.last()
        if record is None:
            return None
        return self.set_current_segment(*record).last()

    def _next(self):
        """Return next record taking partial key into account."""
        if self._current_segment is None:
            return self._first()
        record = self._current_segment.next()
        if record is None:
            record = self._cursor.next()
            if record is None:
                return None
            if self.get_partial() is not None:
                if not record[0].startswith(self.get_converted_partial()):
                    return None
            return self.set_current_segment(*record).first()
        else:
            return record

    def _prev(self):
        """Return previous record taking partial key into account."""
        if self._current_segment is None:
            return self._last()
        record = self._current_segment.prev()
        if record is None:
            record = self._cursor.prev()
            if record is None:
                return None
            if self.get_partial() is not None:
                if not record[0].startswith(self.get_converted_partial()):
                    return None
            return self.set_current_segment(*record).last()
        else:
            return record

    def _set_both(self, key, value):
        """Return current record after positioning cursor at (key, value)."""
        segment, record_number = divmod(value, SegmentSize.db_segment_size)
        # Find the segment reference in secondary database
        cursor = self._dbset.db_object.cursor(txn=self._transaction)
        try:
            record = cursor.set_range(key)
            while record:
                if record[0] != key:
                    return None
                segment_number = int.from_bytes(record[1][:4], byteorder='big')
                if segment_number > segment:
                    return None
                if segment_number == segment:
                    break
                record = cursor.next()
            else:
                return None
        finally:
            cursor.close()
        # Check if record number is in segment
        ref = record[1]
        if len(ref) == LENGTH_SEGMENT_LIST_REFERENCE:
            srn = int.from_bytes(ref[6:], byteorder='big')
            segment = SegmentList(
                segment_number,
                key,
                records=self._segment_list.get(srn, txn=self._transaction))
        elif len(ref) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
            srn = int.from_bytes(ref[7:], byteorder='big')
            segment = SegmentBitarray(
                segment_number,
                key,
                records=self._segment_bits.get(srn, txn=self._transaction))
        else:
            segment = SegmentInt(segment_number, key, records=ref[4:])
        if segment.setat(value) is None:
            return None
        # Move self._cursor to new segment reference
        record = self._cursor.set_both(key, ref)
        if record is None:
            return None
        self._current_segment = segment
        self._current_segment_number = segment_number
        # Return the "secondary database record".
        return key, value

    def _set_range(self, key):
        """Return current record after positioning cursor at nearest to key."""
        # Move self._cursor to nearest segment reference
        record = self._cursor.set_range(key)
        if record is None:
            self._current_segment = None
            self._current_segment_number = None
            self._current_record_number_in_segment = None
            return None
        segment_number = int.from_bytes(record[1][:4], byteorder='big')
        # Set up the segment instance and position at first record
        ref = record[1]
        if len(ref) == LENGTH_SEGMENT_LIST_REFERENCE:
            srn = int.from_bytes(ref[6:], byteorder='big')
            segment = SegmentList(
                segment_number,
                record[0],
                records=self._segment_list.get(srn, txn=self._transaction))
        elif len(ref) == LENGTH_SEGMENT_BITARRAY_REFERENCE:
            srn = int.from_bytes(ref[7:], byteorder='big')
            segment = SegmentBitarray(
                segment_number,
                record[0],
                records=self._segment_bits.get(srn, txn=self._transaction))
        else:
            segment = SegmentInt(segment_number, record[0], records=ref[4:])
        self._current_segment = segment
        self._current_segment_number = segment_number
        # Return the "secondary database record".
        return segment.first()

    def _first_partial(self, partial):
        """Place cursor at first record with partial key and return record."""
        r = self._cursor.set_range(partial)
        if r is None:
            return None
        if not r[0].startswith(partial):
            return None
        return r

    def _last_partial(self, partial):
        """Place cursor at last record with partial key and return record."""
        # This code is wrong but the only place using it does not work yet.
        # Should it be doing _cursot.last() _cursor.set_range() _cursor.prev()?
        k = list(partial)
        while True:
            try:
                k[-1] = chr(ord(k[-1]) + 1)
            except ValueError:
                k.pop()
                if not len(k):
                    return self._last()
                continue
            self._set_range(''.join(k).encode())
            return self._prev()

    def refresh_recordset(self, instance=None):
        """Refresh records for datagrid access after database update.

        The bitmap for the record set may not match the existence bitmap.

        """
        # See set_selection() hack in chesstab subclasses of DataGrid.
        # It seems not needed by this class.
        
        #raise DBapiError('refresh_recordset not implemented')


class RecordsetCursorDB(RecordsetCursor):
    
    """Provide a bsddb3 style cursor for a recordset of arbitrary records.

    The cursor does not support partial keys because the records in the
    recordset do not have an implied order (apart from the accidential order
    of existence on the database).

    """

    def __init__(self, recordset, database_instance):
        """Define a cursor to access recordset in the order given in keymap."""
        super().__init__(recordset)
        self._transaction = database_instance.get_transaction

    @property
    def transaction(self):
        """Return the active transaction or None."""
        return self._transaction()

    # The _get_record hack in sqlite3bitdatasource.py becomes the correct way
    # to do this because the record has bsddb-specific decoding needs.
    def _get_record(self, record_number, use_cache=False):
        """Return record_number from database using cache if requested."""
        dbset = self._dbset
        if use_cache:
            record = dbset.record_cache.get(record_number)
            if record is not None:
                return record # maybe (record_number, record)
        segment, recnum = divmod(record_number, SegmentSize.db_segment_size)
        if segment not in dbset.rs_segments:
            return # maybe raise
        if recnum not in dbset.rs_segments[segment]:
            return # maybe raise
        try:
            record = dbset._database.get(record_number,
                                         txn=self.transaction).decode()
        except AttributeError:
            # Assume get() returned None.
            record = None
        # maybe raise if record is None (if not, None should go on cache)
        if use_cache:
            dbset.record_cache[record_number] = record
            dbset.record.deque.append(record_number)
        return (record_number, record)

            
# Maybe this and DBFile should have same superclass for open_root and close.
class DBSegment(object):
    
    """Define a DB file to store inverted record number list representations.
    """

    def __init__(self, segment_type, dbfile_instance):
        """Define segment file for segment type for dbfile_instance.
        
        segment_type=representation style for inverted list of record numbers
        dbfile_instance=dbfile instance for which segment type file is created

        For large numbers of records fixed sized bit maps representing record
        numbers relative to the base for the segment.  Otherwise a list of
        record numbers relative to the base.
        
        """
        super(DBSegment, self).__init__()
        self._seg_dbfile = (SUBFILE_DELIMITER * 2).join(
            (dbfile_instance.get_database_file(), segment_type))
        self._seg_object = None
        self._transaction = dbfile_instance._transaction

    @property
    def transaction(self):
        """Return the active transaction or None."""
        return self._transaction()

    def close(self):
        """Close inverted index DB."""
        if self._seg_object is not None:
            self._seg_object.close()
            self._seg_object = None

    def get(self, key):
        """Get a segment record from the database."""
        # Exists to match the sqlite3 interface - may get used as wrapper for
        # DB get method eventually.  Sqlite3Segment version hides boilerplate
        # SQL statements.
        raise DBapiError('DBSegment.get not implemented')

    def delete(self, key):
        """Delete a segment record from the database."""
        # Exists to match the sqlite3 interface - may get used as wrapper for
        # DB delete method eventually.  Sqlite3Segment version hides boilerplate
        # SQL statements.
        raise DBapiError('DBSegment.delete not implemented')

    def put(self, key, value):
        """Put a segment record on the database, either replace or insert."""
        # Exists to match the sqlite3 interface - may get used as wrapper for
        # DB put method eventually.  Sqlite3Segment version hides boilerplate
        # SQL statements.
        raise DBapiError('DBSegment.put not implemented')

    def append(self, value):
        """Append a segment record on the database using a new key."""
        # Exists to match the sqlite3 interface - may get used as wrapper for
        # DB append method eventually.  Sqlite3Segment version hides boilerplate
        # SQL statements.
        raise DBapiError('DBSegment.append not implemented')

            
class DBExistenceBitMap(DBSegment):
    
    """DB file to store record number existence bit map keyed by segment number.
    """

    def __init__(self, *a):
        """Define dbfile in environment for segment record bit maps.
        
        dbfile=file name relative to environment home directory
        
        """
        super(DBExistenceBitMap, self).__init__('exist', *a)
        self._segment_count = None

    @property
    def segment_count(self):
        """Return number of records in segment."""
        return self._segment_count

    @segment_count.setter
    def segment_count(self, segment_number):
        """Set segment count from 0-based segment_number if greater."""
        if segment_number > self._segment_count:
            self._segment_count = segment_number + 1
    
    def open_root(self, dbenv):
        """Create inverted index DB in dbenv."""
        try:
            self._seg_object = DB(dbenv)
        except:
            raise
        try:
            self._seg_object.set_re_pad(0)
            self._seg_object.set_re_len(SegmentSize.db_segment_size_bytes)
            self._seg_object.open(
                self._seg_dbfile,
                self._seg_dbfile,
                DB_RECNO,
                DB_CREATE,
                txn=self.transaction)
            self._segment_count = self._seg_object.stat(
                flags=DB_FAST_STAT, txn=self.transaction)['ndata']
        except:
            self._seg_object = None
            raise

            
class DBSegmentList(DBSegment):
    
    """DB file to store inverted record number list with arbitrary key.
    """

    def __init__(self, *a):
        """Define dbfile in environment for segment record number lists.
        
        dbfile=file name relative to environment home directory
        
        """
        super(DBSegmentList, self).__init__('list', *a)

    def open_root(self, dbenv):
        """Create inverted index DB in dbenv."""
        try:
            self._seg_object = DB(dbenv)
        except:
            raise
        try:
            self._seg_object.open(
                self._seg_dbfile,
                self._seg_dbfile,
                DB_RECNO,
                DB_CREATE,
                txn=self.transaction)
        except:
            self._seg_object = None
            raise

            
class DBSegmentBitMap(DBSegment):
    
    """DB file to store inverted record number bit map with arbitrary key.
    """

    def __init__(self, *a):
        """Define dbfile in environment for segment record bit maps.
        
        dbfile=file name relative to environment home directory
        
        """
        super(DBSegmentBitMap, self).__init__('bits', *a)

    def open_root(self, dbenv):
        """Create inverted index DB in dbenv."""
        try:
            self._seg_object = DB(dbenv)
        except:
            raise
        try:
            self._seg_object.set_re_pad(0)
            self._seg_object.set_re_len(SegmentSize.db_segment_size_bytes)
            self._seg_object.open(
                self._seg_dbfile,
                self._seg_dbfile,
                DB_RECNO,
                DB_CREATE,
                txn=self.transaction)
        except:
            self._seg_object = None
            raise

            
class DBControlFile(object):
    
    """Define a DB file for control information about the database DB files.

    The method names used in DBPrimaryFile and superclasses are used
    where possible.  But this file is primary DB_BTREE NODUP.  It is used by
    all FileControl instances.
    
    """

    def __init__(self, database_instance, control_file='control'):
        """File control database for all DB files."""
        super(DBControlFile, self).__init__()
        self._control_file = ''.join((SUBFILE_DELIMITER * 3, control_file))
        self._control_object = None
        self._transaction = database_instance.get_transaction

    @property
    def transaction(self):
        """Return the active transaction or None."""
        return self._transaction()

    def open_root(self, dbenv):
        """Create file control database in environment."""
        try:
            self._control_object = DB(dbenv)
        except:
            raise
        try:
            self._control_object.set_flags(DB_DUPSORT)
            self._control_object.open(
                self._control_file,
                self._control_file,
                DB_BTREE,
                DB_CREATE,
                txn=self.transaction)
        except:
            self._control_object = None
            raise

    def close(self):
        """Close file control database."""
        if self._control_object is not None:
            self._control_object.close()
            self._control_object = None

    def get_control_database(self):
        """Return the database containing file control records."""
        return self._control_object

    @property
    def control_file(self):
        """Return the name which is both primary column and table name."""
        return self._control_file


class FileControl(object):
    
    """Freed resource data for a segmented file.
    """

    def __init__(self, dbfile):
        """Define the file control data."""
        super(FileControl, self).__init__()

        # Primary or Secondary file instance whose segment reuse is handled
        self._dbfile = dbfile


class FileControlPrimary(FileControl):
    
    """Freed resource data for a segmented file.

    Introduced to keep track of pages on the existence bit map that contain
    freed record numbers that can be reused.

    The list of existence bit map page numbers containing freed record numbers
    is cached.

    """

    def __init__(self, *args):
        """Define the file control data for primary files."""
        super(FileControlPrimary, self).__init__(*args)
        self._freed_record_number_pages = None

    @property
    def freed_record_number_pages(self):
        """Return existence bit map record numbers available for re-use."""
        if self._freed_record_number_pages is None:
            return None
        return bool(self._freed_record_number_pages)

    def note_freed_record_number(self, record_number):
        """Adjust segment of high and low freed record numbers."""
        self.note_freed_record_number_segment(
            *divmod(record_number, SegmentSize.db_segment_size))

    def note_freed_record_number_segment(
        self, segment, record_number_in_segment):
        """Adjust segment of high and low freed record numbers."""
        if self._freed_record_number_pages is None:
            self._freed_record_number_pages = []
            cursor = self._dbfile.get_control_database(
                ).cursor(txn=self._dbfile.transaction)
            try:
                record = cursor.set(b'E')
                while record:
                    self._freed_record_number_pages.append(
                        int.from_bytes(record[1], byteorder='big'))
                    record = cursor.next_dup()
            finally:
                cursor.close()
        insert = bisect.bisect_left(self._freed_record_number_pages, segment)
        # Should be:
        # if insert <= len(self._freed_record_number_pages):
        # Leave as it is until dbapi tests give same results as sqlite3 tests,
        # which have the same problem.
        if self._freed_record_number_pages:
            if insert < len(self._freed_record_number_pages):
                if self._freed_record_number_pages[insert] == segment:
                    return
        self._freed_record_number_pages.insert(insert, segment)
        self._dbfile.get_control_database().put(
            b'E',
            segment.to_bytes(1 + segment.bit_length() // 8, byteorder='big'),
            flags=DB_NODUPDATA,
            txn=self._dbfile.transaction)

    def get_lowest_freed_record_number(self):
        """Return low record number in segments with freed record numbers."""
        if self._freed_record_number_pages is None:
            self._freed_record_number_pages = []
            cursor = self._dbfile.get_control_database(
                ).cursor(txn=self._dbfile.transaction)
            try:
                record = cursor.set(b'E')
                while record:
                    self._freed_record_number_pages.append(
                        int.from_bytes(record[1], byteorder='big'))
                    record = cursor.next_dup()
            finally:
                cursor.close()
        while len(self._freed_record_number_pages):
            s = self._freed_record_number_pages[0]
            lfrns = self._read_exists_segment(s)
            if lfrns is None:
                # Do not reuse record number on segment of high record number
                return 0
            try:
                first_zero_bit = lfrns.index(False, 0 if s else 1)
            except ValueError:
                cursor = self._dbfile.get_control_database(
                    ).cursor(txn=self._dbfile.transaction)
                try:
                    if cursor.set_both(
                        b'E',
                        s.to_bytes(1 + s.bit_length() // 8, byteorder='big')):
                        cursor.delete()
                    else:
                        raise
                finally:
                    cursor.close()
                del self._freed_record_number_pages[0]
                continue
            return s * SegmentSize.db_segment_size + first_zero_bit
        else:
            return 0 # record number when inserting into RECNO database

    def _read_exists_segment(self, segment_number):
        """Return existence bit map for segment_number if not high segment."""
        # record keys are 1-based but segment_numbers are 0-based
        if segment_number < self._dbfile.get_existence_bits().segment_count - 1:
            ebm = Bitarray()
            ebm.frombytes(
                self._dbfile.get_existence_bits_database(
                    ).get(segment_number + 1,
                          txn=self._dbfile.transaction))
            return ebm
        return None


class FileControlSecondary(FileControl):
    
    """Freed resource data for a segmented file.

    Introduced to keep track of freed pages on the record number bit map and
    record number list files for each segmented file.

    These pages can be freed when records are deleted or when an inverted list
    for a key is moved between bit map and list representations.

    (The term page is used because record number n on --list and --bits files
    is not in general referring to segment number n on the segmented file.)

    """

    def __init__(self, *args):
        """Define the file control data for secondary files."""
        super(FileControlSecondary, self).__init__(*args)
        self._freed_list_pages = None
        self._freed_bits_pages = None

    @property
    def freed_list_pages(self):
        """List pages available for re-use."""
        return self._freed_list_pages

    @property
    def freed_bits_pages(self):
        """Bit Map pages available for re-use."""
        return self._freed_bits_pages

    def note_freed_bits_page(self, page_number):
        """Add page_number to freed bits pages."""
        self._put_bits_page_number(page_number)

    def note_freed_list_page(self, page_number):
        """Add page_number to freed list pages."""
        self._put_list_page_number(page_number)

    def get_freed_bits_page(self):
        """Return low page from freed bits pages."""
        if self._freed_bits_pages is False:
            return 0 # record number when inserting into RECNO database
        return self._get_bits_page_number()

    def get_freed_list_page(self):
        """Return low page from freed list pages."""
        if self._freed_list_pages is False:
            return 0 # record number when inserting into RECNO database
        return self._get_list_page_number()

    def _put_bits_page_number(self, page):
        """Put page on freed bits page record."""
        try:
            self._dbfile.get_control_database().put(
                b'B',
                page.to_bytes(1 + page.bit_length() // 8, byteorder='big'),
                flags=DB_NODUPDATA,
                txn=self._dbfile.transaction)
        except DBKeyExistError:
            # Assume callers do not check if page is already marked free.
            pass
        self._freed_bits_pages = True

    def _put_list_page_number(self, page):
        """Put page on freed list page record."""
        try:
            self._dbfile.get_control_database().put(
                b'L',
                page.to_bytes(1 + page.bit_length() // 8, byteorder='big'),
                flags=DB_NODUPDATA,
                txn=self._dbfile.transaction)
        except DBKeyExistError:
            # Assume callers do not check if page is already marked free.
            pass
        self._freed_list_pages = True

    def _get_bits_page_number(self):
        """Pop low page from freed bits page record."""
        cursor = self._dbfile.get_control_database(
            ).cursor(txn=self._dbfile.transaction)
        try:
            record = cursor.set(b'B')
            if record:
                cursor.delete()
                self._freed_bits_pages = False
                return int.from_bytes(record[1], byteorder='big')
        finally:
            cursor.close()
        self._freed_bits_pages = False
        return 0 # record number when inserting into RECNO database

    def _get_list_page_number(self):
        """Pop low page from freed list page record."""
        cursor = self._dbfile.get_control_database(
            ).cursor(txn=self._dbfile.transaction)
        try:
            record = cursor.set(b'L')
            if record:
                cursor.delete()
                self._freed_list_pages = False
                return int.from_bytes(record[1], byteorder='big')
        finally:
            cursor.close()
        self._freed_list_pages = False
        return 0 # record number when inserting into RECNO database
