# dbapi_test.py
# Copyright 2012 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""dbapi tests"""

import unittest
import os
import shutil
import subprocess
import ast

from bsddb3.db import (
    DB_CREATE,
    DB_INIT_MPOOL,
    DB_INIT_LOCK,
    DB_INIT_LOG,
    DB_INIT_TXN,
    DB, DB_RDONLY,
    DBEnv,
    DB_RECNO,
    DB_BTREE,
    )

from ..api.test._filespec import (
    samplefilespec,
    SampleValue,
    SampleRecord,
    GAMES_FILE_DEF,
    GAME_FIELD_DEF,
    EVENT_FIELD_DEF,
    SITE_FIELD_DEF,
    DATE_FIELD_DEF,
    ROUND_FIELD_DEF,
    WHITE_FIELD_DEF,
    BLACK_FIELD_DEF,
    RESULT_FIELD_DEF,
    SEVEN_TAG_ROSTER,
    PRIMARY,
    SECONDARY,
    )
from .. import dbapi
from ..api import database, filespec, record, recordset, constants
from ..api.segmentsize import SegmentSize
from . import database_interface


def base(folder, environment=None):
    return dbapi.DBbase(
        dbapi.DBPrimary,
        dbapi.DBSecondary,
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'dbapi_tests', folder)),
        environment)


def api(folder):
    return dbapi.DBapi(
        samplefilespec(),
        os.path.expanduser(
            os.path.join('~', 'dbapi_tests', folder)),
        {'flags': (DB_CREATE |
                   DB_INIT_MPOOL |
                   DB_INIT_LOCK |
                   DB_INIT_LOG |
                   DB_INIT_TXN)})


def _delete_folder(folder):
    shutil.rmtree(
        os.path.expanduser(os.path.join('~', 'dbapi_tests', folder)),
        ignore_errors=True)


class SampleRecordAssumptions(unittest.TestCase):

    def setUp(self):
        self.sr = SampleRecord()
        self.sv = SampleValue()

    def tearDown(self):
        self.sr = None
        self.sv = None

    def test____assumptions_01(self):
        self.assertEqual(len(self.sr.__dict__), 8)
        self.assertIsInstance(self.sr.key, record.KeyData)
        self.assertIsInstance(self.sr.value, record.Value)
        self.assertIsInstance(self.sr.value, SampleValue)
        self.assertEqual(self.sr.record, None)
        self.assertEqual(self.sr.database, None)
        self.assertEqual(self.sr.dbname, None)
        self.assertEqual(self.sr.srkey, None)
        self.assertEqual(self.sr.srvalue, None)
        self.assertEqual(self.sr.srindex, None)
        self.assertEqual(len(self.sv.__dict__), 0)
        self.assertIsInstance(self.sv, record.Value)
        self.assertIsInstance(self.sv, SampleValue)

    def test____assumptions_02(self):
        self.assertRaisesRegex(
            ValueError,
            "malformed node or string: \[\]",
            self.sr.load_record,
            *((0, []),))
        self.assertRaisesRegex(
            TypeError,
            "__dict__ must be set to a dictionary, not a 'list'",
            self.sr.load_record,
            *((0, '[]'),))
        self.sr.load_record((0, '{}'))
        self.assertEqual(len(self.sr.key.__dict__), 1)
        self.assertEqual(self.sr.key.recno, 0)
        self.assertEqual(len(self.sr.value.__dict__), 0)

    def test____assumptions_03(self):
        self.sr.load_record((0, repr({})))
        self.assertEqual(len(self.sr.value.__dict__), 0)

    def test____assumptions_04(self):
        # This happens, but will not work in database.
        self.sr.load_record(('0', '{}'))
        self.assertEqual(self.sr.key.recno, b'0')

    def test____assumptions_05(self):
        # This happens, but will not work in database.
        self.sr.load_record(([], '{}'))
        self.assertEqual(self.sr.key.recno, [])

    def test____assumptions_06(self):
        self.assertEqual(SEVEN_TAG_ROSTER,
                         (EVENT_FIELD_DEF,
                          SITE_FIELD_DEF,
                          DATE_FIELD_DEF,
                          ROUND_FIELD_DEF,
                          WHITE_FIELD_DEF,
                          BLACK_FIELD_DEF,
                          RESULT_FIELD_DEF,
                          ))
        value = {GAME_FIELD_DEF:'game score',
                 WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 EVENT_FIELD_DEF:'Match',
                 ROUND_FIELD_DEF:'1',
                 DATE_FIELD_DEF:'today',
                 RESULT_FIELD_DEF:'1-0',
                 SITE_FIELD_DEF:'club venue',
                 'attribute':None,
                 }
        self.sr.load_record((0, repr(value)))
        self.assertEqual(self.sr.value.__dict__, value)
        pv = self.sr.value.pack()
        # pv[0] is not necessarely equal to repr(value)
        self.assertEqual(ast.literal_eval(pv[0]),
                         ast.literal_eval(repr(value)))
        self.assertEqual(pv[1],
                         {WHITE_FIELD_DEF:['A Jones'],
                          BLACK_FIELD_DEF:['B Smith'],
                          EVENT_FIELD_DEF:['Match'],
                          ROUND_FIELD_DEF:['1'],
                          DATE_FIELD_DEF:['today'],
                          RESULT_FIELD_DEF:['1-0'],
                          SITE_FIELD_DEF:['club venue'],
                          })

    def test____assumptions_07(self):
        value = {WHITE_FIELD_DEF:'A Jones',
                 BLACK_FIELD_DEF:'B Smith',
                 }
        self.sr.load_record((0, repr(value)))
        pv = self.sr.value.pack()
        self.assertEqual(pv[1],
                         {WHITE_FIELD_DEF:['A Jones'],
                          BLACK_FIELD_DEF:['B Smith'],
                          })


class _DatabaseEncoders(unittest.TestCase):

    def setUp(self):
        self.sde = dbapi._DatabaseEncoders()

    def tearDown(self):
        pass

    def test____assumptions(self):
        self.assertEqual(repr(57), '57')

    def test_encode_record_number_01(self):
        self.assertEqual(self.sde.encode_record_number(57), b'57')

    def test_decode_record_number_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'int' object has no attribute 'decode'",
            self.sde.decode_record_number,
            57)
        self.assertEqual(self.sde.decode_record_number(b'57'), 57)

    def test_encode_record_selector(self):
        self.assertEqual(self.sde.encode_record_selector('57'), b'57')


class DBbase_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        p = os.path.abspath('')
        ba = dbapi.DBbase(None, None, {}, '', {})
        self.assertEqual(ba._dbdef, {})
        self.assertEqual(ba._dbenv, None)
        self.assertEqual(ba._dbspec, {})
        self.assertEqual(ba._databasefolder, p)
        self.assertEqual(ba._home, p)
        self.assertEqual(ba._control, None)
        self.assertEqual(ba._dbtxn, None)
        self.assertEqual(ba._DBenvironment, {})
        self.assertEqual(len(ba.__dict__), 8)
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 5 required positional arguments: ",
                "'primary_class', 'secondary_class', 'database_specification'",
                ", 'databasefolder', and 'DBenvironment'",
                )),
            dbapi.DBbase,
            )
        self.assertRaisesRegex(
            database.DatabaseError,
            "".join((
                "Database folder name None is not valid",
                )),
            dbapi.DBbase,
            *(None, None, None, None, None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Database specification must be a dictionary",
            dbapi.DBbase,
            *(None, None, None, '', None))
        self.assertIsInstance(
            dbapi.DBbase(None, None, {}, '', None),
            dbapi.DBbase)
        self.assertRaisesRegex(
            database.DatabaseError,
            "Specification for 't' must be a dictionary",
            dbapi.DBbase,
            *(None, None, dict(t=None), '', None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Specification for t must contain a primary name",
            dbapi.DBbase,
            *(None, None, dict(t={}), '', None))
        class C:
            def __init__(self, *a, **b):
                pass
        self.assertRaisesRegex(
            database.DatabaseError,
            "Specification for t must contain a primary name",
            dbapi.DBbase,
            *(C, None, dict(t={}), '', None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Primary definition key for t must be in fields definition",
            dbapi.DBbase,
            *(C, None, dict(t=dict(primary='key')), '', None))
        self.assertRaisesRegex(
            database.DatabaseError,
            ''.join((
                "Primary definition key for t must not be in ",
                "secondary definition \(ignoring case\)")),
            dbapi.DBbase,
            *(C,
              None,
              dict(t=dict(primary='key',
                          secondary={'key':None})),
              '',
              None))
        self.assertRaisesRegex(
            database.DatabaseError,
            ''.join((
                "Primary definition key for t must not be in ",
                "secondary definition \(ignoring case\)")),
            dbapi.DBbase,
            *(C,
              None,
              dict(t=dict(primary='key',
                          secondary={'key':'key'})),
              '',
              None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Primary definition key for t must be in fields definition",
            dbapi.DBbase,
            *(C,
              None,
              dict(t=dict(primary='key',
                          secondary={'lock':'lock'})),
              '',
              None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Primary definition key for t must be in fields definition",
            dbapi.DBbase,
            *(C,
              C,
              dict(t=dict(primary='key',
                          secondary={'lock':'lock'})),
              '',
              None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Primary definition key for t must be in fields definition",
            dbapi.DBbase,
            *(C,
              C,
              dict(t=dict(primary='key',
                          secondary={'lock':'locks'})),
              '',
              None))

    def test___init___02(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def set_control_database(self, *a):
                pass
            def set_primary_database(self, *a):
                pass
        api = dbapi.DBbase(C,
                           C,
                           filespec.FileSpec(
                               t=dict(primary='key',
                                      secondary={'lock':'locks'},
                                      fields={'key':{}, 'locks':{}})),
                           '',
                           None)
        self.assertEqual(len(api._dbdef), 1)
        self.assertIsInstance(api._dbdef['t'].primary, C)
        self.assertIsInstance(api._dbdef['t'].secondary['locks'], C)
        self.assertEqual(len(api.__dict__), 8)


class DBbase_backout(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def set_control_database(self, *a):
                pass
            def set_primary_database(self, *a):
                pass
        api = dbapi.DBbase(C,
                           C,
                           filespec.FileSpec(
                               t=dict(primary='key',
                                      secondary={'lock':'locks'},
                                      fields={'key':{}, 'locks':{}})),
                           '',
                           None)
        self.assertEqual(api.backout(), None)


class DBbase_close_contexts(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def set_control_database(self, *a):
                pass
            def set_primary_database(self, *a):
                pass
        api = dbapi.DBbase(C,
                           C,
                           filespec.FileSpec(
                               t=dict(primary='key',
                                      secondary={'lock':'locks'},
                                      fields={'key':{}, 'locks':{}})),
                           '',
                           None)
        self.assertEqual(api.close_contexts(None), None)


class DBbase_close_database(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def close(self):
                pass
            def set_control_database(self, *a):
                pass
            def set_primary_database(self, *a):
                pass
        api = dbapi.DBbase(C,
                           C,
                           filespec.FileSpec(
                               t=dict(primary='key',
                                      secondary={'lock':'locks'},
                                      fields={'key':{}, 'locks':{}})),
                           '',
                           None)
        self.assertEqual(api.close_database(), None)


class DBbase_commit(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def set_control_database(self, *a):
                pass
            def set_primary_database(self, *a):
                pass
        api = dbapi.DBbase(C,
                           C,
                           filespec.FileSpec(
                               t=dict(primary='key',
                                      secondary={'lock':'locks'},
                                      fields={'key':{}, 'locks':{}})),
                           '',
                           None)
        self.assertEqual(api.commit(), None)


class DBbase_db_compatibility_hack(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        class C:
            def __init__(self, *a, **b):
                pass
            def set_control_database(self, *a):
                pass
            def set_primary_database(self, *a):
                pass
        api = dbapi.DBbase(C,
                           C,
                           filespec.FileSpec(
                               t=dict(primary='key',
                                      secondary={'lock':'locks'},
                                      fields={'key':{}, 'locks':{}})),
                           '',
                           None)
        self.assertEqual(api.db_compatibility_hack('value', 'key'), 'value')


class DBbase_do_deferred_updates(unittest.TestCase):

    def setUp(self):
        self.api = base('do_deferred_updates')

    def tearDown(self):
        _delete_folder('do_deferred_updates')

    def test_do_deferred_updates_01(self):
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "'pyscript' is not an existing file",
            self.api.do_deferred_updates,
            *('pyscript', 'filepath'))
        self.assertRaisesRegex(
            dbapi.DBapiError,
            ''.join(("'/home/roger/dbapi_tests/do_deferred_updates' is ",
                     "not an existing file",
                     )),
            self.api.do_deferred_updates,
            *(self.api._home, 'filepath'))
        self.assertRaisesRegex(
            dbapi.DBapiError,
            ''.join(("'/home/roger/dbapi_tests/do_deferred_updates' is ",
                     "not an existing file",
                     )),
            self.api.do_deferred_updates,
            *(self.api._home, (self.api._home, 'filepath')))
        script = os.path.join(self.api._home, 'script')
        os.makedirs(self.api._home)
        open(script, 'w').close()
        sp = self.api.do_deferred_updates(script, self.api._home)
        self.assertIsInstance(sp, subprocess.Popen)
        sp.wait()


class DBbase_exists(unittest.TestCase):

    def setUp(self):
        self.api = base('exists')

    def tearDown(self):
        _delete_folder('exists')

    def test_exists_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.exists,
            *(None, None))
        self.assertEqual(self.api.exists('games', None), False)
        self.assertEqual(self.api.exists('games', 'Site'), True)


class DBbase_files_exist(unittest.TestCase):

    def setUp(self):
        self.api = base('files_exist')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for c in self.api._dbdef.values():
            c.open_root = lambda a : None
            c.close = lambda : None

    def tearDown(self):
        _delete_folder('files_exist')

    def test_exists_01(self):
        os.makedirs(self.api._home)
        self.assertEqual(self.api.files_exist(), None)
        self.api.open_context()
        #self.assertEqual(self.api.files_exist(), True)
        self.api.close_database()
        #self.assertEqual(self.api.files_exist(), True)


class DBbase_database_cursor(unittest.TestCase):

    def setUp(self):
        self.api = base('database_cursor')

    def tearDown(self):
        _delete_folder('database_cursor')

    def test_database_cursor_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.database_cursor,
            *(None, None))
        self.assertIsInstance(
            self.api.database_cursor('games', 'Site'),
            dbapi.CursorDBSecondary)
        self.assertIsInstance(
            self.api.database_cursor('games', 'games'),
            dbapi.CursorDBPrimary)


class DBbase_repair_cursor(unittest.TestCase):

    def setUp(self):
        self.api = base('database_cursor')

    def tearDown(self):
        _delete_folder('database_cursor')

    def test_repair_cursor_01(self):
        class C:
            pass
        c = C()
        self.assertIs(self.api.repair_cursor(c), c)


class DBbase_get_database_folder(unittest.TestCase):

    def setUp(self):
        self.api = base('get_database_folder')

    def tearDown(self):
        _delete_folder('get_database_folder')

    def test_get_database_folder_01(self):
        self.assertEqual(
            self.api.get_database_folder(),
            os.path.expanduser(
                os.path.join('~', 'dbapi_tests', 'get_database_folder')))


class DBbase_get_database(unittest.TestCase):

    def setUp(self):
        self.api = base('get_database')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}

    def tearDown(self):
        _delete_folder('get_database')

    def test_get_database_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database,
            *(None, None))
        self.assertEqual(self.api.get_database('games', 'Site'), None)
        self.assertEqual(self.api.get_database('games', 'games'), None)
        self.api.open_context()
        self.assertEqual(
            self.api.get_database('games', 'Site').__class__,
            DB().__class__)
        self.assertEqual(
            self.api.get_database('games', 'games').__class__,
            DB().__class__)


class DBbase_get_database_instance(unittest.TestCase):

    def setUp(self):
        self.api = base('get_database_instance')

    def tearDown(self):
        _delete_folder('get_database_instance')

    def test_get_database_instance_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_database_instance,
            *(None, None))
        self.assertIsInstance(
            self.api.get_database_instance('games', 'Site'),
            dbapi.DBSecondary)
        self.assertIsInstance(
            self.api.get_database_instance('games', 'games'),
            dbapi.DBPrimary)


class DBbase_get_first_primary_key_for_index_key(unittest.TestCase):
    # Test results are suspicious but may be correct.

    def setUp(self):
        def f(a):
            raise dbapi.DBapiError(
                "get_first_primary_key_for_index_key for primary table")
        self.api = base('get_first_primary_key_for_index_key')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for v in self.api._dbdef.values():
            v.get_first_primary_key_for_index_key = f
            v.open_root = lambda a : None

    def tearDown(self):
        _delete_folder('get_first_primary_key_for_index_key')

    def test_get_first_primary_key_for_index_key_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_first_primary_key_for_index_key,
            *(None, None, None))
        #self.assertRaisesRegex(
        #    AttributeError,
        #    ''.join((
        #        "'NoneType' object has no attribute ",
        #        "'execute'")),
        #    self.api.get_first_primary_key_for_index_key,
        #    *('games', 'Site', 'k'))
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "get_first_primary_key_for_index_key for primary index",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))
        self.api.open_context()
        #self.assertEqual(
        #    self.api.get_first_primary_key_for_index_key('games', 'games', 'k'),
        #    None)
        self.assertRaisesRegex(
            dbapi.DBapiError,
            "get_first_primary_key_for_index_key for primary index",
            self.api.get_first_primary_key_for_index_key,
            *('games', 'games', 'k'))


class DBbase_get_primary_record(unittest.TestCase):

    def setUp(self):
        self.api = base('get_primary_record')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}

    def tearDown(self):
        _delete_folder('get_primary_record')

    def test_get_primary_record_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.get_primary_record,
            *(None, None))
        self.api.open_context()
        self.assertEqual(self.api.get_primary_record('games', 1), None)


class DBbase_is_primary(unittest.TestCase):

    def setUp(self):
        self.api = base('is_primary')

    def tearDown(self):
        _delete_folder('is_primary')

    def test_is_primary_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_primary,
            *(None, None))
        self.assertEqual(self.api.is_primary('games', 'games'), True)
        self.assertEqual(self.api.is_primary('games', 'Site'), False)


class DBbase_is_primary_recno(unittest.TestCase):

    def setUp(self):
        self.api = base('is_primary_recno')

    def tearDown(self):
        _delete_folder('is_primary_recno')

    def test_is_primary_recno_01(self):
        self.assertEqual(
            self.api.is_primary_recno('Berkeley DB compatibility'), True)


class DBbase_is_recno(unittest.TestCase):

    def setUp(self):
        self.api = base('is_recno')

    def tearDown(self):
        _delete_folder('is_recno')

    def test_is_recno_01(self):
        self.assertRaisesRegex(
            KeyError,
            "None",
            self.api.is_recno,
            *(None, None))
        self.assertEqual(self.api.is_recno('games', 'games'), True)
        self.assertEqual(self.api.is_recno('games', 'Site'), False)


class DBbase_open_contexts(unittest.TestCase):

    def setUp(self):
        self.api = base('open_contexts')
        for v in self.api._dbdef.values():
            v.open_contexts = lambda a : None

    def tearDown(self):
        _delete_folder('open_contexts')

    def test_open_contexts_01(self):
        self.assertEqual(self.api.open_contexts('DPT compatibility'), None)


class DBbase_allocate_and_open_contexts(unittest.TestCase):

    def setUp(self):
        self.api = base('allocate_and_open_contexts')
        for v in self.api._dbdef.values():
            v.allocate_and_open_contexts = lambda a : None

    def tearDown(self):
        _delete_folder('allocate_and_open_contexts')

    def test_allocate_and_open_contexts_01(self):
        self.assertEqual(self.api.allocate_and_open_contexts(
            'DPT compatibility'), None)


class DBbase_get_packed_key(unittest.TestCase):

    def setUp(self):
        self.api = base('get_packed_key')

    def tearDown(self):
        _delete_folder('get_packed_key')

    def test_get_packed_key_01(self):
        instance = record.Record(record.Key, record.Value)
        self.assertEqual(
            self.api.get_packed_key('Berkeley DB compatibility',
                                    instance).decode(),
            repr(instance.key.pack()))


class DBbase_decode_as_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = base('decode_as_primary_key')

    def tearDown(self):
        _delete_folder('decode_as_primary_key')

    def test_decode_as_primary_key_01(self):
        #self.assertEqual(self.api.decode_as_primary_key(
        #    'Ignored in sqlite3', 10), 10)
        srkey = self.api.encode_record_number(3456)
        self.assertEqual(self.api.decode_as_primary_key(
            'Ignored in sqlite3', srkey), 3456)


class DBbase_encode_primary_key(unittest.TestCase):

    def setUp(self):
        self.api = base('encode_primary_key')

    def tearDown(self):
        _delete_folder('encode_primary_key')

    def test_encode_primary_key_01(self):
        instance = record.Record(record.KeyData, record.Value)
        instance.key.load(23)
        self.assertEqual(
            self.api.encode_primary_key('games', instance),
            b'23')


class DBbase_set_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = base('set_defer_update')
        for v in self.api._dbdef.values():
            v.close = lambda : None

    def tearDown(self):
        _delete_folder('set_defer_update')

    def test_set_defer_update_01(self):
        self.assertEqual(
            self.api.set_defer_update(duallowed='kkk'),
            'kkk')


class DBbase_unset_defer_update(unittest.TestCase):

    def setUp(self):
        self.api = base('unset_defer_update')
        self.api._DBenvironment = {'flags': (DB_CREATE |
                                             DB_INIT_MPOOL |
                                             DB_INIT_LOCK |
                                             DB_INIT_LOG |
                                             DB_INIT_TXN)}
        for v in self.api._dbdef.values():
            v.open_root = lambda a : None

    def tearDown(self):
        _delete_folder('unset_defer_update')

    def test_unset_defer_update_01(self):
        self.assertEqual(
            self.api.unset_defer_update(),
            True)


class DBbase_use_deferred_update_process(unittest.TestCase):

    def setUp(self):
        self.api = base('use_deferred_update_process')

    def tearDown(self):
        _delete_folder('use_deferred_update_process')

    def test_unset_defer_update_01(self):
        self.assertRaisesRegex(
            database.DatabaseError,
            "use_deferred_update_process not implemented",
            self.api.use_deferred_update_process,
            **{})


class DBbase_initial_database_size(unittest.TestCase):

    def setUp(self):
        self.api = base('initial_database_size')

    def tearDown(self):
        _delete_folder('initial_database_size')

    def test_initial_database_size_01(self):
        self.assertEqual(
            self.api.initial_database_size(),
            True)


class DBbase_increase_database_size(unittest.TestCase):

    def setUp(self):
        self.api = base('increase_database_size')

    def tearDown(self):
        _delete_folder('increase_database_size')

    def test_initial_database_size_01(self):
        self.assertEqual(
            self.api.increase_database_size(),
            None)


class DBbase_do_database_task(unittest.TestCase):

    def setUp(self):
        def f(obj, lw, **tma):
            pass
        self.f = f
        self.api = base('do_database_task')

    def tearDown(self):
        _delete_folder('do_database_task')

    def test_initial_database_size_01(self):
        self.assertEqual(
            self.api.do_database_task(self.f),
            None)


class DBbase_make_recordset_key(unittest.TestCase):

    def setUp(self):
        self.api = api('make_recordset_key')
        self.api.open_context()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('make_recordset_key')

    def test_make_recordset_key_01(self):
        self.assertIsInstance(
            self.api.make_recordset_key('games', 'games', key=1),
            recordset.Recordset)


class DBbase_make_recordset_key_startswith(unittest.TestCase):

    def setUp(self):
        self.api = api('make_recordset_key_startswith')
        self.api.open_context()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('make_recordset_key_startswith')

    def test_make_recordset_key_startswith_01(self):
        self.assertIsInstance(
            self.api.make_recordset_key_startswith('games', 'Site', key=b'a'),
            recordset.Recordset)


class DBbase_make_recordset_key_range(unittest.TestCase):

    def setUp(self):
        self.api = api('make_recordset_key_range')
        self.api.open_context()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('make_recordset_key_range')

    def test_make_recordset_key_range_01(self):
        self.assertIsInstance(
            self.api.make_recordset_key_range('games', 'Site', key=b'c'),
            recordset.Recordset)


class DBbase_make_recordset_all(unittest.TestCase):

    def setUp(self):
        self.api = api('make_recordset_all')
        self.api.open_context()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('make_recordset_all')

    def test_make_recordset_all_01(self):
        self.assertIsInstance(
            self.api.make_recordset_all('games', 'games'),
            recordset.Recordset)


class DBbase_file_records_under(unittest.TestCase):

    def setUp(self):
        self.api = api('file_records_under')
        self.api.open_context()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('file_records_under')

    def test_file_records_under_01(self):
        self.assertRaisesRegex(
            database.DatabaseError,
            "file_records_under not implemented for DBPrimary",
            self.api.file_records_under,
            *('games',
              'games',
              recordset.Recordset(dbhome=self.api, dbset='games'),
              None))
        self.assertEqual(
            self.api.file_records_under(
                'games',
                'Site',
                recordset.Recordset(dbhome=self.api, dbset='games'),
                None),
            None)


class DBbase_start_transaction(unittest.TestCase):

    def setUp(self):
        self.api = base('start_transaction')

    def tearDown(self):
        _delete_folder('start_transaction')

    def test_start_transaction_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'txn_begin'",
            self.api.start_transaction)


class DBbase_cede_contexts_to_process(unittest.TestCase):

    def setUp(self):
        self.api = base('cede_contexts_to_process')

    def tearDown(self):
        _delete_folder('cede_contexts_to_process')

    def test_cede_contexts_to_process_01(self):
        self.assertEqual(
            self.api.cede_contexts_to_process({}),
            None)


class DBbase_method_classes(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_method_classes_01(self):
        bit = dbapi.DBapi
        api = dbapi.DBbase
        base = database.Database
        enc = dbapi._DatabaseEncoders

        self.assertEqual(
            [a for a in dir(bit)
                 if not a.startswith('__') and not a.endswith('__')],
            ['allocate_and_open_contexts',
             'backout',
             'cede_contexts_to_process',
             'close_context',
             'close_contexts',
             'close_database',
             'commit',
             'create_recordset_cursor',
             'database_cursor',
             'database_specification',
             'db_compatibility_hack',
             'decode_as_primary_key',
             'decode_record_number',
             'deferred_update_housekeeping',
             'delete_instance',
             'do_database_task',
             'do_deferred_updates',
             'edit_instance',
             'encode_primary_key',
             'encode_record_key',
             'encode_record_number',
             'encode_record_selector',
             'exists',
             'file_records_under',
             'files_exist',
             'get_associated_indicies',
             'get_database',
             'get_database_filenames',
             'get_database_folder',
             'get_database_instance',
             'get_dbsets',
             'get_first_primary_key_for_index_key',
             'get_packed_key',
             'get_primary_record',
             'get_table_index',
             'get_table_indicies',
             'get_transaction',
             'increase_database_size',
             'initial_database_size',
             'is_primary',
             'is_primary_recno',
             'is_recno',
             'make_recordset',
             'make_recordset_all',
             'make_recordset_key',
             'make_recordset_key_like',
             'make_recordset_key_range',
             'make_recordset_key_startswith',
             'make_root',
             'open_context',
             'open_contexts',
             'put_instance',
             'record_finder',
             'record_selector',
             'repair_cursor',
             'set_defer_update',
             'start_transaction',
             'unfile_records_under',
             'unset_defer_update',
             'use_deferred_update_process',
             'values_finder',
             'values_selector'])
        self.assertEqual(
            len([a for a in dir(bit)
                 if not a.startswith('__') and not a.endswith('__')]),
            62)
        self.assertEqual(
            len([a for a in dir(api)
                 if not a.startswith('__') and not a.endswith('__')]),
            62)
        self.assertEqual(
            len([a for a in dir(base)
                 if not a.startswith('__') and not a.endswith('__')]),
            49)
        self.assertEqual(
            len([a for a in dir(enc)
                 if not a.startswith('__') and not a.endswith('__')]),
            4)

        # Methods inherited from DBbase
        self.assertIs(api.backout, bit.backout)
        self.assertIs(api.close_contexts, bit.close_contexts)
        self.assertIs(api.close_database, bit.close_database)
        self.assertIs(api.commit, bit.commit)
        self.assertIs(api.db_compatibility_hack, bit.db_compatibility_hack)
        self.assertIs(api.do_deferred_updates, bit.do_deferred_updates)
        self.assertIs(api.exists, bit.exists)
        self.assertIs(api.files_exist, bit.files_exist)
        self.assertIs(api.database_cursor, bit.database_cursor)
        self.assertIs(api.repair_cursor, bit.repair_cursor)
        self.assertIs(api.get_database_folder, bit.get_database_folder)
        self.assertIs(api.get_database, bit.get_database)
        self.assertIs(api.get_database_instance, bit.get_database_instance)
        self.assertIs(
            api.get_first_primary_key_for_index_key,
            bit.get_first_primary_key_for_index_key)
        self.assertIs(api.get_primary_record, bit.get_primary_record)
        self.assertIs(api.is_primary, bit.is_primary)
        self.assertIs(api.is_primary_recno, bit.is_primary_recno)
        self.assertIs(api.is_recno, bit.is_recno)
        self.assertIs(api.open_contexts, bit.open_contexts)
        self.assertIs(api.allocate_and_open_contexts,
                      bit.allocate_and_open_contexts)
        self.assertIs(api.get_packed_key, bit.get_packed_key)
        self.assertIs(api.decode_as_primary_key, bit.decode_as_primary_key)
        self.assertIs(api.encode_primary_key, bit.encode_primary_key)
        self.assertIs(api.set_defer_update, bit.set_defer_update)
        self.assertIs(api.unset_defer_update, bit.unset_defer_update)
        self.assertIs(
            api.use_deferred_update_process,
            bit.use_deferred_update_process)
        self.assertIs(api.initial_database_size, bit.initial_database_size)
        self.assertIs(api.increase_database_size, bit.increase_database_size)
        self.assertIs(api.do_database_task, bit.do_database_task)
        self.assertIs(api.make_recordset_key, bit.make_recordset_key)
        self.assertIs(
            api.make_recordset_key_startswith,
            bit.make_recordset_key_startswith)
        self.assertIs(
            api.make_recordset_key_range,
            bit.make_recordset_key_range)
        self.assertIs(api.make_recordset_all, bit.make_recordset_all)
        self.assertIs(api.file_records_under, bit.file_records_under)
        self.assertIs(api.start_transaction, bit.start_transaction)
        self.assertIs(api.cede_contexts_to_process,
                      bit.cede_contexts_to_process)
        
        # Methods inherited from _DatabaseEncoders
        self.assertIs(enc.encode_record_number, bit.encode_record_number)
        self.assertIs(enc.decode_record_number, bit.decode_record_number)
        self.assertIs(enc.encode_record_selector, bit.encode_record_selector)
        
        # Methods inherited from Database
        self.assertIs(base.deferred_update_housekeeping,
                      bit.deferred_update_housekeeping)

        # DBbase methods overridden
        self.assertIsNot(api.__init__, bit.__init__)
        self.assertIs(api.open_context, bit.open_context)
        self.assertIs(api.close_context, bit.close_context)
        self.assertIs(api.delete_instance, bit.delete_instance)
        self.assertIs(api.edit_instance, bit.edit_instance)
        self.assertIs(api.put_instance, bit.put_instance)

        # _DatabaseEncoders methods not overridden
        self.assertIs(enc.encode_record_key, bit.encode_record_key)


class DBapi_init(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___init___01(self):
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 3 required positional arguments: ",
                "'database_specification', 'databasefolder', ",
                "and 'DBenvironment'",
                )),
            dbapi.DBapi,
            )
        self.assertRaisesRegex(
            TypeError,
            "".join((
                "__init__\(\) missing 2 required positional arguments: ",
                "'databasefolder' and 'DBenvironment'",
                )),
            dbapi.DBapi,
            *(samplefilespec(),))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Database folder name None is not valid",
            dbapi.DBapi,
            *(samplefilespec(),),
            **dict(DBenvironment='', databasefolder=None))
        self.assertRaisesRegex(
            database.DatabaseError,
            "Database folder name None is not valid",
            dbapi.DBapi,
            *(samplefilespec(), None, ''))

    def test___init___02(self):
        api = dbapi.DBapi(
            filespec.FileSpec(
                t=dict(
                    primary='key',
                    fields=dict(lock={}, key={}, Skey={}),
                    filedesc=dict(fileorg=None),
                    secondary=dict(skey=None),
                    )),
            databasefolder='',
            DBenvironment=None)
        p = os.path.abspath('')
        self.assertEqual(api._databasefolder, p)
        self.assertEqual(api._home, p)
        self.assertEqual(api._dbenv, None)
        self.assertEqual(api._dbspec,
                         dict(t={'fields':{'key':{}, 'lock':{}, 'Skey':{}},
                                 'filedesc':{'fileorg':None, 'bsize':20,
                                             'brecppg':10, 'dsize':160},
                                 'primary':'key',
                                 'secondary':{'skey':None},
                                 'btod_factor':8,
                                 'default_records':200,
                                 })
                         )
        self.assertEqual(len(api._dbdef), 1)
        self.assertIsInstance(api._dbdef['t'], dbapi._DBDefinition)
        df = api._dbdef['t']
        self.assertEqual(df._dbset, 't')
        self.assertEqual(df.dbname_to_secondary_key, {'skey': 'Skey'})
        self.assertIsInstance(df.primary, dbapi.DBPrimary)
        self.assertIsInstance(df.secondary['Skey'], dbapi.DBSecondary)
        self.assertEqual(len(df.secondary), 1)
        self.assertEqual(len(df.__dict__), 4)
        self.assertEqual(len(api.__dict__), 8)
        self.assertIsInstance(
            dbapi.DBapi(samplefilespec(), '', None),
            dbapi.DBapi)
        self.assertIsInstance(
            dbapi.DBapi(database_specification=samplefilespec(),
                           databasefolder='',
                           DBenvironment=None),
            dbapi.DBapi)


class DBbase_encode_record_key(unittest.TestCase):

    def setUp(self):
        self.api = base('encode_record_key')

    def tearDown(self):
        _delete_folder('encode_record_key')

    def test_encode_record_key_01(self):
        self.assertRaisesRegex(
            AttributeError,
            "'NoneType' object has no attribute 'encode'",
            self.api.encode_record_key,
            *(None,))
        self.assertEqual(self.api.encode_record_key('key'), b'key')
        self.assertEqual(
            self.api.encode_record_key(repr(('d', 'e', 'f'))),
            b"('d', 'e', 'f')")


class DBbase_open_context(unittest.TestCase):

    def setUp(self):
        self.api = base('open_context',
                        environment={
                            'flags': DB_CREATE|DB_INIT_MPOOL|DB_INIT_TXN})

    def tearDown(self):
        _delete_folder('open_context')

    def test_open_context_01(self):
        self.assertEqual(self.api.open_context(), True)


class DBbase_close_context(unittest.TestCase):

    def setUp(self):
        self.api = base('close_context',
                        environment={
                            'flags': DB_CREATE|DB_INIT_MPOOL|DB_INIT_TXN})

    def tearDown(self):
        _delete_folder('close_context')

    def test_close_context_01(self):
        self.assertEqual(self.api.close_context(), None)

    def test_close_context_02(self):
        self.assertEqual(self.api.open_context(), True)
        self.assertEqual(self.api.close_context(), None)


class DBbase_get_database_filenames(unittest.TestCase):

    def setUp(self):
        self.api = base('api_get_database_filenames',
                        environment={
                            'flags': DB_CREATE|DB_INIT_MPOOL|DB_INIT_TXN})
        self.api.open_context()

    def tearDown(self):
        _delete_folder('api_get_database_filenames')

    def test_get_database_filenames_01(self):
        self.assertEqual(sorted(self.api.get_database_filenames()),
                         sorted(['___control',
                                 'games_Event',
                                 'games_Round',
                                 'games_Date',
                                 'games_White',
                                 'games_Name',
                                 'Game',
                                 'Game__exist',
                                 'Game__list',
                                 'Game__bits',
                                 'games_Black',
                                 'games_Result',
                                 'games_Site',
                                 'log.0000000001',
                                 ]))


class DBapi_put_instance(unittest.TestCase):

    def setUp(self):
        self.api = api('put_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('put_instance')

    def test_put_instance_05(self):
        database_interface.test_put_instance_05(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_06(self):
        database_interface.test_put_instance_06(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_07(self):
        database_interface.test_put_instance_07(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_08(self):
        database_interface.test_put_instance_08(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_09(self):
        database_interface.test_put_instance_09(self,
                                                collect_counts,
                                                game_number_to_record_number)

    def test_put_instance_10(self):
        database_interface.test_put_instance_10(self,
                                                collect_counts,
                                                game_number_to_record_number)


class DBapi_delete_instance(unittest.TestCase):

    def setUp(self):
        self.api = api('delete_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('delete_instance')

    def test_delete_instance_04(self):
        database_interface.test_delete_instance_01(self,
                                                   collect_counts,
                                                   game_number_to_record_number)

    def test_delete_instance_05(self):
        database_interface.test_delete_instance_02(self,
                                                   collect_counts,
                                                   game_number_to_record_number)


class DBapi_edit_instance(unittest.TestCase):

    def setUp(self):
        self.api = api('edit_instance')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('edit_instance')

    def test_edit_instance_06(self):
        database_interface.test_edit_instance_06(self,
                                                 collect_counts,
                                                 game_number_to_record_number)

    def test_edit_instance_07(self):
        database_interface.test_edit_instance_07(self,
                                                 collect_counts,
                                                 game_number_to_record_number)


class DBapi_put_instance_with_holes(unittest.TestCase):

    def setUp(self):

        # DPT segment size is 65280 (8k bytes - 32 reserved) while apsw, bsddb3,
        # and sqlite3, segment size is 32768 (4k bytes).  Reuse record number
        # tests cannot produce the same answer in general on the two segment
        # sizes, even if the DPT algorithm deciding when to fill a hole is the
        # one used for apsw, bsddb3, and sqlite3.
        # The api() function calls samplefilespec with the default fileorg
        # argument because DPT is not the database engine.
        self.dpt_reuse_record_numbers = False
        
        self.api = api('put_instance_holes')
        self.sr = SampleRecord()

    def tearDown(self):
        self.api.close_database()
        _delete_folder('put_instance_holes')

    def test_holes_01(self):
        database_interface.test_holes_01(self,
                                         collect_counts,
                                         game_number_to_record_number)


def game_number_to_record_number(game_number):
    # Record numbers start at 1 and game numbers start at 1
    return game_number


def record_number_to_game_number(record_number):
    # Record numbers start at 1 and game numbers start at 1
    return record_number


def collect_counts(testcase,
                   record_adder,
                   data_adder,
                   answers,
                   counts,
                   segments,
                   recordsets,
                   rowexceptions):
    list_seg_length = constants.LENGTH_SEGMENT_LIST_REFERENCE
    bits_seg_length = constants.LENGTH_SEGMENT_BITARRAY_REFERENCE
    files = dict(Round='games_Round',
                 White='games_White',
                 Event='games_Event',
                 Site='games_Site',
                 Black='games_Black',
                 Date='games_Date',
                 Result='games_Result',
                 )
    db_directory = testcase.api.get_database_folder()
    dbenv = DBEnv()
    dbenv.open(db_directory)
    filespec = samplefilespec()
    primaries = {filespec[t][PRIMARY]:t for t in filespec}
    for t in answers['data']:
        testcase.assertEqual(t in primaries, True)
        segments[t] = dict()
        rowexceptions[t] = dict()
        db = DB(dbenv)
        db.open(t, dbname=t, flags=DB_RDONLY)
        stat = db.stat()
        testcase.assertEqual(stat['ndata'], answers['record_counts'][t])
        cursor = db.cursor()
        while True:
            r = cursor.next()
            if not r:
                break
            # Allow for place holder null record for reuse record number
            if r[1] is None:
                continue
            data_adder(rowexceptions[t],
                       segments[t],
                       (r[0], r[1].decode()),
                       answers['data'][t]['defaultrow'],
                       record_number_to_game_number)
        cursor.close()
        db.close()
    for t in answers['data']:
        for u in answers['records']:
            if u not in filespec[primaries[t]][SECONDARY]:
                continue
            segments[u] = dict()
            recordsets[u] = dict()
            db = DB(dbenv)
            db.open(files.get(u), dbname=u, flags=DB_RDONLY)
            ubits = ''.join((t, '__bits'))
            dbbits = DB(dbenv)
            dbbits.open(ubits, dbname=ubits, flags=DB_RDONLY)
            ulist = ''.join((t, '__list'))
            dblist = DB(dbenv)
            dblist.open(ulist, dbname=ulist, flags=DB_RDONLY)
            uexist = ''.join((t, '__exist'))
            cursor = db.cursor()
            while True:
                r = cursor.next()
                if not r:
                    break
                rk, rv = r
                if len(r[1]) == list_seg_length:
                    # List of record numbers.
                    srn = int.from_bytes(rv[6:], byteorder='big')
                    seg = int.from_bytes(rv[:4], byteorder='big')
                    rct = int.from_bytes(rv[4:6], byteorder='big')
                    slist = dblist.get(srn)
                    segbase = seg * SegmentSize.db_segment_size
                    for i in range(0, len(slist), 2):
                        record_adder(
                            recordsets[u],
                            segments[u],
                            (rk.decode(),
                             segbase + int.from_bytes(slist[i:i+2], 'big')),
                            record_number_to_game_number)
                elif len(r[1]) == bits_seg_length:
                    # Bitmap of record numbers.
                    srn = int.from_bytes(rv[7:], byteorder='big')
                    seg = int.from_bytes(rv[:4], byteorder='big')
                    rct = int.from_bytes(rv[4:7], byteorder='big')
                    sbits = dbbits.get(srn)
                    segbase = seg * SegmentSize.db_segment_size
                    for e, b in enumerate(sbits):
                        if not b:
                            continue
                        for eb, bitvalue in enumerate(
                            (128, 64, 32, 16, 8, 4, 2, 1)):
                            if bitvalue & b:
                                record_adder(
                                    recordsets[u],
                                    segments[u],
                                    (rk.decode(), segbase + eb + e * 8),
                                    record_number_to_game_number)
                else:
                    # Direct record number.
                    srn = int.from_bytes(rv[4:], byteorder='big')
                    seg = int.from_bytes(rv[:4], byteorder='big')
                    rct = 1
                    segbase = seg * SegmentSize.db_segment_size
                    record_adder(recordsets[u],
                                 segments[u],
                                 (rk.decode(),
                                  seg * SegmentSize.db_segment_size + srn),
                                 record_number_to_game_number)
            cursor.close()
            dblist.close()
            dbbits.close()
            db.close()
    dbenv.close()


if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    
    runner().run(loader(SampleRecordAssumptions))
    
    runner().run(loader(_DatabaseEncoders))

    # The DBbase class tests.

    runner().run(loader(DBbase_init))

    runner().run(loader(DBbase_open_context))
    runner().run(loader(DBbase_close_context))
    runner().run(loader(DBbase_close_database))
    runner().run(loader(DBbase_get_database_filenames))

    runner().run(loader(DBbase_backout))
    runner().run(loader(DBbase_close_contexts))
    runner().run(loader(DBbase_commit))
    runner().run(loader(DBbase_db_compatibility_hack))
    runner().run(loader(DBbase_exists))
    runner().run(loader(DBbase_files_exist))
    runner().run(loader(DBbase_database_cursor))
    runner().run(loader(DBbase_repair_cursor))
    runner().run(loader(DBbase_get_database_folder))
    runner().run(loader(DBbase_get_database))
    runner().run(loader(DBbase_get_database_instance))
    runner().run(loader(DBbase_get_first_primary_key_for_index_key))
    runner().run(loader(DBbase_get_primary_record))
    runner().run(loader(DBbase_is_primary))
    runner().run(loader(DBbase_is_primary_recno))
    runner().run(loader(DBbase_is_recno))
    runner().run(loader(DBbase_open_contexts))
    runner().run(loader(DBbase_allocate_and_open_contexts))
    runner().run(loader(DBbase_get_packed_key))
    runner().run(loader(DBbase_decode_as_primary_key))
    runner().run(loader(DBbase_encode_primary_key))
    runner().run(loader(DBbase_use_deferred_update_process))
    runner().run(loader(DBbase_initial_database_size))
    runner().run(loader(DBbase_increase_database_size))
    runner().run(loader(DBbase_do_database_task))
    runner().run(loader(DBbase_make_recordset_key))
    runner().run(loader(DBbase_make_recordset_key_startswith))
    runner().run(loader(DBbase_make_recordset_key_range))
    runner().run(loader(DBbase_make_recordset_all))
    runner().run(loader(DBbase_file_records_under))
    runner().run(loader(DBbase_start_transaction))
    runner().run(loader(DBbase_cede_contexts_to_process))

    runner().run(loader(DBbase_method_classes))
    runner().run(loader(DBbase_encode_record_key))
    runner().run(loader(DBbase_set_defer_update))
    runner().run(loader(DBbase_unset_defer_update))
    runner().run(loader(DBbase_do_deferred_updates))

    # The DBapi class tests. These take a total of nearly 15 minutes.

    runner().run(loader(DBapi_init))

    runner().run(loader(DBapi_put_instance))
    runner().run(loader(DBapi_delete_instance))
    runner().run(loader(DBapi_edit_instance))

    runner().run(loader(DBapi_put_instance_with_holes))
